/*********************************************************************
*      This file is part of the source code of SYNAPS kernel.        *
*   Author(s): B. Mourrain, GALAAD, INRIA                            *
**********************************************************************
History: 
$Id: FFT.h,v 1.1 2005/07/11 08:33:33 mourrain Exp $
**********************************************************************/
#ifndef SYNAPS_LINALG_FFT_H
#define SYNAPS_LINALG_FFT_H
#include <complex>
#include <synaps/init.h>

__BEGIN_NAMESPACE_SYNAPS
//====================================================================
/** \file FFT.h 
    \brief Generic Fast Fourier Transform.
**/
//----------------------------------------------------------------------
/// Root of Unity. Replace @c@ by the primitive root of order @n@.
template<class T> inline void Root_Unity(std::complex<T> &c, int n) {
  c = std::complex<T>(cos(8.0*atan(1.0)/n), sin(8.0*atan(1.0)/n));
}
//----------------------------------------------------------------------
/**
   Functions which permute an  array 
   by Bit Reversal method.
**/
inline unsigned int BitReverse(int k, int n)
{
  int s, j = 0, m = k;
  double l =  std::log((double)n)/std::log((double)2);
  for (int q = 0; q < l; q++)
    {
      s = (int) std::floor(m/2.);
      j = 2*j + (m - 2*s);
      m = s;
    }
  return j;
}
//----------------------------------------------------------------------
/// Shuflle on N entries.
template <class T>
inline void Permut(T *Tab, int N)
{
  int j = 0;
  for (int k = 0; k < N; k++)
    {
      j = BitReverse(k, N);
      if (j > k)
	std::swap(Tab[j], Tab[k]);
    }
}
//-------------------------------------------------------------------
/**
  Implementation of {\tt FFT } computation of a complex array.
  It is an "in-place" transformation, based on the Butterfly algorithm.
  A complex array W, of size n (a power of 2), is replaced by
  the FFT of its values. The complex number @w@ is a primitive 
  root of unity of order $n$.
**/
template <class C, class U>
void FFT(int n, const std::complex<U> & w,  std::complex<C>  *W, int inverse)
{
  switch (n) {
  case 1:
    break;
  case 2: 
    {
      std::complex<C> t(W[0]);
      W[0] = t + W[1];
      W[1] = t - W[1];
    }
    break;
  case 4:
    {
       std::complex<C> x0(W[0]+W[1]), x1(W[0]-W[1]), y0(W[2]+W[3]), y1(W[2]-W[3]);
      if (inverse) {
       	W[0] = (x0 + y0);
	W[1] = (x1 - std::complex<C>(-y1.imag(), y1.real()));
	W[2] = (x0 - y0);
	W[3] =(x1 + std::complex<C>(-y1.imag(), y1.real()));}
      else {
       	W[0] = x0 + y0;
	W[1] = x1 + std::complex<C>(-y1.imag(), y1.real());
	W[2] = x0 - y0;
	W[3] = x1 - std::complex<C>(-y1.imag(), y1.real()); }
      }
    break;
  default:
  { 
    /* Since the array w has been permutated before the call to FFT,
       elements with even index in the initial array, are now the
       first n/2 elements and elements with odd index the last n/2.
       So the recursive process takes the initial array divided in two parts.
    */

    FFT(n/2,pow(w, 2), W, inverse);
    FFT(n/2, pow(w, 2), W+n/2,inverse);
    
    for (int k = 0; k< n/2; k++)
      {
	std::complex<C> z = pow(w,k)*W[k+n/2] ;
	W[k+n/2] = W[k] - z;
	W[k] = W[k] + z;
      }
  }
  }
}
//-----------------------------------------------------------------------------
__END_NAMESPACE_SYNAPS
#endif // SYNAPS_LINALG_FFT_H

