/**********************************************************************
 *   This file is part of the source code of the SYNAPS library.      *
 *   Author(s): H. Khalil, B. Mourrain, GALAAD, INRIA                 *
 *   $Id: bivariate_res.H,v 1.3 2005/09/28 06:30:39 mourrain Exp $
 **********************************************************************/
/** \file                                                            **/
#ifndef synaps_solve_bivariate_res_H
#define synaps_solve_bivariate_res_H
//----------------------------------------------------------------------
#include <synaps/init.h>
#include <synaps/base/Seq.h>
#include <synaps/arithm/traits.h>
#include <synaps/linalg/VectDse.h>
#include <synaps/linalg/MatrDse.h>
#include <synaps/linalg/lapack.h>
#include <synaps/linalg/Svd.h>
#include <synaps/linalg/Eigen.h>
#include <synaps/upol/resultant.h>
#include <synaps/mpol/bezoutian.h>
#include <synaps/upol.h>
#include <synaps/mpol.h>

#include <cmath>

//----------------------------------------------------------------------
__BEGIN_NAMESPACE_SYNAPS

template<class C>
struct EigenRes
{
  C epsilon;
  EigenRes(): epsilon(1.e-6) {}
  EigenRes(const C & eps): epsilon(eps) {}

};


using std::abs;


struct no_multiplicity
{
	void push_back(int i, int j) {}
};

template <class MU>
struct with_x_multiplicity
{
	MU & mu;
	with_x_multiplicity(MU & m): mu(m) {}
	void push_back(int i, int j) {mu.push_back(i);}
};

template <class MU>
struct with_xy_multiplicity
{
	MU & mux; MU& muy;
	with_xy_multiplicity(MU & mx, MU & my): mux(mx), muy(my) {}
	void push_back(int i, int j) 
	{
		mux.push_back(i);
		muy.push_back(j);
	}
};


/**
  \param Matrix a matrix.
  \return the maximum degree of a coefficient of the matrix \c Matrix.
 **/
	template<class MAT> 
int max_degree(const MAT & Matrix)
{
	int n= Matrix.nbcol(),result=0; 
	for (int i=0;i<=n-1;++i)
		for ( int j=0;j<=n-1;++j)
		{
			if (degree(Matrix(i,j)) >result)
				result=degree(Matrix(i,j));    
		}
	return result;
}
/**
  \param mat_eigen the list of matrices.
  \param M the initial matrix.
  \param d the number of matrix.
  \param n the size of the matrix.

  Creates a list of \c d+1 matrix of size \c nxn, which indices are 0 to \c d 
  It representes the matrix S0,..,Sd.
 **/
	template<class matrix_t,class MAT>
void init_list_matrix(matrix_t *mat_eigen, MAT & M, int d, int n)
{
	for(int i=0;i<d+1;++i)
		mat_eigen[i]=matrix_t(n,n);
	for ( int i=0;i<=n-1;++i)
		for (int j=0;j<=n-1;++j)
			for (int k=0;k<=d;++k)
			{
				typename MAT::value_type tmp=M(i,j);
				if ( degree(tmp)<k ) 
					mat_eigen[k](i,j)=0;
				else mat_eigen[k](i,j)=tmp[k];
			}
}

/**
  \param Mat the initial list of matrices.
  \param B the matrix result.
  \param N the size of the matrix \c B .
  \param n the size of the matrices in \c Mat .
  \param d the number of matrices in \c Mat .

  Initialization of the matrix \c B starting from the matrix \c Mat. 
 **/
template<class matrix_t>
void initialization_B(matrix_t *Mat, matrix_t & B,
		int N, int n, int d)
{
  for ( int i=0;i<=N-1-n;++i)
    B(i,i)=1;
  for ( int i=N-n;i<=N-1;++i)
    for (int j=(d-1)*n;j<=d*n-1;++j)
      B(i,j)=-Mat[d](i-N+n,j-(d-1)*n);
}

/**
  \param Mat the initial list of matrices.
  \param A the matrix result.
  \param N the size of the matrix \c A .
  \param n the size of the matrices in \c Mat .
  \param d the number of matrices in \c Mat .

  Initialization of the matrix \c A starting from the matrix \c Mat. 
 **/
template<class matrix_t>
void initialization_A(matrix_t *Mat, matrix_t & A,
		      int N, int n, int d)
{
  for ( int i=0;i<=N-1-n;++i)
    A(i,i+n)=1;
  for ( int k=0;k<=d-1;++k)
    for(int i=N-n;i<=N-1;++i)
      for(int j=k*n;j<=(k+1)*n-1;++j)
	A(i,j)=Mat[k](i-N+n ,j-k*n);
}



/* cette fonction donne le noyau de Sylvester(a[0]) si le rang de la matrice
   des vecteurs propres asocies a a[0], est plus grand que la taille de la 
   matrice de bezout */ 
	template<class matrix_t,class pol_t>
matrix_t syl(pol_t p,pol_t q,double a)
{
	typedef typename pol_t::coeff_t coeff_t;
	double epsilon=0.00001;
	MatrDse<UPolDse<coeff_t> > M = u_sylvester<UPolDse<coeff_t> > (p,q,1);
	int n = M.nbcol();
	matrix_t matrix(n,n);
	for(int i=0;i<n;++i)
		for(int j=0;j<n;++j)
			matrix(i,j)=M(i,j)(a);
	matrix_t U,V;
	VectDse<coeff_t> s=Svd(matrix,U,V);
	int ord=1;
	for(int i=1;i<n;++i)
		if (s[i]>epsilon)
			ord=ord+1;
	int N=n-ord;
	matrix_t kernel(n,N);
	for(int i=0;i<n;++i)
		for(int j=ord;j<n;++j)
			kernel(i,j-ord)=V(j,i);
	return kernel;
}


template<class vector_cpx_t, class matrix_cpx_t, class pol_t>
bool eigen_from_pol(vector_cpx_t & eigval, 
		matrix_cpx_t & eigvec,
		const pol_t & Pol1, const pol_t & Pol2)
{
	typedef typename pol_t::coeff_t coeff_t;
	typedef std::complex<coeff_t> complex_t;
	typedef MatrDse<coeff_t,lapack::rep2d<coeff_t> > matrix_t;
	typedef MatrDse<UPolDse<coeff_t> > UPolMat_t;

	std::cout.precision(5);
	UPolMat_t M = UBezout< UPolDse<coeff_t> > (Pol1,Pol2);

	//  int rb=M.nbcol();
	int n= M.nbcol(), d=max_degree(M), N=n*d;

	if(d) 
	  {
	    /* Creates a list of d+1 matrix of size nxn, which indices 
	       are 0 to d. It representes the matrix S0,..,Sd. */
	    matrix_t* Mat= new matrix_t[d+1];
	    init_list_matrix(Mat,M,d,n);
	    /* Initialization of two matrix A and B */
	    matrix_t A(N,N), B(N,N);
	    initialization_A(Mat,A,N,n,d);
	    //	    cout <<A<<endl;
	    initialization_B(Mat,B,N,n,d);
	    //	    cout <<B<<endl;
	    /* Calcul des valeurs propres generalisees de A -x*B. */
	    eigval=Eigenval(A,B);
	    //	    cout <<eigval<<endl;
	    /* Calcul des vecteurs propres generalises de A -x*B. */
	    eigvec=Eigenvct(A,B);
	    return true;
	  } 
	else
	  return false;
}

/* calcul des points d'intersection 
   a est le vecteur des valeurs propres(les abscisses)
   b est la matrice des vecteurs propres correspondants 
   epsilon est la prescision */
template<class vector_cpx_t, class matrix_cpx_t, class  pol_t, 
	class MU>
	Seq<VectDse<typename pol_t::coeff_t> >
	roots_from_eigen(vector_cpx_t&  a, matrix_cpx_t& b, 
			const pol_t & p, const pol_t & q, 
			double epsilon,
			MU mu
			)
{
	typedef typename pol_t::coeff_t coeff_t;
	typedef MatrDse<coeff_t,lapack::rep2d<coeff_t> > matrix_t;
	typedef VectDse<coeff_t> vector_t;
	typedef VectDse<coeff_t> point_t;

	point_t point(2,AsSize());

	Seq<point_t> res;

	int n=size(a), m=b.nbrow(), nb=1, j, comp;
	int rb=std::max(degree(p,0),degree(q,0));;

	/* debarras des valeurs infinis et imaginaires */
	for (int i=0;i<n;++i)
	{
		if(abs(real(a[i])) > 1000 || abs(imag(a[i])) > epsilon )
		{
		  vector_cpx_t v(n-1,AsSize());
			matrix_cpx_t M(m,n-1);
			for(int l=0;l<i;++l)
			{
				v[l]=a[l];
				for(int k=0;k<m;++k)
					M(k,l)=b(k,l);
			}
			for(int l=i+1;l<n;++l)
			{
				v[l-1]=a[l];
				for(int k=0;k<m;++k)
					M(k,l-1)=b(k,l);
			}
			n=n-1;i=i-1;
			a=v;b=M;
		}
	}


	n=size(a);

	/* c'est une boucle jusqu'a la fin de la fonction, qui assure le travail
	   avec chaque valeur propre (a[0] avec les autres valeurs proches d'elle)
	   a part */
	while (n>0)
	{
		/* j est la multiplicite de a[0]. 
		   on prend la moyenne des points proches de a[0] et on 
		   rassemble les vecteurs propres correspondants dans la matrice d.
		   Et on soustrait du vecteur a a[0] avec les points proche d'elle, et
		   de b les vecteurs correspondants */ 
		j=1; 
		for (int i=1;i<n;++i)
			if ( abs( real(a[i]) - real(a[0]) ) < epsilon )
				j=j+1;
		matrix_cpx_t d(m,j);
		comp=-1;
		double a0=real(a[0]);
		for (int i=0;i<n;++i)
			if ( abs( real(a[i]) - a0 ) < epsilon )
			{
				comp=comp+1;
				a0=(a0+real(a[i]))/2;
				for (int k=0;k<m;++k)
					d(k,comp)=b(k,i);	    
				vector_cpx_t v(n-1);
				matrix_cpx_t ma(m,n-1);
				for (int l=0;l<i;++l)
				{
					v[l]=a[l];
					for (int e=0;e<m;++e)
						ma(e,l)=b(e,l);
				}
				for (int l=i+1;l<n;++l)
				{
					v[l-1]=a[l];
					for (int e=0;e<m;++e)
						ma(e,l-1)=b(e,l);
				}
				n=n-1;i=i-1;a=v;b=ma;
			}




		/*Si a[0] est simple cad j=1 on a le point d'intersection 
		  (a[0],d(1)/d(0))*/
		if (j==1)
		{
			if ( abs(imag( d(1,0)/d(0,0) ))<epsilon )
			{
				point[0]=a0;point[1]=real(d(1,0))/real(d(0,0));
				res.push_back(point);
				mu.push_back(1,1);
				//std::cout<<"(x"<<nb<<","<<"y"<<nb<<") = "<<"("<<a0<<","<<real(d(1,0))/real(d(0,0))<<")"<<"       de multiplicite 1"<<endl;
				nb=nb+1;
			} 
		}

		/* Sinon cad a[0] est multiple */
		else
		{
			matrix_t D(m,j),U,V;
			/* D = real(d), ca change rien car si le calcul est exact, 
			   on'a d est une matrice reelle, on calcule D pour utiliser le svd */
			for (int i=0;i<m;++i)
				for (int k=0;k<j;++k)
					D(i,k)=real(d(i,k));       
			vector_t s=Svd(D,U,V);   

			/* ord est le rang de D la matrice des vecteurs propres 
			   associes a a[0] */
			int ord=1;
			for (int i=1;i<j;++i)
				if (s[i]>epsilon)
					ord=ord+1;                

			/* si ord=1 on'a qu'un seul vecteur propre associe a a[0] */
			if (ord==1)
			{
				point[0]=a0;point[1]=real(d(1,0))/real(d(0,0));
				res.push_back(point);
				mu.push_back(j,1);
				//std::cout<<"(x"<<nb<<","<<"y"<<nb<<") = "<<"("<<a0<<","<<real(d(1,0))/real(d(0,0))<<")"<<"       de multiplicite " <<j<<endl;
				nb=nb+1;
			}

			/* Sinon cad on'a plus qu'un vecteur propre libre, alors
			   on calcule D0 et D1 en utilisant D*V */
			else
			{
				int teste=0;
				/* si ord >= rb=taille de la matrice de Bezout, alors on peut
				   pas utiliser D pour calculer D0 et D1, pour cela on utilse
				   la fonction syl*/
				if(ord>=rb)
				{
					D=syl<matrix_t>(p,q,a0);
					ord=D.nbcol();
					if(ord!=j)
						teste=1;	
				}
				else
				{
					if (ord!=j)
					{
						teste=1;
						D=D*V; 
					}
				}            
				matrix_t D0(ord,ord), D1(ord,ord);
				D0=D(Range(0,(ord-1)),Range(0,(ord-1)));
				D1=D(Range(1,ord),Range(0,(ord-1)));
				/* y est le vecteur des valeurs propres generalisees de (D1,D0)
				   cad les ordonnees des points d'int. au dessus de a[0] 
				   il faut aussi debarrasser des valeurs immaginaires, et
				   rassembler les points proches */
				vector_cpx_t y=Eigenval(D1,D0);                                   	      int n1=size(y);
				int j1=0;
				for (int i=0;i<n1;++i)
					if(abs(imag(y[i]))<2*epsilon)
						j1=j1+1;
				if(j1<n1)
					teste=1;
				vector_t t(j1);
				comp=0;
				for(int i=0; i<n1; ++i)
					if ( abs( imag(y[i]) ) < 2*epsilon )
					{
						t[i]=real(y[i]);
						comp=comp+1;
					}
				n1=size(t);
				int n3=n1;
				double p;
				int n2=0;
				while ( n1>0 )
				{
					comp=0;
					p=t[0];
					for (int i=0;i<n1;++i)
						if ( abs(t[i]-p)<epsilon )
						{
							comp=comp+1;
							n1=n1-1;
							p=(p+t[i])/2;
							vector_t v(n1);
							for (int k=0;k<i;++k)
								v[k]=t[k];
							for (int k=i+1;k<=n1;++k)
								v[k-1]=t[k];
							i=i-1;
							t=v;
						} 
					if (comp==n3) 
						n2=1;
					if(n2==1) 
					{
						teste=0;
						point[0]=a0;point[1]=p;
						res.push_back(point);
						mu.push_back(j,1);
						//cout<<"(x"<<nb<<","<<"y"<<nb<<") = "<<"("<<a0<<","<<p<<")"<<"       de multiplicite "<<j<<endl;
					}
					else
					{
						if(teste==1)
						{
							point[0]=a0;point[1]=p;
							res.push_back(point);
							mu.push_back(j,1);
							//cout<<"(x"<<nb<<","<<"y"<<nb<<") = "<<"("<<a0<<","<<p<<")"<<"       (multiplicite en x="<<j<<", multiplicite en y="<<comp<<")"<<endl;
						}
						else
						{
							point[0]=a0;point[1]=p;
							res.push_back(point);
							mu.push_back(comp,1);
							//cout<<"(x"<<nb<<","<<"y"<<nb<<") = "<<"("<<a0<<","<<p<<")"<<"       de multiplicite "<<comp<<endl;
						}
					}
					/*  cout<<"(x"<<nb<<","<<"y"<<nb<<") = "<<"("<<a0<<","<<p<<")"<<"       (multiplicite en x="<<j<<", multiplicite en y="<<comp<<")"<<endl;*/
					nb=nb+1;
				}
			}
		}
	}
	return res;
}

/** Compute the real roots of two bivariate polynomials, by eigencomputation
  from their Bezout resultant matrix.
  \param  Pol1,Pol2 are the input bivariate polynomials
  \return a sequence of arrays containing the first and second coordinates of the real roots of \c Pol1=0, \c Pol2 =0.
 **/
template<class pol_t>
Seq<VectDse<typename pol_t::coeff_t> >
solve(const pol_t & p, const pol_t & q,
      EigenRes<typename pol_t::coeff_t> mthd)
{
  typedef typename pol_t::coeff_t coeff_t;
  typedef std::complex<coeff_t> complex_t;


  VectDse<complex_t> eigval;
  MatrDse<complex_t,lapack::rep2d<complex_t> > eigvec;
  
  //synaps::eigen_from_pol(eigval,eigvec,p,q);
  //return synaps::roots_from_eigen(eigval,eigvec,p,q, mthd.epsilon,
  //		synaps::no_multiplicity());
  bool have_sol=eigen_from_pol(eigval,eigvec,p,q);
  if(have_sol)
    return roots_from_eigen(eigval,eigvec,p,q, mthd.epsilon, 
			    no_multiplicity());
  else
    return Seq<VectDse<typename pol_t::coeff_t> >();
}

/** Same function.
  Store the multiplicity of the resultant as a polynomial in the first
  variables, in \c mu. Generically, this corresponds to the multiplicity
  of intersection of \c p=q=0.
 **/
template<class pol_t>
Seq<VectDse<typename pol_t::coeff_t> >
solve(const pol_t & p, const pol_t & q, Seq<int> & mu,
      EigenRes<typename pol_t::coeff_t> mthd )
{
  typedef typename pol_t::coeff_t coeff_t;
  typedef std::complex<coeff_t> complex_t;
  
  VectDse<complex_t> eigval;
  MatrDse<complex_t,lapack::rep2d<complex_t> > eigvec;

  // synaps::eigen_from_pol(eigval,eigvec,p,q);
  // return synaps::roots_from_eigen(eigval,eigvec,p,q, mthd.epsilon,
  //		synaps::with_x_multiplicity<Seq<int> >(mu));
  bool have_sol=eigen_from_pol(eigval,eigvec,p,q);
  if(have_sol)
    return roots_from_eigen(eigval,eigvec,p,q, mthd.epsilon, 
			    with_x_multiplicity< Seq<int> >(mu));
  else
    return Seq<VectDse<typename pol_t::coeff_t> >();
}

/** Same function.
  Store the multiplicity of the resultant as a polynomial in the first
  variables, in \c mux and the multiplicity on the fiber, in \c muy.
 **/
template<class pol_t>
Seq<VectDse<typename pol_t::coeff_t> >
solve(const pol_t & p, const pol_t & q,
      Seq<int> & mux, Seq<int> & muy,
      EigenRes<typename pol_t::coeff_t> mthd )
{
	typedef typename pol_t::coeff_t coeff_t;
	typedef std::complex<coeff_t> complex_t;

	VectDse<complex_t> eigval;
	MatrDse<complex_t,lapack::rep2d<complex_t> > eigvec;

	// synaps::eigen_from_pol(eigval,eigvec,p,q);
	// return synaps::roots_from_eigen(eigval,eigvec,p,q, mthd.epsilon,
	//		synaps::with_xy_multiplicity<Seq<int> >(mux,muy));
  bool have_sol=eigen_from_pol(eigval,eigvec,p,q);
  if(have_sol)
    return roots_from_eigen(eigval,eigvec,p,q, mthd.epsilon, 
			    with_xy_multiplicity<Seq<int> >(mux,muy));
  else
    return Seq<VectDse<typename pol_t::coeff_t> >();
}

__END_NAMESPACE_SYNAPS
//====================================================================
#endif // synaps_solve_bivariate_res_H

