/********************************************************************
 *   This file is part of the source code of the SYNAPS library.
 *   Author(s): J.P. Pavone, GALAAD, INRIA
 *   $Id: vstack.h,v 1.1 2005/07/11 10:03:57 jppavone Exp $
 ********************************************************************/
#ifndef SYNAPS_SOLVE_SBDSLV_VSTACK_HPP
#define SYNAPS_SOLVE_SBDSLV_VSTACK_HPP
//--------------------------------------------------------------------
#include <synaps/init.h>
#include "linear_allocator.h"

//--------------------------------------------------------------------
__BEGIN_NAMESPACE_SYNAPS
//--------------------------------------------------------------------
namespace sbdslv
{
  template< class creal_t >
  struct vstack /* pile de projection centrée variable */
  {
    typedef unsigned sz_t;
    /* nombre de projections disponibles */
    sz_t m_nbp; 
    inline sz_t nbp() const { return m_nbp; };
    /* allocateur des coefficients et des projections */
    memory::linear_allocator<creal_t> m_cpchnk;
    memory::linear_allocator<sz_t>    m_ipchnk;
    vstack(): m_nbp(0) {};
    /* accesseurs pour les projections */
    inline       sz_t* pinf( sz_t iptr )       { return m_ipchnk.data(iptr); };
    inline const sz_t* pinf( sz_t iptr ) const { return m_ipchnk.data(iptr); };
    inline       sz_t& pmsz( sz_t iptr )       { return pinf(iptr)[0]; };
    inline       sz_t  pmsz( sz_t iptr ) const { return pinf(iptr)[0]; };
    inline       sz_t& pmad( sz_t iptr )       { return pinf(iptr)[1]; };
    inline       sz_t  pmad( sz_t iptr ) const { return pinf(iptr)[1]; };
    inline       sz_t& pMsz( sz_t iptr )       { return pinf(iptr)[2]; };
    inline       sz_t  pMsz( sz_t iptr ) const { return pinf(iptr)[2]; };
    inline       sz_t& pMad( sz_t iptr )       { return pinf(iptr)[3]; };
    inline       sz_t  pMad( sz_t iptr ) const { return pinf(iptr)[3]; };
    
    inline       creal_t* pmdata( sz_t iptr )       { return m_cpchnk.data(pmad(iptr)); };
    inline const creal_t* pmdata( sz_t iptr ) const { return m_cpchnk.data(pmad(iptr)); };
    inline       creal_t* pMdata( sz_t iptr )       { return m_cpchnk.data(pMad(iptr)); };
    inline const creal_t* pMdata( sz_t iptr ) const { return m_cpchnk.data(pMad(iptr)); };
    
    inline       creal_t& plower( sz_t iptr )       { return *(pmdata(iptr)-2); };
    inline const creal_t& plower( sz_t iptr ) const { return *(pmdata(iptr)-2); };
    inline       creal_t& pupper( sz_t iptr )       { return *(pmdata(iptr)-1); };
    inline const creal_t& pupper( sz_t iptr ) const { return *(pmdata(iptr)-1); };
    
    inline       sz_t     allocc( sz_t sz )         { return m_cpchnk.alloc(sz); };
    inline       sz_t     alloci( sz_t sz )         { return m_ipchnk.alloc(sz); };
    
    /* interface avec le solveur */
    void push( creal_t *& pmn, creal_t *& pmx, sz_t sz ) 
    {
      sz_t iptr;
      if ( nbp() == 0 ) { 
	iptr = allocc(2); 
	m_cpchnk.data(iptr)[0] = 0.0; 
	m_cpchnk.data(iptr)[1] = 1.0; 
      };
      iptr       = alloci(4);
      pMsz(iptr) = pmsz(iptr) = sz;
      pmad(iptr) = allocc(sz);
      pMad(iptr) = allocc(sz);
      pmn        = pmdata(iptr);
      pmx        = pMdata(iptr);
      m_nbp ++ ;
    };
    
    void add( creal_t const * pmn, creal_t const * pmx, sz_t sz )
    {
      creal_t *_pmn,*_pmx;
      push(_pmn,_pmx,sz);
      std::copy(pmn,pmn+sz,_pmn);
      std::copy(pmx,pmx+sz,_pmx);
    };
    
    inline void read( sz_t i, creal_t const *& pmn, creal_t const *& pmx, sz_t& sz )  const
    { pmn = pmdata(4*i), pmx = pMdata(4*i), sz = pmsz  (4*i); };
    
    void reset() { 
      m_nbp = 0; 
      m_cpchnk.free(); 
      m_ipchnk.free();
    };
    

    
    bool thickness( const creal_t& eps  ) { 
      if ( nbp() == 0 ) return true;
      creal_t mn = vct::min(pmdata(0),pmsz(0));
      creal_t mx = vct::max(pMdata(0),pMsz(0));
      if ( mn*mx > 0 ) return false;
      for ( sz_t i = 4; i < 4*nbp(); i += 4 )
	{
	  creal_t imn,imx;
	  imn = vct::min(pmdata(i),pmsz(i)); if ( imn < mn ) mn = imn;
	  imx = vct::max(pMdata(i),pMsz(i)); if ( imx > mx ) mx = imx;
	  if ( imx*imn > 0 ) return false;
	};
      return mx-mn < eps;
    };
    
    void copy( const vstack& vs  )
    {
      /* vs.print(); */
      reset();
      creal_t const *pmn;
      creal_t const *pmx;
      sz_t    sz,i;
      for ( i = 0; i < vs.nbp(); i ++ )
	{
	  vs.read(i,pmn,pmx,sz);
	  add(pmn,pmx,sz);
	};
    };
  };
};
//--------------------------------------------------------------------
__END_NAMESPACE_SYNAPS
/********************************************************************/
#endif //
