#include <synaps/arithm/gmp.h>
#include <synaps/mpol/xml.h>
#include <synaps/topology/TopSbdBdg2d.h>
#include <synaps/base/io/geomview.h>
#include <synaps/base/io/axel.h>

void usage (char* cmd)
{
  std::cout<<"usage:\n"<<cmd<<" file [-bZa] ..."<<std::endl;
  std::cout<<"\n  The file should contain the equation of an implicit curve, with (big) integer,"<<std::endl;
  std::cout<<"  in the format <polynomial>...</polynomial> (example: <polynomial>x^2+2*y^2-1</polynomial>)"<<std::endl;
  std::cout<<"options:\n";
  std::cout<<"   -b x0 x1 y0 y1   box where the topology is computed (default -10 10 -10 10)"<<std::endl;
  std::cout<<"   -a asr           aspect ratio"<<std::endl;
  std::cout<<"   -e eps           maximal size of box at the end of the subdivision"<<std::endl;
  std::cout<<"   -no_box          don't draw the singular boxes"<<std::endl;

}

int main(int argc, char **argv)
{
  typedef MPol<QQ>      mpol_t;
  typedef std::string   str_t;

  using std::cout;
  using std::endl;
  if(argc<2)
    {
      usage(argv[0]);
      return 0;
    }
  std::ifstream in(argv[1],std::ios::in);
  if(!in.good())
    {
      if(str_t(argv[1])!=str_t("-h"))
	{
	  std::cerr<<argv[0]<<": file "<<argv[1]<<" not found"<<std::endl;
	  return 1;
	}
      usage(argv[0]);
      return 0;
    }
  Seq<mpol_t> lp;
  mpol_t p;
  while(xml::read(in,p,Variables("x y")))
    {
      std::cout<<"polynomial: "<<p<<"\n  "<<degree(p)<<" ["<<degree(p,0)<<","<<degree(p,1)<<"] "<<p.size()<<std::endl;     
      lp.push_back(p);
    }

  double eps=1e-3,asr=0.1,a0=-3.1,b0=3,a1=-3.2,b1=3.1;
  int N=1;
  unsigned Z=1;
  bool axel=true;
  bool rel=false;
  bool dr_bx=true;
  for(unsigned c=2;c<argc;c++)
    {
      if(str_t(argv[c])==str_t("-N"))
	{
	  c++; N=atoi(argv[c]);
	  std::cout<<"N "<<N<<std::endl;
	}
      else if(str_t(argv[c])==str_t("-Z"))
	{
	  c++; Z=atoi(argv[c]);
	  std::cout<<"Z "<<Z<<std::endl;
	}
      else if(str_t(argv[c])==str_t("-a"))
	{
	  c++; asr=atof(argv[c]);
	  std::cout<<"A "<<asr<<std::endl;
	}
      else if(str_t(argv[c])==str_t("-e"))
	{
	  c++; eps=atof(argv[c]);
	  std::cout<<"E "<<eps<<std::endl;
	}
      else if(str_t(argv[c])==str_t("-gmv"))
	{
	  axel=false;
	  std::cout<<"Using geomview"<<std::endl;
	}
      else if(str_t(argv[c])==str_t("-rel"))
	{
	  rel=true;
	  std::cout<<"Using relative precision"<<std::endl;
	}
      else if(str_t(argv[c])==str_t("-no_box"))
	{
	  dr_bx=false;
	  std::cout<<"Don't draw singular boxes"<<std::endl;
	}
      else if(str_t(argv[c])==str_t("-b"))
	{
	  c++; if(c<argc) a0=atof(argv[c]);
	  c++; if(c<argc) b0=atof(argv[c]);
	  a1=a0; b1=b0;
	  c++; if(c<argc) a1=atof(argv[c]);
	  c++; if(c<argc) b1=atof(argv[c]);
	}
      else if(str_t(argv[c])==str_t("-h"))
	{
	  usage(argv[0]);
	  return 0;
	}
    }

  if(rel)
    {
     eps *= std::max(std::abs(b0-a0), std::abs(b1-a1));
    }

  std::cout<<"Eps: "<<eps<<std::endl;
  std::cout<<"Box: "<<a0<<" "<<b0<<" "<<a1<<" "<<b1<<std::endl;

  topology::point_graph<double> g;
  topology::assign(g, lp[0],  TopSbdBdg2d<double>(eps,asr,dr_bx), a0,b0,a1,b1); 
     
  if(axel)
    {
      axel::ostream vw("tmp.axl"); 
      vw << g;
      vw.view();
    }
  else
    {
      geomview::ostream vw("tmp.vect");
      vw <<g;
      vw.view();
    }
}

