/***************************************************************************
 *   Copyright (C) 2005 by Chen Liang                                      *
 *   cliang@cs.hku.hk                                                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef synaps_topology_TopSbdCurve3d_h
#define synaps_topology_TopSbdCurve3d_h

//--------------------------------------------------------------------
#include <vector>
#include <list>
#include <math.h>
#include <synaps/upol/bezier/BEZIER.m>
#include <synaps/mpol/MPol.h> 
#include <synaps/mpol/MPolDse.h> 
#include <synaps/msolve/sbdslv.h> 
#include <synaps/topology/Cell3dN.h>
#include <synaps/topology/Octree.h>
//--------------------------------------------------------------------
__BEGIN_NAMESPACE_SYNAPS
//--------------------------------------------------------------------
template<class C, 
	 class CELL= Cell3dN<5,C>, 
	 class MSLV= SbdSolver<typename CELL::coeff_t, SBDSLV_RDL> >
struct TopSbdCurve3d
{

  //type definitions
  typedef TopSbdCurve3d<C, CELL,MSLV>         self_t;

  typedef typename CELL::coeff_t         coeff_t;

  typedef typename CELL::MPOL_T          MPOL_T;
  typedef MPolDse<C, mpoldse::bernstein<C> > mpolb_t;

  typedef topology::box3d<coeff_t>       box_t;
  typedef topology::point<C>             point_t;
  typedef topology::point_graph<C>       pointgraph_t;

  
  typedef CELL                           cell_t;
  typedef OCTREE::Node<cell_t*>           node_t;
  typedef OCTREE::Octree<cell_t*>         octree_t;
  
  typedef intersection3d<C>       intersection_t;
  typedef shared_object<intersection_t*>  intersection_ptr;
  //  typedef intersection3d<coeff_t>         intersection_t;
  typedef Seq<intersection3d<coeff_t>*>   intersectionlist_t;
  //  typedef shared_object<intersection_t*>  intersection_ptr;
  typedef Seq<intersection_ptr>           intersection_ptr_list;
  typedef topology::point_graph<coeff_t,  intersection_ptr_list> pointgraph_int_t;

protected:
// data members

    C           m_dEpsilon; // maximum tolerance
    int         m_nMinSbd;
    octree_t    m_cOcTree;  // the octree structure

    int         m_nCount; // traveral count

public:
// constructors/distructor

    TopSbdCurve3d():m_dEpsilon(1e-3), m_nMinSbd(0),m_nCount(0) {}
    TopSbdCurve3d(coeff_t e):m_dEpsilon(e), m_nMinSbd(0),m_nCount(0) {}
    TopSbdCurve3d(coeff_t e, int d):m_dEpsilon(e), m_nMinSbd(d),m_nCount(0) {}

    ~TopSbdCurve3d();

    //int regularity(cell_t & c);

    // compute intersection on the dedicated face for the given polynomial
    void     computeinter_face(intersectionlist_t& inters, cell_t& c, int nAxis, int nFace);
 
    void     computeintersections(intersectionlist_t& inters, cell_t& c);

    template<class POINTGRAPH>
    void run(POINTGRAPH& g, CELL* rc);

protected:
// helpers

    // this function subdivide a given node
    void subdivide(node_t* n, std::vector<node_t*>& aNodeList);
    // this function subdivide a cell
    void subdivide(CELL* cell, std::vector<CELL*>& subCells);

    // housekeeping: octree
    void deleteNode(node_t* n, std::vector<intersection_t*>& inters);
};

//////////////////////////////////
// template instantiation       //
//////////////////////////////////

//inline template class TopSbdCurve3d< double, double, bezier::bzslv<double> >;

//////////////////////////////////
// constructor/distructor       //
//////////////////////////////////
template< class X, class CELL, class MSLV>
TopSbdCurve3d<X, CELL, MSLV>::~TopSbdCurve3d()
{
  // release the memory occupied by the octree
  m_nCount++;
  std::vector<intersection_t*> inters;
  deleteNode(m_cOcTree.GetRoot(), inters);
  for (int i = 0; i < inters.size(); i++)
    delete inters[i];
}

//--------------------------------------------------------------------
/////////////////////////////////////////
/***************************************/
/*               Z^                    */
/*                |_____               */
/*               /|    /|              */
/*     N  ^     /_|_B_/ |              */
/*        |    |  |___|_|___> Y        */
/*        |    | /F   | /              */
/*             |/_____|/               */
/*            |/_                      */
/*           X                         */
/*                                     */
/*    Fig.1  The coordinate system     */
/***************************************/
/////////////////////////////////////////

template< class C, class CELL, class MSLV>
void TopSbdCurve3d<C, CELL, MSLV>::deleteNode(node_t* n, std::vector<intersection_t*>& inters)
{
    if (n->GetF_NEchild())
        deleteNode(n->F_NEchild, inters);
    if (n->GetF_NWchild())
        deleteNode(n->F_NWchild, inters);
    if (n->GetF_SEchild())
        deleteNode(n->F_SEchild, inters);
    if (n->GetF_SWchild())
        deleteNode(n->F_SWchild, inters);
    if (n->GetB_NEchild())
        deleteNode(n->B_NEchild, inters);
    if (n->GetB_NWchild())
        deleteNode(n->B_NWchild, inters);
    if (n->GetB_SEchild())
        deleteNode(n->B_SEchild, inters);
    if (n->GetB_SWchild())
        deleteNode(n->B_SWchild, inters);

    for (int i = 0; i < n->GetCell()->m_aInterList.size(); i++)
    {
        if (n->GetCell()->m_aInterList[i]->nCount != m_nCount)
        {
            inters.push_back(n->GetCell()->m_aInterList[i]);
            n->GetCell()->m_aInterList[i]->nCount=m_nCount;
        }
    }

    delete n->GetCell();
    delete n;
}
//--------------------------------------------------------------------
template<class C, class CELL,  class MSLV>
void TopSbdCurve3d<C, CELL, MSLV>::subdivide(node_t* n, std::vector<node_t*>& aNodeList)
{
  using namespace OCTREE;
  cell_t* c = n->GetCell();
  std::vector<cell_t*> aSubCells;
  
  // subdivide a cell into 8 subcells
  subdivide(c,aSubCells);
  
  // encapsulate the cells with the node of the octree
  aNodeList.push_back(new node_t(B_SW, n, aSubCells[0]));
  aNodeList.push_back(new node_t(F_SW, n, aSubCells[1]));
  aNodeList.push_back(new node_t(B_SE, n, aSubCells[2]));
  aNodeList.push_back(new node_t(F_SE, n, aSubCells[3]));
  aNodeList.push_back(new node_t(B_NW, n, aSubCells[4]));
  aNodeList.push_back(new node_t(F_NW, n, aSubCells[5]));
  aNodeList.push_back(new node_t(B_NE, n, aSubCells[6]));
  aNodeList.push_back(new node_t(F_NE, n, aSubCells[7]));
}
//--------------------------------------------------------------------

/////////////////////////////////////////
/***************************************/
/*          ^ Z                        */
/*          |                          */
/*          |                          */
/*          |                          */
/*          |4----6                    */
/*         /||___/|_______> Y          */
/*        5/-0--7 2                    */
/*        /|/   |/                     */
/*       / 1----3                      */
/*     \/_                             */
/*     X                               */
/*                                     */
/*    Fig.2  The layout of the         */
/*        subdivided cells             */
/***************************************/
/////////////////////////////////////////

/////////////////////////////////////////
/***************************************/
/*          ^ Z                        */
/*          |   5     0                */
/*          |  |     /                 */
/*          |  v   |/_                 */
/*          |------                    */
/*         /||___/|_______> Y          */
/*  2-->  -/ ---- | <--3               */
/*        /|/ 1 |/                     */
/*       / ------                      */
/*     \/_    ^                        */
/*     X      |                        */
/*             4                       */
/*    Fig.3  Faces index of a cell     */
/***************************************/
/////////////////////////////////////////
// this function spawns 8 evenly subdivided child cell of this call
template < class C, class CELL, class MSLV >
void TopSbdCurve3d<C, CELL, MSLV>::subdivide(CELL* cell,
					    std::vector<CELL*>& subCells)
{
  typedef typename CELL::intersection_t intersection_t;
  typedef typename CELL::intersectionlist_t intersectionlist_t;
  typedef typename CELL::MPOL_T MPOL_T;
    // define the subdivision points
#ifdef SYNAPS_SBD3D_RNDFACTORWHENSUBDIVISION
    // a perturbation ranged from 0.005-0.1
    C t[] = {(C(rand())/C(RAND_MAX)-C(0.5))*C(0.095)+C(0.5),
             (C(rand())/C(RAND_MAX)-C(0.5))*C(0.095)+C(0.5),
             (C(rand())/C(RAND_MAX)-C(0.5))*C(0.095)+C(0.5)};
    t[0]+= (t[0]>0.5?0.005:-0.005);
    t[1]+= (t[1]>0.5?0.005:-0.005);
    t[2]+= (t[2]>0.5?0.005:-0.005);
    C mid[] = {cell->m_cBox.x0*(C(1)-t[0])+cell->m_cBox.x1*t[0], cell->m_cBox.y0*(C(1)-t[1])+cell->m_cBox.y1*t[1], cell->m_cBox.z0*(C(1)-t[2])+cell->m_cBox.z1*t[2]};
#else
    C mid[] = {(cell->m_cBox.x0+cell->m_cBox.x1)/C(2), (cell->m_cBox.y0+cell->m_cBox.y1)/C(2), (cell->m_cBox.z0+cell->m_cBox.z1)/C(2)};
#endif
    C box[] = {cell->m_cBox.x0, cell->m_cBox.x1, cell->m_cBox.y0, cell->m_cBox.y1, cell->m_cBox.z0, cell->m_cBox.z1};

    // create new cells
    subCells.push_back(new CELL(box[0], mid[0], box[2], mid[1], box[4], mid[2]));
    subCells.push_back(new CELL(mid[0], box[1], box[2], mid[1], box[4], mid[2]));
    subCells.push_back(new CELL(box[0], mid[0], mid[1], box[3], box[4], mid[2]));
    subCells.push_back(new CELL(mid[0], box[1], mid[1], box[3], box[4], mid[2]));
    subCells.push_back(new CELL(box[0], mid[0], box[2], mid[1], mid[2], box[5]));
    subCells.push_back(new CELL(mid[0], box[1], box[2], mid[1], mid[2], box[5]));
    subCells.push_back(new CELL(box[0], mid[0], mid[1], box[3], mid[2], box[5]));
    subCells.push_back(new CELL(mid[0], box[1], mid[1], box[3], mid[2], box[5]));

    // update the depth count for each subdivided cells
    for (int i = 0; i < 8; i++)
    {
        subCells[i]->m_nDepth = cell->m_nDepth+1;
    }

    // the subcells inherits the previous computed intersections on its parent cell
    for (int i = 0; i < cell->m_aInterList.size(); i++)
    {
        intersection_t* inter = cell->m_aInterList[i];

        // assign the intersections on the parent cell to subdivided cells, according to the plane
        // they are on and the planar coordinates
        int nidx = inter->nPlane; // testing the plane YZ (nidx=0) or XZ (nidx=1) or XZ (nidx=2)
        int *cidx=subcellidx+(nidx<<3); // the right set of subcell indices to be assigned each intersection
        int axisidx0 = plane_axisidx[nidx<<1], axisidx1 = plane_axisidx[(nidx<<1)+1]; // the first and the second axis indices, {1, 2} for YZ plane, {0, 2} for XZ plane and {0, 1} for XY plane

        assert(_is_verysmall(inter->pnt[nidx]-box[nidx<<1]) || _is_verysmall(box[(nidx<<1)+1]-inter->pnt[nidx]));
        assert(inter->pnt[axisidx0] >= box[axisidx0<<1] && inter->pnt[axisidx0] <= box[(axisidx0<<1)+1] &&
               inter->pnt[axisidx1] >= box[axisidx1<<1] && inter->pnt[axisidx1] <= box[(axisidx1<<1)+1]);

        if (inter->pnt[axisidx0] < mid[axisidx0]) // if smaller than the first mid-point
        {
            if (inter->pnt[axisidx1] < mid[axisidx1]) // if smaller than the second mid-point
            {
                if (inter->pnt[nidx]-box[nidx<<1] < box[(nidx<<1)+1]-inter->pnt[nidx]) // if the plane corresponds to x (or y, or z) = a, where x (or y, or z) is in [a, b]
                    subCells[cidx[0]]->m_aInterList.push_back(inter);
                else // if the plane corresponds to x (or y, or z) = b, where x (or y, or z) is in [a, b]
                    subCells[cidx[1]]->m_aInterList.push_back(inter);
            }
            else // if larger than the first mid-point
            {
                if (inter->pnt[nidx]-box[nidx<<1] < box[(nidx<<1)+1]-inter->pnt[nidx]) // if the plane corresponds to x (or y, or z) = a, where x (or y, or z) is in [a, b]
                    subCells[cidx[2]]->m_aInterList.push_back(inter);
                else
                    subCells[cidx[3]]->m_aInterList.push_back(inter);
            }
        }
        else // if lager than the first mid-point
        {
            if (inter->pnt[axisidx1] < mid[axisidx1]) // if smaller than the second  mid-point
            {
                if (inter->pnt[nidx]-box[nidx<<1] < box[(nidx<<1)+1]-inter->pnt[nidx]) // if the plane corresponds to x (or y, or z) = a, where x (or y, or z) is in [a, b]
                    subCells[cidx[4]]->m_aInterList.push_back(inter);
                else // if the plane corresponds to x (or y, or z) = b, where x (or y, or z) is in [a, b]
                    subCells[cidx[5]]->m_aInterList.push_back(inter);
            }
            else // if larger than the second mid-point
            {
                if (inter->pnt[nidx]-box[nidx<<1] < box[(nidx<<1)+1]-inter->pnt[nidx])
                    subCells[cidx[6]]->m_aInterList.push_back(inter);
                else // if the plane corresponds to x (or y, or z) = b, where x (or y, or z) is in [a, b]
                    subCells[cidx[7]]->m_aInterList.push_back(inter);
            }
        }
    }

    // update the 'computed faces' flag for all subcells
    for (int i = 0; i < 8*MAX_FACEIDX; i++)
    {
        int nCellIdx=_cellidx[i], nInheritIdx = _inheritidx[i];
        // assert that a face never inherit from more than one parent cell's face
        assert((subCells[nCellIdx]->m_aFaceComputed[nInheritIdx]) == false);
        /*if ((subCells[nCellIdx]->m_aFaceComputed[nInheritIdx]))
            std::cout << "assertion failed at depth " << cell->m_nDepth << std::endl;*/
        if (cell->m_aFaceComputed[nInheritIdx]) // if the parent cell's face this cell inherited from has 'computed' flag
            subCells[nCellIdx]->m_aFaceComputed[nInheritIdx] = true;
    }


    // split the representations for each surface
    for (int i = 0; i < 5; i++)
    {
        MPOL_T splited[8];
        int vr_idx[3] = {VARIDX(cell->m_F[i], 0), VARIDX(cell->m_F[i], 1), VARIDX(cell->m_F[i], 2)};

        subCells[7]->m_F[i] = cell->m_F[i];
#ifdef SYNAPS_SBD3D_RNDFACTORWHENSUBDIVISION
        // subdivide along z direction with a perturbation
        mpoldse::split(subCells[3]->m_F[i].rep(), subCells[7]->m_F[i].rep(), vr_idx[2], t[2]);

        // subdivide along y direction with a perturbation
        mpoldse::split(subCells[1]->m_F[i].rep(), subCells[3]->m_F[i].rep(), vr_idx[1], t[1]);
        mpoldse::split(subCells[5]->m_F[i].rep(), subCells[7]->m_F[i].rep(), vr_idx[1], t[1]);

        // subdivide along x direction with a perturbation
        mpoldse::split(subCells[0]->m_F[i].rep(), subCells[1]->m_F[i].rep(), vr_idx[0], t[0]);
        mpoldse::split(subCells[2]->m_F[i].rep(), subCells[3]->m_F[i].rep(), vr_idx[0], t[0]);
        mpoldse::split(subCells[4]->m_F[i].rep(), subCells[5]->m_F[i].rep(), vr_idx[0], t[0]);
        mpoldse::split(subCells[6]->m_F[i].rep(), subCells[7]->m_F[i].rep(), vr_idx[0], t[0]);
#else
        // subdivide along z direction
        mpoldse::split(subCells[3]->m_F[i].rep(), subCells[7]->m_F[i].rep(), vr_idx[2]);

        // subdivide along y direction
        mpoldse::split(subCells[1]->m_F[i].rep(), subCells[3]->m_F[i].rep(), vr_idx[1]);
        mpoldse::split(subCells[5]->m_F[i].rep(), subCells[7]->m_F[i].rep(), vr_idx[1]);

        // subdivide along x direction
        mpoldse::split(subCells[0]->m_F[i].rep(), subCells[1]->m_F[i].rep(), vr_idx[0]);
        mpoldse::split(subCells[2]->m_F[i].rep(), subCells[3]->m_F[i].rep(), vr_idx[0]);
        mpoldse::split(subCells[4]->m_F[i].rep(), subCells[5]->m_F[i].rep(), vr_idx[0]);
        mpoldse::split(subCells[6]->m_F[i].rep(), subCells[7]->m_F[i].rep(), vr_idx[0]);
#endif
    }

    // compute the intersections on the cell borders (Dynamic Programming applied here
    // to avoid repetitive computation, see arrangements of _cellidx/_faceidx/_shareidx)
    for (int i = 0; i < 8*MAX_FACEIDX; i++)
    {
        intersectionlist_t aInters;
        int nCellIdx=_cellidx[i], nFaceIdx = _faceidx[i], nShareIdx = _shareidx[i];

        if (nFaceIdx >= 0)
        {
            computeinter_face(aInters, *subCells[nCellIdx], nFaceIdx>>1, nFaceIdx%2);
            // the cell shares a face with this cell (i.e. subcells[_cellidx[i]]) should also
            // share the some intersections (on a particular face) with this cell
            cell->mergelist(subCells[nCellIdx]->m_aInterList, aInters);
            cell->mergelist(subCells[nShareIdx]->m_aInterList, aInters);
            // mark the this face (i.e. face nFaceIdx of cell nCellIdx) as 'computed'
            subCells[nCellIdx]->m_aFaceComputed[nFaceIdx] = true;
            // for two subcells that shares a face, the shared face should of coz be parallel,
            // so their face indices should have same rounded value when devided by 3 (since
            // the face indices are (YZ_Plane_0, YZ_Plane_1, XZ_Plane0, XZ_Plane1, XY_Plane0,
            // XY_Plane1)). And of coz, they have different mod against 2.
            // by a simple deduction, f(x)=x+(x%2==1?-1:1) gives the correct face index for face x;
            // now we mark the shared face also as 'computed'
            assert(subCells[nShareIdx]->m_aFaceComputed[nFaceIdx+(nFaceIdx%2?-1:1)] == false);
            //if (subCells[nShareIdx]->m_aFaceComputed[nFaceIdx+(nFaceIdx%2?-1:1)])
            //    std::cout << "assertion failed at depth " << cell->m_nDepth << " nCell " << nCellIdx << " nFace " << nFaceIdx << " nShare " << nShareIdx << " i"  << i << ", cell " << nShareIdx << " face " << (nFaceIdx+(nFaceIdx%2?-1:1)) << " has already been computed." << std::endl;
            subCells[nShareIdx]->m_aFaceComputed[nFaceIdx+(nFaceIdx%2?-1:1)] = true;
        }
    }
}
//====================================================================
template <class C, class CELL, class MSLV >
void TopSbdCurve3d<C, CELL, MSLV>::computeintersections(intersectionlist_t& inters, CELL& c)
{
  // for each facet of the cell, compute its intersections with the curve
  for (int v = 0; v < 3; v++) // for each axis
    {
      for (int f = 0; f < 2; f++) // for each facet perpendicular to that axis
        {
	  // compute the intersections on this facet if it has not yet been computed
	  if (!c.m_aFaceComputed[v*2+f])
            {
	      computeinter_face(inters, c, v, f);
	      c.m_aFaceComputed[v*2+f] = true;
            }
        }
    }
}
//====================================================================
template < class C, class CELL, class MSLV >
void TopSbdCurve3d<C, CELL, MSLV>::computeinter_face(intersectionlist_t& inters, 
						    CELL& c, 
						    int nAxis, int nFace)
{
    point_t l(3u);

    // obtain the polynomail representing the face
    MPOL_T facepol[2]; // = {m_F[0], m_F[1]};
    face(facepol[0].rep(), c.m_F[0].rep(), nAxis, nFace);
    face(facepol[1].rep(), c.m_F[1].rep(), nAxis, nFace);

    // build a system of equations
    std::list<MPOL_T> eqnList;
    eqnList.push_back(facepol[0]);
    eqnList.push_back(facepol[1]);

    // solve the equation using a 2-variate solver
    MSLV solver(facepol, facepol+2, 1e-4);
    std::vector<C> results;
    int nSol, nVar;

    solver.solve(results, nSol, nVar, eqnList.begin(), eqnList.end());
    assert(nVar==2);

    // rescale the solution to the coordinates of the cell
    // also update the multiplicity accordingly
    for (int s = 0; s < nSol; s++)
    {
        C x[3] = {(C)nFace,(C)nFace,(C)nFace}; // the 3D coord of the point

        switch(nAxis)
        {
        case PLANEYZ: // yz-planes
            x[1] = (C)(results[nVar*s]);
            x[2] = (C)(results[nVar*s+1]);
            break;
        case PLANEXZ: // xz-planes
            x[0] = (C)(results[nVar*s]);
            x[2] = (C)(results[nVar*s+1]);
            break;
        case PLANEXY: // xy-planes
            x[0] = (C)(results[nVar*s]);
            x[1] = (C)(results[nVar*s+1]);
            break;
        }

        // determine the multiplicity
        //int nMaxDeg = c.m_T[nAxis].rep().env.sz(VARIDX(c.m_T[nAxis], nAxis)); 
	// the degree of nAxis_th variable of nAxis_th component of t = {t1, t2, t3} = gradF cross gradG
        int nMulti = 1;

        // currently, only account for multiplicity 1 or 2...
        //for (int v = 0; v< nMaxDeg; v++)
        //{
            C val[3] = {-1,-1,-1};
            eval(val[0], c.m_T[0], x); // evaluate the t_nAxis(x0, y0, z0) at the intersection (x0, y0, z0)
            eval(val[1], c.m_T[1], x); // evaluate the t_nAxis(x0, y0, z0) at the intersection (x0, y0, z0)
            eval(val[2], c.m_T[2], x); // evaluate the t_nAxis(x0, y0, z0) at the intersection (x0, y0, z0)

            /*if (abs(val[nAxis]) < 0.1)
            {
                std::cout << "sin val: " << val[nAxis];
                val[nAxis] /= sqrt(val[0]*val[0]+val[1]*val[1]+val[2]*val[2]); // normalize the vector
                std::cout << " nor: " << val[nAxis] << std::endl;
            }
            else
            {
                val[nAxis] /= sqrt(val[0]*val[0]+val[1]*val[1]+val[2]*val[2]); // normalize the vector
            }*/

            val[nAxis] /= sqrt(val[0]*val[0]+val[1]*val[1]+val[2]*val[2]); // normalize the vector

            if (_is_verysmall(std::abs(val[nAxis])))
            {
                nMulti = 2;
                //std::cout << "multiple point: val =" << val << ",  f(" << nAxis << ")=" << c.m_T[nAxis] << std::endl;
            }
        //}


        // scale to the cell's coordinate
        l[0] = c.m_cBox.x0*((C)1-x[0]) + c.m_cBox.x1*x[0];
        l[1] = c.m_cBox.y0*((C)1-x[1]) + c.m_cBox.y1*x[1];
        l[2] = c.m_cBox.z0*((C)1-x[2]) + c.m_cBox.z1*x[2];

        intersection_t *ni = new intersection_t(l, nAxis, nMulti);
        inters.push_back(ni);
        ni->nDepth = c.m_nDepth;
    }
}

//====================================================================
template< class C, class CELL, class MSLV>
template< class POINTGRAPH>
void TopSbdCurve3d<C, CELL, MSLV>::run(POINTGRAPH& g, CELL* rc)
{
    // initialization
    node_t* root = m_cOcTree.GetRoot();

    // step0:
   
    computeintersections(rc->m_aInterList,*rc);
    // std::cout << "no. of roots at root: " << rc->m_aInterList.size() << std::endl;
    root->SetCell(rc);

    // step3:
    // subdivision
    std::list<node_t*> aNodeQueue;
    std::vector<node_t*> aSingular;
    pointgraph_int_t g_tmp;
    int nCellProc = 0;

    aNodeQueue.push_front(root);

    while(aNodeQueue.size())
    {
        // process the first node in the queue
        node_t* n = *aNodeQueue.begin();
        aNodeQueue.pop_front();

        bool bSubdivide = false;
        bool bBuildGraph = false;

        // if the cell is regular, i.e. the topology of the surface within the cell is certified
        if (regularity(*n->GetCell()))
        {
	  if (is_variety_in(*n->GetCell())) // it may contains the polar curve
            {
                if (n->GetCell()->m_nDepth < m_nMinSbd) // if we did less than the minimum number of subdivision
                {
                    // subdivide further
                    bSubdivide = true;
                }
                else // build the graph
                {
                    bBuildGraph = true;
                }
            }
            else // if regular yet does not contain the surface, of no interest for further subdivision
            {
                // its safe to say that not any facet of this cell should contain a intersection point
                memset(n->GetCell()->m_aFaceComputed, true, sizeof(bool)*6);
            }
        }
        else // if the cell is not regular
        {
            if (n->GetCell()->testsize(m_dEpsilon)) // if the cell size is smaller than epsilon
            {
                // keep a count of singular cells
	      if (is_variety_in(*n->GetCell()))
                {
                    bBuildGraph = true;
                    aSingular.push_back(n);
                }
            }
            else // if the cell size is still larger than the set threshold
            {
                // further subdivision if it may contains the intersection
	      if (is_variety_in(*n->GetCell()))
                    bSubdivide = true;
            }
        }

        // in fact, bSubdivide and bBuildGraph cannot be both true (but can be both false)
        // assert(!(bSubdivide && bBuildGraph))
        if (bSubdivide)
        {
            // append the subdivided node at the back of the queue
            // in doing this, we can guarantee that larger nodes are processed first
            std::vector<node_t*> aNodeList;
            subdivide(n, aNodeList);
            for (int i = 0; i < (int)aNodeList.size(); i++)
                aNodeQueue.push_front(aNodeList[i]);
        }
        else if (bBuildGraph)
        {
	  build_graph(g_tmp,*n->GetCell());
        }

        // a counter
        nCellProc++;
    }

    // convert g_tmp to g
    // calcuate the size
    int nVert = (int)rep(rep(g_tmp).Vertices).size();
    int nEdges = 0;

    for(int i=0; i<(int)rep(rep(g_tmp).Edges).size(); ++i)
        nEdges += (int)rep(rep(g_tmp).Edges)[i].size();

    if (nVert)
    {
        int nVertIdx = 0;
        int nBase = (int)rep(rep(g).Vertices).size();; // base index of g when g is not empty
        m_nCount++;
        for(int i=0; i<nVert; ++i)
        {
            if (rep(rep(rep(g_tmp).Vertices)[i])->nCount != m_nCount)
            {
                rep(rep(rep(g_tmp).Vertices)[i])->nCount = m_nCount;
                rep(rep(rep(g_tmp).Vertices)[i])->nFlag = nVertIdx++;
                g.push_back(rep(rep(rep(g_tmp).Vertices)[i])->pnt);
            }
        }

        for(int i=0; i<(int)rep(rep(g_tmp).Edges).size(); ++i)
        {
            for(int j=0; j<(int)rep(rep(g_tmp).Edges)[i].size(); ++j)
            {
                g.insert(rep(rep(rep(g_tmp).Vertices)[i])->nFlag+nBase,
			 rep(rep(rep(g_tmp).Vertices)[rep(rep(g_tmp).Edges)[i][j]])->nFlag+nBase);
            }
        }

        //std::cout << "No. of Vertices: " << (int)rep(rep(g).Vertices).size() - nBase << std::endl;
        //std::cout << "No. of Edges: " << nEdges << std::endl;
    }

    //std::cout << "Total Cell Processed: " << nCellProc << std::endl;
    //std::cout << "Sing Cells: " << aSingular.size() << std::endl;
}


//--------------------------------------------------------------------
namespace topology {

  template<class C, class U, class CELL, class SLV>
  void assign(point_graph<C>& g, 
	      const MPol<U>& p1, 
	      const MPol<U>& p2, 
	      TopSbdCurve3d<C,CELL,SLV> slv,
	      double x0, double x1,
	      double y0, double y1,
	      double z0, double z1)
  {
    typedef TopSbdCurve3d<C,CELL,SLV> MTH;
    typedef typename MTH::cell_t  cell_t;
    typedef typename MTH::mpolb_t mpolb_t;
    
    // step1:
    // computing grad F1 cross grad F2
    mpolb_t p1dse, p2dse;

    toBernstein(p1, p1dse, x0,x1,y0,y1,z0,z1);
    toBernstein(p2, p2dse, x0,x1,y0,y1,z0,z1);

    // debug only
//     std::cout << "F("
//               << degree(p1,0) << ", "
//               << degree(p1,1) << ", "
//               << degree(p1,2) << "): "
//               << p1 << std::endl;

//     std::cout << "G("
//               << degree(p2,0) << ", "
//               << degree(p2,1) << ", "
//               << degree(p2,2) << "): "
//               << p2 << std::endl;

//     std::cout << "Fb("
//               << p1dse.rep().env.sz(VARIDX(p1dse, 0)) << ", "
//               << p1dse.rep().env.sz(VARIDX(pd1dse, 1)) << ", "
//               << p1dse.rep().env.sz(VARIDX(pd1dse, 2)) << "): "
//               << p1dse << std::endl;

//     std::cout << "Gb("
//               << p2dse.rep().env.sz(VARIDX(pd2dse, 0)) << ", "
//               << p2dse.rep().env.sz(VARIDX(pd2dse, 1)) << ", "
//               << p2dse.rep().env.sz(VARIDX(pd2dse, 2)) << "): "
//               << p2dse << std::endl;

    mpolb_t p1_dx, p1_dy, p1_dz;
    mpolb_t p2_dx, p2_dy, p2_dz;

    diff(p1_dx, p1dse, VARIDX(pd1dse, 0));
    diff(p1_dy, p1dse, VARIDX(pd1dse, 1));
    diff(p1_dz, p1dse, VARIDX(pd1dse, 2));

    diff(p2_dx, p2dse, VARIDX(pd1dse, 0));
    diff(p2_dy, p2dse, VARIDX(pd2dse, 1));
    diff(p2_dz, p2dse, VARIDX(pd2dse, 2));

//     std::cout << "grad Fb = {" << p1_dx << ", " <<  std::endl
//               << "           " << p1_dy << ", " <<  std::endl
//               << "           " << p1_dz << "} " <<  std::endl;

//     std::cout << "grad Gb = {" << p2_dx << ", " <<  std::endl
//               << "           " << p2_dy << ", " <<  std::endl
//              << "           " << p2_dz << "} " <<  std::endl;

    mpolb_t c1, c2, c3; // grad F1 cross grad F2
    using arithm::operator*;
    c1 = p1_dy * p2_dz - p1_dz * p2_dy;
    c2 = p2_dx * p1_dz - p2_dz * p1_dx;
    c3 = p1_dx * p2_dy - p1_dy * p2_dx;

    // debug only
//     std::cout << "grad(F) X grad(G) = {" << std::endl
//               << "                     " << c1 << "," << std::endl
//               << "                     " << c2 << "," << std::endl
//               << "                     " << c3 << "}" << std::endl;

    // step2:
    // assign to the octree
    cell_t* rc = new cell_t(x0,x1,y0,y1,z0,z1);
    rc->m_F[0] = p1dse;
    rc->m_F[1] = p2dse;
    rc->m_T[0] = c1;
    rc->m_T[1] = c2;
    rc->m_T[2] = c3;

    slv.run(g,rc);
  }
}

//====================================================================
template < class C >
bool is_variety_in(Cell3dN<5,C>& c)
{
  return c.containsVariety(0) && c.containsVariety(1);
}

//--------------------------------------------------------------------
template < class POINTGRAPH, class CELL>
void build_graph(POINTGRAPH& g, CELL& c)
{

  typedef typename CELL::coeff_t         coeff_t;
  typedef typename CELL::MPOL_T          MPOL_T;
  typedef intersection3d<coeff_t>        intersection_t;
  typedef topology::point<coeff_t>       point_t;
  typedef shared_object<intersection_t*> intersection_ptr;
  // determine the regularity if it is not known
  if (c.m_nRegularIdx == -2)
    regularity(c);
  
  if (c.m_nRegularIdx == -1) // a singular cell
    {
      if (c.m_aInterList.size()) // if the cell has other point connecting to this cell
        {
	buildsingular:
	  ;//	  build_box(g,c.m_cBox);
        }
    }
    else // a regular cell
    {
        Seq<intersection_t*> inters = c.m_aInterList;

        // compute the intersections of the curve and the cell facets
        // computeintersections(inters); // no longer necessary

        // sort according to the regular direction
        switch(c.m_nRegularIdx)
        {
        case 0:
            sort(inters.begin(), inters.end(), XMAJOR());
            break;
        case 1:
            sort(inters.begin(), inters.end(), YMAJOR());
            break;
        case 2:
            sort(inters.begin(), inters.end(), ZMAJOR());
            break;
        }

        // construct the graph
        bool bConnectWithLast = false;
        int nMulti = 1;
        int nLastIdx = -1;
        for (int v = 0; v < (int)inters.size(); v++)
        {
            int nBIdx = g.push_back(intersection_ptr(inters[v]));

            if (bConnectWithLast)
                g.insert(nLastIdx, nBIdx);

            if (!bConnectWithLast && nMulti < inters[v]->nMulti)
                bConnectWithLast = false;
            else
                bConnectWithLast = !bConnectWithLast;

            if (nMulti < inters[v]->nMulti)
            {
                nMulti++;
                v--;
            }
            else
            {
                nMulti = 1;
            }

            nLastIdx = nBIdx;
        }

        if (inters.size() > 2)
        {
	  //            std::cout << "a cell with " << inters.size() << " intersection points." << std::endl;
//             for (int i = 0; i < inters.size(); i++)
//             {
//                 std::cout << "(" << inters[i]->pnt[0] << ", " << inters[i]->pnt[1]
//                           << ", " << inters[i]->pnt[2] << ") d=" << inters[i]->nDepth << std::endl;
//             }
            goto buildsingular;
        }
    }
}
//--------------------------------------------------------------------
template < class C >
bool regularity(Cell3dN<5,C>& c)
{
    // check if the curve tangent, i.e. t = (t0, t1, t2) = grad F cross grad G
    // has no sign change for at least one of its component (or say, along one axis)
    for (int i = 0; i < 3; i++)
    {
        C* coeff = c.m_T[i].begin();
        C lastC = coeff[0];
        bool bSignChange = false;

        for (int j = 1; j < c.m_T[i].rep().env.sz(); j++)
        {
            C cirC = coeff[j];
            if (_is_singular(cirC)) // if cirC is a zero value
            {
                continue;
            }
            else
            {
                if (lastC*cirC < 0) // a sign change is detected
                {
                    bSignChange = true;
                    lastC = cirC;
                    break;
                }

                lastC = cirC; // always remember the last non-singular coefficient
            }
        }

        if (_is_singular(lastC))  //all coefficients are singular (or 0)
            continue;
        else if (!bSignChange) // no sign changes
        {
            c.m_nRegularIdx = i;
            return TRUE;
        }
    }
    c.m_nRegularIdx = -1;
    return FALSE;
}
//--------------------------------------------------------------------
template<typename C, typename MPOL, typename MPOLB>
void toBernstein(const MPOL& p, MPOLB& bp, 
		 const C& a0, const C& b0, 
		 const C& a1, const C& b1, 
		 const C& a2, const C& b2)
{
    MPOLB tmp(p);
    //    PRINT_DEBUG(p);
    mpoldse::restrict(tmp.rep(),VARIDX(tmp,0),a0,b0); // RESCALE(p,v,a,b) ==> p = p((1-t)*a+t*b)
    mpoldse::restrict(tmp.rep(),VARIDX(tmp,1),a1,b1); // RESCALE(p,v,a,b) ==> p = p((1-t)*a+t*b)
    mpoldse::restrict(tmp.rep(),VARIDX(tmp,2),a2,b2); // RESCALE(p,v,a,b) ==> p = p((1-t)*a+t*b)

    bp = tmp;
}
//--------------------------------------------------------------------
template < class POINTGRAPH, class CELL>
void build_box(POINTGRAPH& g, CELL* c)
{
  typedef typename CELL::coeff_t         coeff_t;
  typedef topology::point<coeff_t>       point_t;
  int nBIdx;
  point_t l(3u);
  
  // form the graph representing the shape of the cell
  for (int i = 0; i < 8; i++)
    {
      l[0] = ((i&0x1) ? c->m_cBox.x1:c->m_cBox.x0);
      l[1] = ((i&0x2) ? c->m_cBox.y1:c->m_cBox.y0);
      l[2] = ((i&0x4) ? c->m_cBox.z1:c->m_cBox.z0);
      //nBIdx = g.push_back(shared_object(l));
      //nBIdx = g.push_back(intersection_ptr(new intersection_t(l,0,1)));
      nBIdx = g.push_back(l);
    }
  
  nBIdx -= 7;
  
  g.insert(nBIdx,nBIdx+1);
  g.insert(nBIdx,nBIdx+2);
  g.insert(nBIdx+1,nBIdx+3);
  g.insert(nBIdx+2,nBIdx+3);
  
  g.insert(nBIdx+4,nBIdx+5);
  g.insert(nBIdx+4,nBIdx+6);
  g.insert(nBIdx+5,nBIdx+7);
  g.insert(nBIdx+6,nBIdx+7);
  
  g.insert(nBIdx,nBIdx+4);
  g.insert(nBIdx+1,nBIdx+5);
  g.insert(nBIdx+2,nBIdx+6);
  g.insert(nBIdx+3,nBIdx+7);
}
//--------------------------------------------------------------------
__END_NAMESPACE_SYNAPS
/********************************************************************/
#endif // synaps_topology_TopSbdCurve3d_h

