/*********************************************************************
*      This file is part of the source code of AXEL library          *
*   Author(s): E. Tobis                                              *
*              B. Mourrain, GALAAD, INRIA                            *
**********************************************************************/
/** \FILE                                                            */
#ifndef axel_mesh_hermite_H__
#define axel_mesh_hermite_H__
/********************************************************************/

#include <synaps/init.h>

#include <synaps/usolve/bezier/SlvBzBdg.h>
#include <synaps/msolve/sbdslv.h>
#include <synaps/linalg/Solve.h>
#include <synaps/linalg/array3d.h> 
#include <synaps/linalg/lapack.h>
#include <synaps/topology/mesh/bezier3d.h>
#include <synaps/topology/mesh/regularity.h>
#include <synaps/topology/mesh/miscellaneous.h>
#include <synaps/topology/cell2d.h>
#include <synaps/topology/sbd2d.h>
#include <synaps/topology/mesh/rootsOnEdges.h>
#include <synaps/topology/mesh/BEZIER.h>
#include <synaps/topology/point_graph.h>
//#include <synaps/viewer/geomview/point_graph.h>


 extern ::SYNAPS::MPol<double> poly;

__BEGIN_NAMESPACE_SYNAPS

// polynomial is the tensor of coefficients of a polynomial
// value is the point at which we want to evaluate the polynomial
// degree is the degree of the polynomial in the variable corresponding
// to the edge
template <class REP3D,class R>
R evaluateTensorOnEdge(REP3D polynomial,
		       unsigned edge,R value)
{
  // We first extract the one dimensional representation of the polynomial
  // on the edge
  bezier::rep1d<R> repOnTheEdge;
  R result;

  extractRepOfEdgeFromTensor(polynomial,edge,repOnTheEdge);

  // We now evaluate that polynomial at the given point, using 
  // de Casteljau's algorithm

  if (repOnTheEdge.size() == 0)
    result = 0;
  else
    result = brnops::eval(repOnTheEdge.tab_,repOnTheEdge.size(),value);
  
  return result;
}

// This function modifies point as follows. It converts the value into a
// two-dimensional point, and adds it to the parameter.
template <class C,class MP,class REP3D>
void modifyBarycenter(const cell3d<MP,REP3D> & cell,
		      topology::point<C> & point,
		      C value,unsigned edge,unsigned face)
{
  if (face < 1 || face > 6) {
    std::cerr << "Wrong face number: " << face
	      << ". Must be 1-6" << std::endl;
    
    assert(face >= 1 && face <= 6);
  }    
  
  if (edge < 1 || edge > 12) {
    std::cerr << "Wrong face number: " << edge 
	      << ". Must be 1-12" << std::endl;
    assert(edge >= 1 && edge <= 12);
  }
  
  topology::point<C> p3d(3);

  convertRootToPoint(cell,value,edge,p3d);

  topology::point<C> p2d(2);

  dropExtraCoordinate(p3d,face,p2d);

  point[0] += p2d[0];point[1] += p2d[1];
}

template <class C,class MP,class REP3D>
void getPointInsideACube(const cell3d<MP,REP3D> & cell,
			 const Seq<Seq<C> > & roots,
			 const Seq<Seq<Seq<C> > > & evalDerivatives,
			 topology::point<C> & result)
{
  result.resize(3);
  MatrDse<C,lapack::rep2d<C> > A,B,X;

  unsigned numberOfRoots = 0;

  for (unsigned i = 0;i < roots.size();i++)
    numberOfRoots += roots[i].size();

  A.rep().resize(numberOfRoots,3);
  B.rep().resize(numberOfRoots,1);
  X.rep().resize(3,1);

  unsigned rootsProcessed = 0;

  topology::point<C> barycenter(3);

  barycenter[0] = 0;barycenter[1] = 0;barycenter[2] = 0;

  for (unsigned edge = 0;edge < 12;edge ++) {
    for (unsigned rootNumber = 0;rootNumber < roots[edge].size();
	 rootNumber++) {
      // We add the equation
      A(rootsProcessed,0) = evalDerivatives[0][edge][rootNumber];
      A(rootsProcessed,1) = evalDerivatives[1][edge][rootNumber];
      A(rootsProcessed,2) = evalDerivatives[2][edge][rootNumber];
      
      // Then we add the constant
      topology::point<C> p3d(3);
      convertRootToPoint(cell,roots[edge][rootNumber],edge + 1,p3d);
      
      B(rootsProcessed,0) = 
	evalDerivatives[0][edge][rootNumber]*p3d[0]
	+ evalDerivatives[1][edge][rootNumber]*p3d[1]
	+ evalDerivatives[2][edge][rootNumber]*p3d[2];

      barycenter[0] += p3d[0];
      barycenter[1] += p3d[1];
      barycenter[2] += p3d[2];

      rootsProcessed++;
    }
  }
  
  barycenter[0] = barycenter[0] / rootsProcessed;
  barycenter[1] = barycenter[1] / rootsProcessed;
  barycenter[2] = barycenter[2] / rootsProcessed;
  
  // Now, we perturbate the system with the barycenter
  
  rootsProcessed = 0;
  
  for (int edge = 0;edge < 12;edge++)
    for (int i = 0;i < evalDerivatives[0][edge].size();i++)
      B(rootsProcessed++,0) -=
	evalDerivatives[0][edge][i] * barycenter[0]
 	+ evalDerivatives[1][edge][i] * barycenter[1]
 	+ evalDerivatives[2][edge][i] * barycenter[2];

  //  VECTOR::print(std::cout,Svd(A));

  //std::cout << std::endl;
  
  // We solve it
  X = Solve<LS>(A,B);
  
  // And we shift the result back to (0,0), by adding the barycenter
  result[0] = /*X(0,0) + */barycenter[0];
  result[1] = /*X(1,0) + */barycenter[1];
  result[2] = /*X(2,0) + */barycenter[2];
}

// // This function probably works only with "interesting faces", i.e. 
// // faces having more than one edge with roots
// template <class C,class MP,class REP3D>
// void getPointOnAFace(const cell<MP,REP3D> & cell,
// 		     const Seq<Seq<C> > & roots,
// 		     const Seq<Seq<Seq<C> > > & evalDerivatives,
// 		     unsigned face,
// 		     topology::point<C> & result)
// {
//   Seq<unsigned> vars;
//   variablesOfAFace(face,vars);

//   std::vector<unsigned> edges;
//   edgesOfAFace(face,edges);

//   // We initialize the barycenter's coordinates
//   topology::point<C> barycenter(2);
//   barycenter[0] = 0;barycenter[1] = 0;

//   unsigned numberOfRootsOnTheFace;

//   numberOfRootsOnTheFace = roots[edges[0]].size() + roots[edges[1]].size() + 
//     roots[edges[2]].size() + roots[edges[3]].size();

//   if (numberOfRootsOnTheFace == 2) {
//     // Assuming we have an interesting face, we can only have one root
//     // on each of just two edges of the face
//     result.resize(2);
//     for (unsigned i = 0;i < 4;i++) 
//       if (roots[edges[i]].size() == 1) {
// 	topology::point<C> p2d;
// 	topology::point3d<C> p3d;

// 	convertRootToPoint(cell,roots[edges[i]][0],edges[i] + 1,p3d);
	
// 	dropExtraCoordinate(p3d,face,p2d);

// // 	barycenter[0] += p2d[0];
// // 	barycenter[1] += p2d[1];
// 	result[0] = p2d[0];
// 	result[1] = p2d[1];
//       }
    
// //     result[0] = barycenter[0] / 2;
// //     result[1] = barycenter[1] / 2;

//     return;
//   }

//   // These matrices will contain our system
//   MatrDse<C,lapack::rep2d<C> > A,B,X;
//   A.rep().resize(roots[edges[0]].size() + roots[edges[1]].size() + 
// 		 roots[edges[2]].size() + roots[edges[3]].size(),2);
//   B.rep().resize(roots[edges[0]].size() + roots[edges[1]].size() + 
// 		 roots[edges[2]].size() + roots[edges[3]].size(),1);
//   X.rep().resize(2,1);
  
//   unsigned rootsProcessed = 0;

//   for (unsigned edge = 0;edge < 4;edge++) { // For each of the four edges
//     // We process the current edge
//     for (unsigned rootNumber = 0;rootNumber < roots[edges[edge]].size();
// 	 rootNumber++) {
      
//       // First we add the coefficients of the equation
//       A(rootsProcessed,0) = evalDerivatives[vars[0]][edges[edge]][rootNumber];
//       A(rootsProcessed,1) = evalDerivatives[vars[1]][edges[edge]][rootNumber];
      
//       // Then we add the constant
//       topology::point3d<C> p3d;
//       convertRootToPoint(cell,roots[edges[edge]][rootNumber],
// 			 edges[edge] + 1,p3d);

//       topology::point<C> p2d;
//       dropExtraCoordinate(p3d,face,p2d);

//       B(rootsProcessed,0) = 
// 	evalDerivatives[vars[0]][edges[edge]][rootNumber]*p2d[0]
// 	+ evalDerivatives[vars[1]][edges[edge]][rootNumber]*p2d[1];

//       // We add the corresponding coordinate to the barycenter
//       //      modifyBarycenter(cell,point,value edge face
//       modifyBarycenter(cell,barycenter,roots[edges[edge]][rootNumber],
// 		       edges[edge] + 1,face);
//       rootsProcessed++;
//     }
//   }
  
//   barycenter[0] = barycenter[0] / rootsProcessed;
//   barycenter[1] = barycenter[1] / rootsProcessed;
  
//   // We have to add the barycentrical perturbation here
//   // For each root on the face, we have to add the corresponding perturbation
//   // to its equation
  
//   rootsProcessed = 0;
 
//   for (int edge = 0;edge < 4;edge++)
//     for (int i = 0;i < evalDerivatives[vars[0]][edges[edge]].size();i++)
//       B(rootsProcessed++,0) -=
// 	evalDerivatives[vars[0]][edges[edge]][i] * barycenter[0]
//  	+ evalDerivatives[vars[1]][edges[edge]][i] * barycenter[1];
  
//   X = Solve<LS>(A,B);
  
//   result.resize(2);
  
//   // We shift the result back to (0,0), by adding the barycenter
//   result[0] = X(0,0) + barycenter[0];
//   result[1] = X(1,0) + barycenter[1];
// }

template <class C>
bool isAnInterestingFace(const Seq<Seq<C> > & roots,unsigned face)
{
  unsigned numberOfInterestingEdges = 0;
  std::vector<unsigned> edges;
  
  edgesOfAFace(face,edges);

  unsigned i = 0;
  while (i < 4 && numberOfInterestingEdges < 2) {
    if (roots[edges[i]].size() != 0) {
      numberOfInterestingEdges++;
    }
    i++;
  }
  
  return (numberOfInterestingEdges >= 2);
}

template <class C,class MP,class REP3D>
void convertRootsToPoints(const cell3d<MP,REP3D> & cell,
			  const Seq<Seq<C> > & roots,
			  Seq<Seq<topology::point<C> > > & points)
{
  // We have to process each edge
  points.resize(12);

  for (unsigned edge = 0;edge < 12;edge++) {
    Seq<topology::point<C> > sec(roots[edge].size());
    
    points[edge] = sec;
    
    for (unsigned i = 0;i < points[edge].size();i++) {
      convertRootToPoint(cell,roots[edge][i],edge + 1,points[edge][i]);
    }
  }
}

// This function creates new vertices from the points in the three sequences
// received as parameters, and returns isomorphic sequences containing
// pointers to those new vertices.
template<class C>
void convertPointsToVertices(const Seq<Seq<topology::point<C> > > & roots,
			     const Seq<topology::point<C> > & onTheFaces,
			     const topology::point<C> & insideTheCube,
			     Seq<Seq<topology::vertex<topology::point<C> > * > >
			     & vRoots,
			     Seq<topology::vertex<topology::point<C> > * >
			     & vOnTheFaces,
			     topology::vertex<topology::point<C> > *
			     & vInsideTheCube)
{
  
  vInsideTheCube = new topology::vertex<topology::point<C> >(insideTheCube[0],
							 insideTheCube[1],
							 insideTheCube[2]);

  // Now, we convert the points on the faces

  vOnTheFaces.resize(onTheFaces.size());

  for (unsigned face = 0;face < onTheFaces.size();face++) {
    vOnTheFaces[face] = new topology::vertex<topology::point<C> >
      (onTheFaces[face][0],
       onTheFaces[face][1],
       onTheFaces[face][2]);
  }

  vRoots.resize(roots.size());

  for (unsigned edge = 0;edge < roots.size();edge++) {
    vRoots[edge].resize(roots[edge].size());
    for (unsigned i = 0;i < roots[edge].size();i++) {
      vRoots[edge][i] = new topology::vertex<topology::point<C> >
	(roots[edge][i][0],
	 roots[edge][i][1],
	 roots[edge][i][2]);
    }
  }
}

// If this function returns true, we will have the computed points in
// onTheFaces, for the interesting faces; and in insideTheCube. Of course
// also in rootsOnTheEdges.
template <class C,class MP,class REP3D>
bool shouldWeMesh(const cell3d<MP,REP3D> & cell,
		  Seq<Seq<topology::point<C> > > & rootsOnTheEdges,
		  Seq<topology::point<C> > & onTheFaces,
		  topology::point<C> & insideTheCube)
{
  insideTheCube.resize(3);
  unsigned interestingFaces = 0;
  bool everythingInside = true;
  topology::point<C> p2d;
  Seq<Seq<C> > roots;
  Seq<Seq<Seq<C> > > evalDerivatives;
  
  roots.resize(12);
  for (unsigned i = 1;i <= 12;i++)
    computeRootsOnEdge(cell,roots,i);

  // We convert the roots we have into 3d-points

  rootsOnTheEdges.resize(12);
  for (unsigned edge = 0;edge < 12;edge ++) {
    rootsOnTheEdges[edge].resize(roots[edge].size());

    for (unsigned i = 0;i < roots[edge].size();i++)
      convertRootToPoint(cell,roots[edge][i],edge + 1,
			 rootsOnTheEdges[edge][i]);
  }

  computeAndEvaluateDerivativesAtRoots(cell,roots,evalDerivatives);

  onTheFaces.resize(6);
  
  // This loop studies the "meshability" of each face
  for (unsigned face = 1;face <= 6;face++) {
    if (isAnInterestingFace(roots,face)) {
      interestingFaces++;
      
      getPointOnAFace(cell,roots,evalDerivatives,face,p2d);

      convert2DPointInto3DPoint(cell,p2d,onTheFaces[face - 1],face);
      
      //      std::cout << onTheFaces[face - 1] << cell.box << std::endl;

      everythingInside = everythingInside &&
	liesInsideTheCell(cell,onTheFaces[face - 1]);
    }
  }
  //  std::cout << interestingFaces << "," << everythingInside;

  // This is actually a bit conservative, and we could improve it
  if (interestingFaces >= 2) {
    getPointInsideACube(cell,roots,evalDerivatives,insideTheCube);
    everythingInside = everythingInside &&
      liesInsideTheCell(cell,insideTheCube);
  }
  //  std::cout <<"," << /*everythingInside <<"," <<*/ insideTheCube << std::endl;
  return /* everythingInside && */ (interestingFaces >= 2);
}

template <class C,class MP,class REP3D>
bool shouldWeMeshASingularCell(const cell3d<MP,REP3D> & cell,
			       Seq<Seq<topology::point<C> > > &rootsOnTheEdges,
			       topology::point<C> & insideTheCube)
{
  insideTheCube.resize(3);
  bool everythingInside = true;
  Seq<Seq<C> > roots;
  Seq<Seq<Seq<C> > > evalDerivatives;

  roots.resize(12);
  for (unsigned i = 1;i <= roots.size();i++)
    computeRootsOnEdge(cell,roots,i);
  
  // We convert the roots we have into 3d-points
  rootsOnTheEdges.resize(roots.size());
  for (unsigned edge = 0;edge < roots.size();edge ++) {
    rootsOnTheEdges[edge].resize(roots[edge].size());

    for (unsigned i = 0;i < roots[edge].size();i++)
      convertRootToPoint(cell,roots[edge][i],edge + 1,
			 rootsOnTheEdges[edge][i]);
  }

  computeAndEvaluateDerivativesAtRoots(cell,roots,evalDerivatives);

  getPointInsideACube(cell,roots,evalDerivatives,insideTheCube);
  everythingInside =  liesInsideTheCell(cell,insideTheCube);

  return everythingInside;
}

// // In this function, we assume that the cell has a singularity inside.
// // To mesh it, we compute a point inside the cube by using least squares
// // on the planes tangent to the implicit surface at the roots on the edges
// // of the cell. We then use that point as the "inside" vertex of our triangles.
// // The other points are the roots we found on the edges, connected clockwise
// // if we look at each face from inside the cube.
// template <class MP,class REP3D>
// bool processSingularCell(scene<MP,REP3D> & scn,const cell3d<MP,REP3D> & cell)
// {
//   bool weShouldMesh;

//   Seq<Seq<topology::point3d<double> > > rootsOnTheEdges;
//   topology::point3d<double> insideTheCube;

//   weShouldMesh = shouldWeMeshASingularCell(cell,rootsOnTheEdges,insideTheCube);

//   if (weShouldMesh) {
//     Seq<Seq<topology::vertex<topology::point<double> > * > > vRoots;
//     topology::vertex<topology::point<double> > * vInsideTheCube;
    
//     // These two sequences are just dummies, to be able to use the 
//     // same functions we use with normal cells
//     Seq<topology::point3d<double> > onTheFaces;
//     Seq<topology::vertex<topology::point<double> > * > vOnTheFaces;

//     convertPointsToVertices(rootsOnTheEdges,onTheFaces,insideTheCube,
// 			    vRoots,vOnTheFaces,vInsideTheCube);
    
//     buildMeshInSingularCell(scn,cell,vRoots,vInsideTheCube);
//   }
//   return weShouldMesh;
// }

template <class MP,class REP3D>
void meshFaceUsingTangentInformation(scene<MP,REP3D> & scn,
				     const cell3d<MP,REP3D> &cell,
				     const Seq<Seq<Seq<typename MP::coeff_t>
				     > > & evalDerivatives,
				     const Seq<Seq<topology::point
				     <typename MP::coeff_t> > > & roots,
				     const topology::point
				     <typename MP::coeff_t> & singularity,
				     unsigned face)
{
  // We try all the possible segments on the face, and we discard those that
  // are not "very" compatible with the information we obtain from the
  // tangents to the curve on the surface.
  Seq<unsigned> vars;
  variablesOfAFace(face,vars);
  std::vector<unsigned> edges;
  edgesOfAFace(face,edges);
  int initialRoot,firstRoot,secondRoot;
  unsigned initialEdge,firstEdge,secondEdge;

  typedef typename MP::coeff_t C;
  
  initialEdge = 0;
  initialRoot = -1;
  
  // We initialize the initial root
  nextRoot(roots,initialEdge,initialRoot,face);

  firstRoot = initialRoot;
  firstEdge = initialEdge;

  do {

    secondEdge = firstEdge;
    secondRoot = firstRoot;
    for (nextRoot(roots,secondEdge,secondRoot,face);
	 secondEdge != initialEdge || secondRoot != initialRoot;
	 nextRoot(roots,secondEdge,secondRoot,face)) {
      // We now have to test the segments, to see if they are near
      // the tangents
      bool theyAreClose = true;

      topology::point<typename MP::coeff_t> segment,tangent;
      segment.resize(2);
      tangent.resize(2);
      
      segment[0] = roots[edges[secondEdge]][secondRoot][vars[0]] - 
	roots[edges[firstEdge]][firstRoot][vars[0]];
      segment[1] = roots[edges[secondEdge]][secondRoot][vars[1]] - 
	roots[edges[firstEdge]][firstRoot][vars[1]];

      // We invert the coordinates of the normal, to get the direction
      // of the tangent
      // First point
      tangent[0] = - evalDerivatives[vars[1]][edges[firstEdge]][firstRoot];
      tangent[1] = evalDerivatives[vars[0]][edges[firstEdge]][firstRoot];

      theyAreClose = theyAreClose && 
	doTheyHaveASimilarDirection(tangent,segment);

      // Second point
      tangent[0] = - evalDerivatives[vars[1]][edges[secondEdge]][secondRoot];
      tangent[1] = evalDerivatives[vars[0]][edges[secondEdge]][secondRoot];
      
      theyAreClose = theyAreClose && 
	doTheyHaveASimilarDirection(tangent,segment);

      segment.resize(3);
      tangent.resize(3);

      // segment will now hold the normal to the triangle
      

      if (theyAreClose) {
	// We build the triangle
	///////////////// HIGHLY INEFFICIENT - ONLY TOLERABLE IN A TEST ////
	// WE HAVE TO ADD THE VERTICES ONLY ONCE. OTHERWISE, WE ARE WASTING
	// SPACE
	topology::vertex<topology::point<C> > *vertex1, *vertex2, *vertex3;
	vertex1 = new topology::vertex<topology::point<C> > (singularity[0],
							 singularity[1],
							 singularity[2]);
	insertVertexInList(scn,vertex1);

 	vertex2 = new topology::vertex<
	  topology::point<C> > (roots[edges[firstEdge]][firstRoot][0],
			      roots[edges[firstEdge]][firstRoot][1],
			      roots[edges[firstEdge]][firstRoot][2]);
	insertVertexInList(scn,vertex2);

	vertex3 = new topology::vertex<
	  topology::point<C> > (roots[edges[secondEdge]][secondRoot][0],
			      roots[edges[secondEdge]][secondRoot][1],
			      roots[edges[secondEdge]][secondRoot][2]);
	insertVertexInList(scn,vertex3);

	makeTriangle(scn,vertex1,vertex2,vertex3);
      }
    }
    nextRoot(roots,firstEdge,firstRoot,face);
  } while (firstEdge != initialEdge || firstRoot != initialRoot);
}

template <class MP,class REP3D,class REP2D>
void getRepresentationOnTheFace(const cell3d<MP,REP3D> & tensor,
				cell2d<typename MP::coeff_t,REP2D> & matrix,
				unsigned face)
{
  int i,j;
  switch (face) {
  case 1:
    // We need the xs and ys, with z = z0
    matrix.bz_pol.resize(tensor.bezier.nbcol(),tensor.bezier.nbrow());
    for (int x = 0;x < tensor.bezier.nbcol();x++)
      for (int y = 0;y < tensor.bezier.nbrow();y++)
	matrix.bz_pol(x,y) = tensor.bezier(y,x,0);
    matrix.x0 = tensor.box.x0;
    matrix.x1 = tensor.box.x1;
    matrix.y0 = tensor.box.y0;
    matrix.y1 = tensor.box.y1;
    break;
  case 2:
    // We need the xs and zs, with y = y0
    matrix.bz_pol.resize(tensor.bezier.nbcol(),tensor.bezier.nbhgt());
    for (int x = 0;x < tensor.bezier.nbcol();x++)
      for (int z = 0;z < tensor.bezier.nbhgt();z++)
	matrix.bz_pol(x,z) = tensor.bezier(0,x,z);
    matrix.x0 = tensor.box.x0;
    matrix.x1 = tensor.box.x1;
    matrix.y0 = tensor.box.z0;
    matrix.y1 = tensor.box.z1;
    break;
  case 3:
    // We need the xs and ys, with z = z1
    matrix.bz_pol.resize(tensor.bezier.nbcol(),tensor.bezier.nbhgt());
    for (int x = 0;x < tensor.bezier.nbcol();x++)
      for (int y = 0;y < tensor.bezier.nbrow();y++)
	matrix.bz_pol(x,y) = tensor.bezier(y,x,tensor.bezier.nbhgt() - 1);
    matrix.x0 = tensor.box.x0;
    matrix.x1 = tensor.box.x1;
    matrix.y0 = tensor.box.y0;
    matrix.y1 = tensor.box.y1;
    break;
  case 4:
    // We need the ys and zs, with x = x1
    matrix.bz_pol.resize(tensor.bezier.nbrow(),tensor.bezier.nbhgt());
    for (int y = 0;y < tensor.bezier.nbrow();y++)
      for (int z = 0;z < tensor.bezier.nbhgt();z++)
	matrix.bz_pol(y,z) = tensor.bezier(y,tensor.bezier.nbcol() - 1,z);
    matrix.x0 = tensor.box.y0;
    matrix.x1 = tensor.box.y1;
    matrix.y0 = tensor.box.z0;
    matrix.y1 = tensor.box.z1;
    break;
  case 5:
    // We need the xs and zs, with y = y1
    matrix.bz_pol.resize(tensor.bezier.nbcol(),tensor.bezier.nbhgt());
    for (int x = 0;x < tensor.bezier.nbcol();x++)
      for (int z = 0;z < tensor.bezier.nbhgt();z++)
	matrix.bz_pol(x,z) = tensor.bezier(tensor.bezier.nbrow() - 1,x,z);
    matrix.x0 = tensor.box.x0;
    matrix.x1 = tensor.box.x1;
    matrix.y0 = tensor.box.z0;
    matrix.y1 = tensor.box.z1;
    break;
  case 6:
    // We need the ys and zs, with x = x0
    matrix.bz_pol.resize(tensor.bezier.nbrow(),tensor.bezier.nbhgt());
    for (int y = 0;y < tensor.bezier.nbrow();y++)
      for (int z = 0;z < tensor.bezier.nbhgt();z++)
	matrix.bz_pol(y,z) = tensor.bezier(y,0,z);
    matrix.x0 = tensor.box.y0;
    matrix.x1 = tensor.box.y1;
    matrix.y0 = tensor.box.z0;
    matrix.y1 = tensor.box.z1;
    break;
  }
}

struct SLV 
{
  typedef Seq<double> u_solution;
  template<class POL>
  static u_solution u_solve(const POL & p,double x0, double x1)
  {
    return Solve(p,SlvBzBdg<double>(1e-6),x0,x1);
  }
};

template <class MP,class REP3D>
void meshAFaceOfACell(scene<MP,REP3D> & scn,
		      const cell3d<MP,REP3D> & cell,
		      const topology::point<typename MP::coeff_t>
		      & insideTheCube,
		      unsigned face)
{
  // This point graph will be used to hold the results of the 2D meshing
  // algorithm
  topology::point_graph<typename MP::coeff_t> pg;
  cell2d<typename MP::coeff_t> bz_face;

  // We have to build a 2D cell with the information of the face
  getRepresentationOnTheFace(cell,bz_face,face);

  //  std::cout << bz_face << std::endl;
  // Once we have the representation in the 2d cell, we call the subdivision
  // solver
  //XXX  subdivide(pg,bz_face,SLV());


//   char  x[10];
//   sprintf(x,"f%d",face);
//   geomview::ostream output(x);
//   output << pg;
//   output.view();
  
  // Now, we have to work with the solutions we have in pg

  // First, we convert those solutions into vertices
  Seq<topology::vertex<topology::point<typename MP::coeff_t> > * > vertices;
  convert2Dinto3DReversing(scn,pg,cell,vertices,face);

  topology::vertex<topology::point<typename MP::coeff_t> > * vInsideTheCube =
    new topology::vertex<topology::point<typename MP::coeff_t> >
    (insideTheCube[0],insideTheCube[1],insideTheCube[2]);

  insertVertexInList(scn,vInsideTheCube);

  // Now, we use the edges to create triangles

  for (int i = 0;i < pg.Edges.size();i++) {
    for (int j = 0;j < pg.Edges[i].size();j++) {
      makeTriangle(scn,vInsideTheCube,vertices[i],vertices[pg.Edges[i][j]]);
    }
  }
}

template <class MP,class REP3D>
bool processXRegularCell(scene<MP,REP3D> & scn,const cell3d<MP,REP3D> & cell)
{
  typedef typename MP::coeff_t coeff_t;

  bool result = false;
  unsigned edgesToAnalyze[] = {0,1,3,2};

  // We first compute the roots on the edges
  Seq<Seq<coeff_t> > roots;
  roots.resize(12);
  for (unsigned i = 1;i <= 12;i++)
    computeRootsOnEdge(cell,roots,i);

  Seq<unsigned> relevantEdges; // This sequence will hold the nonempty edges

  // We find out which relevant edges have just one root
  for (unsigned i = 0;i <4;i++) {
    if (roots[edgesToAnalyze[i]].size() == 1)
      relevantEdges.push_back(edgesToAnalyze[i]);
  }

  result = relevantEdges.size() > 2;

  if (result) {
    Seq<topology::point<coeff_t> > points(relevantEdges.size());
    Seq<topology::vertex<topology::point<coeff_t> > * > 
      vertices(relevantEdges.size());

    for (unsigned i = 0;i < relevantEdges.size();i++) {
      convertRootToPoint(cell,roots[relevantEdges[i]][0],
			 relevantEdges[i] + 1,points[i]);
      vertices[i] = new topology::vertex<topology::point<coeff_t> >
	(points[i][0],points[i][1],points[i][2]);
      insertVertexInList(scn,vertices[i]);
    }

    // Now we mesh
    makeTriangle(scn,vertices[0],vertices[1],vertices[2]);
    if (relevantEdges.size() == 4) {
      makeTriangle(scn,vertices[2],vertices[3],vertices[0]);
    }
  }

  return result;
}

template <class MP,class REP3D>
bool processYRegularCell(scene<MP,REP3D> & scn,const cell3d<MP,REP3D> & cell)
{
  typedef typename MP::coeff_t coeff_t;

  bool result = false;
  unsigned edgesToAnalyze[] = {8,9,11,10};

  // We first compute the roots on the edges
  Seq<Seq<coeff_t> > roots;
  roots.resize(12);
  for (unsigned i = 1;i <= 12;i++)
    computeRootsOnEdge(cell,roots,i);

  Seq<unsigned> relevantEdges; // This sequence will hold the nonempty edges

  // We find out which relevant edges have just one root
  for (unsigned i = 0;i <4;i++) {
    if (roots[edgesToAnalyze[i]].size() == 1)
      relevantEdges.push_back(edgesToAnalyze[i]);
  }

  result = relevantEdges.size() > 2;

  if (result) {
    Seq<topology::point<coeff_t> > points(relevantEdges.size());
    Seq<topology::vertex<topology::point<coeff_t> > * > 
      vertices(relevantEdges.size());

    for (unsigned i = 0;i < relevantEdges.size();i++) {
      convertRootToPoint(cell,roots[relevantEdges[i]][0],
			 relevantEdges[i] + 1,points[i]);
      vertices[i] = new topology::vertex<topology::point<coeff_t> >
	(points[i][0],points[i][1],points[i][2]);
      insertVertexInList(scn,vertices[i]);
    }

    // Now we mesh
    makeTriangle(scn,vertices[0],vertices[1],vertices[2]);
    if (relevantEdges.size() == 4) {
      makeTriangle(scn,vertices[2],vertices[3],vertices[0]);
    }
  }

  return result;
}

template <class MP,class REP3D>
bool processZRegularCell(scene<MP,REP3D> & scn,const cell3d<MP,REP3D> & cell)
{
  typedef typename MP::coeff_t coeff_t;

  bool result = false;
  unsigned edgesToAnalyze[] = {4,6,7,5};

  // We first compute the roots on the edges
  Seq<Seq<coeff_t> > roots;
  roots.resize(12);
  for (unsigned i = 1;i <= 12;i++)
    computeRootsOnEdge(cell,roots,i);

  Seq<unsigned> relevantEdges; // This sequence will hold the nonempty edges

  // We find out which relevant edges have just one root
  for (unsigned i = 0;i <4;i++) {
    if (roots[edgesToAnalyze[i]].size() == 1)
      relevantEdges.push_back(edgesToAnalyze[i]);
  }

  result = relevantEdges.size() > 2;

  if (result) {
    Seq<topology::point<coeff_t> > points(relevantEdges.size());
    Seq<topology::vertex<topology::point<coeff_t> > * > 
      vertices(relevantEdges.size());

    for (unsigned i = 0;i < relevantEdges.size();i++) {
      convertRootToPoint(cell,roots[relevantEdges[i]][0],
			 relevantEdges[i] + 1,points[i]);
      vertices[i] = new topology::vertex<topology::point<coeff_t> >
	(points[i][0],points[i][1],points[i][2]);
      insertVertexInList(scn,vertices[i]);
    }

    // Now we mesh
    makeTriangle(scn,vertices[0],vertices[1],vertices[2]);
    if (relevantEdges.size() == 4) {
      makeTriangle(scn,vertices[2],vertices[3],vertices[0]);
    }
  }

  return result;
}

inline
bool doTheyShareAFace(unsigned e1,unsigned e2,unsigned & face)
{
  bool result = false;

  if ((e1 == 1 || e1 == 3 || e1 == 9 || e1 == 10) && 
      (e2 == 1 || e2 == 3 || e2 == 9 || e2 == 10)) {
    // Face 1
    result = true;
    face = 1;
  } else if ((e1 == 1 || e1 == 2 || e1 == 5 || e1 == 7) && 
      (e2 == 1 || e2 == 2 || e2 == 5 || e2 == 7)) {
    // Face 2
    result = true;
    face = 2;
  } else if ((e1 == 2 || e1 == 4 || e1 == 11 || e1 == 12) && 
      (e2 == 2 || e2 == 4 || e2 == 11 || e2 == 12)) {
    // Face 3
    result = true;
    face = 3;
  } else if ((e1 == 7 || e1 == 8 || e1 == 10 || e1 == 12) && 
      (e2 == 7 || e2 == 8 || e2 == 10 || e2 == 12)) {
    // Face 4
    result = true;
    face = 4;
  } else if ((e1 == 3 || e1 == 4 || e1 == 6 || e1 == 8) && 
      (e2 == 3 || e2 == 4 || e2 == 6 || e2 == 8)) {
    // Face 5
    result = true;
    face = 5;
  } else if ((e1 == 5 || e1 == 6 || e1 == 9 || e1 == 11) && 
      (e2 == 5 || e2 == 6 || e2 == 9 || e2 == 11)) {
    // Face 6
    result = true;
    face = 6;
  }
  return result;
}

// It returns true if there is no root on the edges and no sign change inside
// or if there are roots on the edges
template <class MP,class REP3D>
bool analyzeFace(const cell3d<MP,REP3D> & cell,unsigned face,
		 const Seq<Seq<typename MP::coeff_t> > & roots)
{

  bool result = true;

  bool allPositive;
  bool allNegative;

  allPositive = true;
  allNegative = true;

  std::vector<unsigned> edges;
  edgesOfAFace(face,edges);

  if ((roots[edges[0]].size() + roots[edges[1]].size() + roots[edges[2]].size()
       + roots[edges[3]].size()) == 0) {
    switch (face) {
    case 1:
      for (unsigned i = 1;i < cell.bezier.nbcol() - 1;i++)
	for (unsigned j = 1;j < cell.bezier.nbrow() - 1;j++)
	  if (cell.bezier(j,i,0) >= 0)
	    allNegative = false;
	  else
	    allPositive = false;
      break;
    case 2:
      for (unsigned i = 1;i < cell.bezier.nbcol() - 1;i++)
	for (unsigned k = 1;k < cell.bezier.nbhgt() - 1;k++)
	  if (cell.bezier(0,i,k) >= 0)
	    allNegative = false;
	  else
	    allPositive = false;
      break;
    case 3:
      for (unsigned i = 1;i < cell.bezier.nbcol() - 1;i++)
	for (unsigned j = 1;j < cell.bezier.nbrow() - 1;j++)
	  if (cell.bezier(j,i,cell.bezier.nbhgt() - 1) >= 0)
	    allNegative = false;
	  else
	    allPositive = false;
      break;
    case 4:
      for (unsigned j = 1;j < cell.bezier.nbrow() - 1;j++)
	for (unsigned k = 1;k < cell.bezier.nbhgt() - 1;k++)
	  if (cell.bezier(j,cell.bezier.nbcol() - 1,k) >= 0)
	    allNegative = false;
	  else
	    allPositive = false;
      break;
    case 5:
      for (unsigned i = 1;i < cell.bezier.nbcol() - 1;i++)
	for (unsigned k = 1;k < cell.bezier.nbhgt() - 1;k++)
	  if (cell.bezier(cell.bezier.nbrow() - 1,i,k) >= 0)
	    allNegative = false;
	  else
	    allPositive = false;
      break;
    case 6:
      for (unsigned j = 1;j < cell.bezier.nbrow() - 1;j++)
	for (unsigned k = 1;k < cell.bezier.nbhgt() - 1;k++)
	  if (cell.bezier(j,0,k) >= 0)
	    allNegative = false;
	  else
	    allPositive = false;
      break;
    default:
      std::cerr << "Wrong face number" << std::endl;
      assert(face >= 1 && face <= 6);
    }
    result = (allNegative || allPositive);
  }
  return result;
}

// variable indicates 
template <class MP,class REP3D>
bool processRegularCell(scene<MP,REP3D> & scn,const cell3d<MP,REP3D> & cell,
			unsigned variable)
{
  typedef typename MP::coeff_t C;
  bool result = false;

  // Now, we compute the roots of the polynomial on the edges
  // of the cell.
  unsigned numberOfRoots = 0,numberOfEdgesWithRoots = 0;
  bool atMostOnePerEdge = true;

  Seq<Seq<typename MP::coeff_t> > roots;
  roots.resize(12);
  for (unsigned i = 1;i <= 12;i++) {
    computeRootsOnEdge(cell,roots,i);
    numberOfRoots += roots[i-1].size();
    atMostOnePerEdge = atMostOnePerEdge && roots[i-1].size() <= 1;
    if (roots[i-1].size() > 0)
      numberOfEdgesWithRoots++;
  }
  
  bool facesOk = true;
  for (unsigned face = 1;face <= 6;face++)
    facesOk = facesOk && analyzeFace(cell,face,roots);

  if (!facesOk)
    return false;

  switch (variable) {
  case 0:
    if (!(analyzeFace(cell,4,roots) && analyzeFace(cell,6,roots)))
      return false;
    break;
  case 1:
    if (!(analyzeFace(cell,2,roots) && analyzeFace(cell,5,roots)))
      return false;
    break;
  case 2:
    if (!(analyzeFace(cell,1,roots) && analyzeFace(cell,3,roots)))
      return false;
    break;
  default:
    std::cerr << "Wrong variable number" << std::endl;
  }
  
  if (numberOfEdgesWithRoots <= 2) // We discard the cell
    result = true;
  else { // We don't discard it, we process it
    // First, we find a diagonal that intersects the surface
    int diagonal;
    topology::point<C> p;
    p.resize(3);

    diagonal = findDiagonal(cell.bezier);
    // Now, depnding on which diagonal we found, we find the intersection
    // with the surface. To do that, we generate the equations of two planes
    // intersecting on the diagonal

    std::vector<typename MP::coeff_t> domains;
    domains.resize(6);
    domains[0] = cell.box.x0*scn.u+scn.x;domains[1] = cell.box.x1*scn.u+scn.x;
    domains[2] = cell.box.y0*scn.v+scn.y;domains[3] = cell.box.y1*scn.v+scn.y;
    domains[4] = cell.box.z0*scn.w+scn.z;domains[5] = cell.box.z1*scn.w+scn.z;

    topology::point<C> p0,p1,direction,normal1,normal2;
    p0.resize(3);
    p1.resize(3);
    direction.resize(3);
    normal1.resize(3);
    normal2.resize(3);

    switch (diagonal) {
    case 0:
      return false;
    case 1:
      p0[0] = domains[0];p0[1] = domains[2];p0[2] = domains[4];
      p1[0] = domains[1];p1[1] = domains[3];p1[2] = domains[5];
      break;
    case 2:
      p0[0] = domains[0];p0[1] = domains[2];p0[2] = domains[5];
      p1[0] = domains[1];p1[1] = domains[3];p1[2] = domains[4];
      break;
    case 3:
      p0[0] = domains[1];p0[1] = domains[2];p0[2] = domains[4];
      p1[0] = domains[0];p1[1] = domains[3];p1[2] = domains[5];
      break;
    case 4:
      p0[0] = domains[0];p0[1] = domains[3];p0[2] = domains[4];
      p1[0] = domains[1];p1[1] = domains[2];p1[2] = domains[5];
      break;
    }

    direction[0] = p1[0] - p0[0];
    direction[1] = p1[1] - p0[1];
    direction[2] = p1[2] - p0[2];

    normal1[0] = -direction[1];normal1[1] = direction[0];normal1[2] = 0;
    normal2[0] = 0;normal2[1] = -direction[2];normal2[2] = direction[1];

    std::list<MP> system;
    MP plane1,plane2;
    MP var0,var1,var2;
    var0 = "x0";var1 = "x1";var2="x2";

    plane1 = var0 * normal1[0] + var1 * normal1[1] + var2 * normal1[2] +
      direction[1]*p0[0] - direction[0]*p0[1];

    plane2 = var0 * normal2[0] + var1 * normal2[1] + var2 * normal2[2] +
      direction[2]*p1[1] - direction[1]*p1[2];

    system.push_back(poly);
    system.push_back(plane1);
    system.push_back(plane2);

    // Once we have a system with the polynomial, and the planes defining
    // the diagonal, we solve it.
    Seq<VectDse<typename MP::coeff_t> > results 
      = solve(system, 
	      SBDSLV<typename MP::coeff_t,SBDSLV_RDRDL>((typename MP::coeff_t) 0.0000001 ), 
		    domains);

    // We convert the solutions into points
    Seq<topology::point<typename MP::coeff_t> > intersections;
    intersections.resize(results.size());
    // We now have to move back our intersections into the box.
    for (int i = 0;i < intersections.size();i++) {
      intersections[i].resize(3);
      intersections[i][0] = (cell.box.x1 - cell.box.x0) * 
	(results[i][0] - domains[0]) / 
	(domains[1] - domains[0]) + cell.box.x0;
      
      intersections[i][1] = (cell.box.y1 - cell.box.y0) * 
	(results[i][1] - domains[2]) / 
	(domains[3] - domains[2]) + cell.box.y0;
      
      intersections[i][2] = (cell.box.z1 - cell.box.z0) *
	(results[i][2] - domains[4]) / 
	(domains[5] - domains[4]) + cell.box.z0;
    }

    // Now, we have to mesh using that point as the center of the meshing

    // For that, we convert the points into vertices.
    Seq<Seq<topology::point<C> > > rootsOnTheEdges;
    Seq<Seq<topology::vertex<topology::point<C> > *> > vRoots;
    Seq<topology::vertex<topology::point<C> > *> vOnTheFaces; // dummy
    Seq<topology::point<C> > onTheFaces; // dummy
    topology::vertex<topology::point<C> > * vInsideTheCube;

    convertRootsToPoints(cell,roots,rootsOnTheEdges);

    convertPointsToVertices(rootsOnTheEdges,onTheFaces,intersections[0],
			    vRoots,vOnTheFaces,vInsideTheCube);

    // We add all the vertices to the scene
    for (unsigned i = 0;i < vRoots.size();i++)
      for (unsigned j = 0;j < vRoots[i].size();j++)
	insertVertexInList(scn,vRoots[i][j]);
    insertVertexInList(scn,vInsideTheCube);



    // Finally, we mesh
    for (unsigned face = 1;face <= 6;face++) {
      std::vector<unsigned> edges;
      edgesOfAFace(face,edges);
      if (vRoots[edges[0]].size() + vRoots[edges[1]].size()+vRoots[edges[2]].size()+vRoots[edges[3]].size() > 1)
	meshFaceOnASingularCell(scn,cell,vRoots,vInsideTheCube,face);
    }
    
    
//     std::cout << std::boolalpha;
//     std::cout << liesInsideTheCell(cell,intersections[0]) << std::endl;
    result = true;
  }
  return result;
}

// In this function, we assume that the cell has a singularity inside.
// To mesh it, we compute a point inside the cube by using least squares
// on the planes tangent to the implicit surface at the roots on the edges
// of the cell. We then use that point as the "inside" vertex of our triangles.
// The other points are the roots we found on the edges, connected clockwise
// if we look at each face from inside the cube.
template <class MP,class REP3D>
bool processSingularCell(scene<MP,REP3D> & scn,const cell3d<MP,REP3D> & cell)
{
  std::cout <<"Singular cell"<<std::endl;
  typedef typename MP::coeff_t C;

  // First, we check whether there actually is a singularity inside
  // the cell. For that, we use a subdivision solver for the system
  // f = fx = fy = fz = 0

  MP fx,fy,fz;
  fx = diff(poly,"x0");
  fy = diff(poly,"x1");
  fz = diff(poly,"x2");

  std::list<MP> system;
  system.push_back(poly);
  system.push_back(fx);system.push_back(fy);system.push_back(fz);

  MatrDse<typename MP::coeff_t> results;
  std::vector<typename MP::coeff_t> domains;

  domains.resize(6);
  domains[0] = cell.box.x0*scn.u+scn.x;domains[1] = cell.box.x1*scn.u+scn.x;
  domains[2] = cell.box.y0*scn.v+scn.y;domains[3] = cell.box.y1*scn.v+scn.y;
  domains[4] = cell.box.z0*scn.w+scn.z;domains[5] = cell.box.z1*scn.w+scn.z;

  results = Solve(system, SBDSLV<typename MP::coeff_t,
		  SBDSLV_RDRDL>(
				(typename MP::coeff_t) 0.0000001 ), domains);
  if(!results.nbrow())
    return true;
  else
    return false;
  // We convert the solutions into points
  Seq<topology::point<typename MP::coeff_t> > singularities;
  singularities.resize(results.nbrow());
  for (int i = 0;i < singularities.size();i++) {
    singularities[i].resize(3);
    // We now have to move back our singularity into the box.
    singularities[i][0] = (cell.box.x1 - cell.box.x0) * 
      (results[i][0] - domains[0]) / 
      (domains[1] - domains[0]) + cell.box.x0;
    
    singularities[i][1] = (cell.box.y1 - cell.box.y0) * 
      (results[i][1] - domains[2]) / 
      (domains[3] - domains[2]) + cell.box.y0;
    
    singularities[i][2] = (cell.box.z1 - cell.box.z0) *
      (results[i][2] - domains[4]) / 
      (domains[5] - domains[4]) + cell.box.z0;
  }
  
  // Now, we compute the roots of the polynomial on the edges
  // of the cell.
  unsigned numberOfRoots = 0;
  bool atMostOnePerEdge = true;

  Seq<Seq<typename MP::coeff_t> > roots;
  roots.resize(12);
  for (unsigned i = 1;i <= 12;i++) {
    computeRootsOnEdge(cell,roots,i);
    numberOfRoots += roots[i-1].size();
    atMostOnePerEdge = atMostOnePerEdge && roots[i-1].size() <= 1;
  }
  
  // We convert the roots we have into 3d-points
  Seq<Seq<topology::point<C> > > rootsOnTheEdges;
  rootsOnTheEdges.resize(12);
  for (unsigned edge = 0;edge < 12;edge ++) {
    rootsOnTheEdges[edge].resize(roots[edge].size());
    for (unsigned i = 0;i < roots[edge].size();i++)
      convertRootToPoint(cell,roots[edge][i],edge + 1,
			 rootsOnTheEdges[edge][i]);
  }

  // Now, we proceed, depending on whether we found exactly one singularity 
  // or not
  if (singularities.size() == 1) { // We found a single singularity
    // In this case, we'll use the singularity as one vertex of the triangles
 
    Seq<Seq<Seq<C> > > evalDerivatives;
    computeAndEvaluateDerivativesAtRoots(cell,roots,evalDerivatives);

    // We convert the roots into vertices
    Seq<Seq<topology::vertex<topology::point<C> > * > > vRoots;
    topology::vertex<topology::point<C> > * vInsideTheCube;
    
    // These two sequences are just dummies, to be able to use the 
    // same functions we use with normal cells
    Seq<topology::point<C> > onTheFaces;
    Seq<topology::vertex<topology::point<C> > * > vOnTheFaces;
    
    convertPointsToVertices(rootsOnTheEdges,onTheFaces,singularities[0],
			    vRoots,vOnTheFaces,vInsideTheCube);
    
    // We now mesh each face
    for (int face = 1;face <= 6;face++) {
//        meshFaceUsingTangentInformation(scn,cell,evalDerivatives,rootsOnTheEdges,
//  				      singularities[0],face);
      meshAFaceOfACell(scn,cell,singularities[0],face);
    }
  } else if (singularities.size() == 0) { // We found no singularities
    // We have no singularities, but the subdivision hasn't been able to deal
    // with this cell

    // First we try a strategy for the case of 4 roots on the borders of the
    // box.

//     if (numberOfRoots == 4 && atMostOnePerEdge) {

//       Seq<unsigned> edgesOfTheRoots;
//       Seq<unsigned> adjacencies;
//       edgesOfTheRoots.resize(4);

//       // This populates a short array with the edge numbers corresponding
//       // to the roots we have
//       unsigned j = 0;
//       for (unsigned i = 0;i < 12;i++)
// 	if (roots[i].size() == 1) {
// 	  edgesOfTheRoots[j] = i+1;
// 	  j++;
// 	}
      
//       // Now, we are going to establish which of the pairs of edges from
//       // edgesOfTheRoots belong to a same face.

//       Seq<unsigned> adjacency;

//       for (unsigned i = 0;i < 4;i++)
// 	for (j = i;j < 4;j++) {
// 	  unsigned face;
// 	  if (doTheyShareAFace(edgesOfTheRoots[i],edgesOfTheRoots[j],face)) {
// 	    adjacency.push_back(i);
// 	    adjacency.push_back(j);
// 	    adjacency.push_back(face);
// 	  }
// 	}

//       // Now, we look for opposing faces
//       unsigned e1 = adjacency[0];
//       unsigned e2 = adjacency[1];
//       unsigned firstFace = adjacency[2];
//       unsigned opposingFace,adjacentFace;

//       unsigned i = 3;
//       bool foundAdjacent = false,foundOpposing = false;
//       while (i < adjacency.size() && !foundAdjacent && !foundOpposing) {
// 	if (!foundOpposing && (e1 != adjacency[i] && e2 != adjacency[i+1])) {
// 	  foundOpposing = true;
// 	  opposingFace = adjacency[i+2];
// 	}
	
// 	if (!foundAdjacent && 
// 	    (e1 == adjacency[i] && e2 != adjacency[i+1]) ||
// 	    (e1 != adjacency[i] && e2 == adjacency[i+1])) {
// 	  foundAdjacent = true;
// 	  adjacentFace = adjacency[i+2];
// 	}
// 	i += 3;
//       }
//     } else {  
      bool itWorked = false;
      
//       // We first try some of the "simple" cases
//       itWorked = processXRegularCell(scn,cell);
      
//       if (!itWorked) {
// 	itWorked = processYRegularCell(scn,cell);
//       }
      
//       if (!itWorked) {
// 	itWorked = processZRegularCell(scn,cell);
//       }
      
      return itWorked;
//     }
  } else { // We found many singularities
    // We'll think about this later
    
  } 


//   if (results.nbrow() == 0 || results.nbrow() > 1) { // We found no singularities
//     bool weShouldMesh = false;
    
// //     Seq<Seq<topology::point3d<C> > > rootsOnTheEdges;
// //     topology::point3d<double> insideTheCube;
    
// //     weShouldMesh = shouldWeMeshASingularCell(cell,rootsOnTheEdges,insideTheCube);
    
// //     if (weShouldMesh) {
// //       Seq<Seq<topology::vertex<topology::point<double> > * > > vRoots;
// //       topology::vertex<topology::point<double> > * vInsideTheCube;
      
// //       // These two sequences are just dummies, to be able to use the 
// //       // same functions we use with normal cells
// //       Seq<topology::point3d<double> > onTheFaces;
// //       Seq<topology::vertex<topology::point<double> > * > vOnTheFaces;
      
// //       convertPointsToVertices(rootsOnTheEdges,onTheFaces,insideTheCube,
// // 			      vRoots,vOnTheFaces,vInsideTheCube);
      
// //       buildMeshInSingularCell(scn,cell,vRoots,vInsideTheCube);
// //     }
//     return weShouldMesh;
//   } else if (results.nbrow() == 1) { // We found an isolated singularity
//     Seq<Seq<topology::vertex<topology::point<double> > * > > vRoots;
//     topology::vertex<topology::point<double> > * vInsideTheCube;
//     topology::point3d<double> insideTheCube;
//     Seq<Seq<topology::point3d<double> > > rootsOnTheEdges;

//     bool weShouldMesh = true;

//     Seq<Seq<typename MP::coeff_t> > roots;
//     roots.resize(12);
//     for (unsigned i = 1;i <= 12;i++)
//       computeRootsOnEdge(cell,roots,i);
    
//     // We convert the roots we have into 3d-points
    
//     rootsOnTheEdges.resize(12);
//     for (unsigned edge = 0;edge < 12;edge ++) {
//       rootsOnTheEdges[edge].resize(roots[edge].size());
      
//       for (unsigned i = 0;i < roots[edge].size();i++)
// 	convertRootToPoint(cell,roots[edge][i],edge + 1,
// 			   rootsOnTheEdges[edge][i]);
//     }

//     // We now have to move back our singularity into the box.
//     insideTheCube[0] = (cell.box.x1 - cell.box.x0) * 
//       (results(0,0) - domains[0]) / 
//       (domains[1] - domains[0]) + cell.box.x0;

//     insideTheCube[1] = (cell.box.y1 - cell.box.y0) * 
//       (results(0,1) - domains[2]) / 
//       (domains[3] - domains[2]) + cell.box.y0;

//     insideTheCube[2] = (cell.box.z1 - cell.box.z0) *
//       (results(0,2) - domains[4]) / 
//       (domains[5] - domains[4]) + cell.box.z0;
    
//     // These two sequences are just dummies, to be able to use the 
//     // same functions we use with normal cells
//     Seq<topology::point3d<double> > onTheFaces;
//     Seq<topology::vertex<topology::point<double> > * > vOnTheFaces;
    
//     convertPointsToVertices(rootsOnTheEdges,onTheFaces,insideTheCube,
// 			    vRoots,vOnTheFaces,vInsideTheCube);

//     buildMeshInSingularCell(scn,cell,vRoots,vInsideTheCube);

//     return weShouldMesh;
//   }
}

// This function takes two 2D-vectors, and outputs true
// iff they have a similar direction. What similar
// means is determined by the variable little
template <class C>
bool doTheyHaveASimilarDirection(topology::point<C> a,topology::point<C> b)
{
  bool result;
  const double little = 0.4;
  C scalarProduct,moduleasq,modulebsq;

  if (a.size() == 2 && b.size() == 2) {
    scalarProduct = a[0]*b[0] + a[1]*b[1];
    
    moduleasq = a[0]*a[0] + a[1]*a[1];
    modulebsq = b[0]*b[0] + b[1]*b[1];
    
    result = scalarProduct*scalarProduct > moduleasq * modulebsq * (1 - little);
  } else if (a.size() == 3 && b.size() == 3) {
    scalarProduct = a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
    
    moduleasq = a[0]*a[0] + a[1]*a[1] + a[2]*a[2];
    modulebsq = b[0]*b[0] + b[1]*b[1] + b[2]*b[2];
    
    result = scalarProduct*scalarProduct > moduleasq * modulebsq * (1 - little);
  }
  return result;
}

template <class MP,class REP3D,class C>
void buildMeshInSingularCell(scene<MP,REP3D> & scn,
			     const cell3d<MP,REP3D> & cell,
			     const Seq<Seq<topology::vertex<C> * > > & roots,
			     topology::vertex<C> * insideTheCube)
{
  unsigned numberOfRoots = 0;
  
  // We insert the vertex inside the cube
  insertVertexInList(scn, insideTheCube);

  // We insert the vertices from the edges
  for (unsigned edge = 0;edge < roots.size();edge++) {
    numberOfRoots += roots.size();
    for (unsigned root = 0;root < roots[edge].size();root++)
      insertVertexInList(scn,roots[edge][root]);
  }

  for (unsigned face = 1;face <= 6;face++) {
    std::vector<unsigned> v;
    edgesOfAFace(face,v);
    if (roots[v[0]].size() + roots[v[1]].size() + roots[v[2]].size() + 
	roots[v[3]].size() >= 2) {
      if (isARegularFaceOnThe1stDirection(cell.bezier,face)) {
	meshRegularFace1stDirection(scn,roots,insideTheCube,face);
      } else if (isARegularFaceOnThe2ndDirection(cell.bezier,face)) {
	meshRegularFace2ndDirection(scn,roots,insideTheCube,face);
      } else {
	meshFaceOnASingularCell(scn,cell,roots,insideTheCube,face);
      }
      
    }
  }
}

template <class MP,class REP3D,class C>
void meshFaceOnASingularCell(scene<MP,REP3D> & scn,
			     const cell3d<MP,REP3D> & cell,
			     const Seq<Seq<topology::vertex<C> * > > & vRoots,
			     topology::vertex<C> * insideTheCube,
			     unsigned face)
{
  std::vector<unsigned> edges;
  edgesOfAFace(face,edges);

  int initialRoot,firstRoot,secondRoot;
  unsigned initialEdge,firstEdge,secondEdge;

  initialEdge = 0;
  initialRoot = -1;

  // We initialize the initial root
  nextRoot(vRoots,initialEdge,initialRoot,face);

  secondRoot = initialRoot;
  secondEdge = initialEdge;

  do {

  firstRoot = secondRoot;
  firstEdge = secondEdge;

  nextRoot(vRoots,secondEdge,secondRoot,face);

  // Before actually drawing the triangle, we have to check its normal
  // Since we want triangles with the same orientation as the surface
  // we will check the sign of the dot product between the normal to the
  // and the normal to the triangle. For the normal to the surface, we'll
  // just take the normal to one of the roots on the edges.
  C normalToSurface,normalToTriangle;
  normalToSurface.resize(3);
  normalToTriangle.resize(3);
  
  Seq<Seq<typename MP::coeff_t> > roots;
  roots.resize(12);
  for (unsigned i = 1;i <= 12;i++)
    computeRootsOnEdge(cell,roots,i);
  
  Seq<Seq<Seq<typename MP::coeff_t> > > evalDerivatives;
  computeAndEvaluateDerivativesAtRoots(cell,roots,evalDerivatives);
  
  normalToSurface[0] = evalDerivatives[0][edges[firstEdge]][firstRoot];
  normalToSurface[1] = evalDerivatives[1][edges[firstEdge]][firstRoot];
  normalToSurface[2] = evalDerivatives[2][edges[firstEdge]][firstRoot];
  
  // These two vectors represent two sides of the triangle
  C side1,side2;
  side1.resize(3);
  side2.resize(3);
  side1[0] = vRoots[edges[secondEdge]][secondRoot]->point[0] - 
    vRoots[edges[firstEdge]][firstRoot]->point[0];
  side1[1] = vRoots[edges[secondEdge]][secondRoot]->point[1] - 
    vRoots[edges[firstEdge]][firstRoot]->point[1];
  side1[2] = vRoots[edges[secondEdge]][secondRoot]->point[2] - 
    vRoots[edges[firstEdge]][firstRoot]->point[2];
  side2[0] = insideTheCube->point[0] - 
    vRoots[edges[firstEdge]][firstRoot]->point[0];
  side2[1] = insideTheCube->point[1] - 
    vRoots[edges[firstEdge]][firstRoot]->point[1];
  side2[2] = insideTheCube->point[2] - 
    vRoots[edges[firstEdge]][firstRoot]->point[2];

  // We get the normal to the triangle by means of a vector product
  normalToTriangle[0] = side1[1]*side2[2] - side1[2]*side2[1];
  normalToTriangle[1] = side1[0]*side2[2] - side1[2]*side2[0];
  normalToTriangle[2] = side1[0]*side2[1] - side1[1]*side2[0];

  // Now, we compute the scalar product of both normals
  typename MP::coeff_t scalarProduct;
  scalarProduct = normalToSurface[0]*normalToTriangle[0] + 
    normalToSurface[1]*normalToTriangle[1] +
    normalToSurface[2]*normalToTriangle[2];
  
  if (scalarProduct > 0) {
    makeTriangle(scn,vRoots[edges[firstEdge]][firstRoot],
		 vRoots[edges[secondEdge]][secondRoot],insideTheCube);
  } else {
    makeTriangle(scn,insideTheCube,vRoots[edges[secondEdge]][secondRoot],
		 vRoots[edges[firstEdge]][firstRoot]);
  }
  
   } while (secondEdge != initialEdge || secondRoot != initialRoot);
}

template <class MP,class REP3D,class C>
void buildMeshUsingHermiteData(scene<MP,REP3D> & scn,
			       const Seq<Seq<topology::vertex<C> * > > & roots,
			       const Seq<topology::vertex<C> * > & onTheFaces,
			       topology::vertex<C> * insideTheCube)
{
  // We insert the vertex inside the cube
  insertVertexInList(scn, insideTheCube);

  // Now we process at the same time the vertices in the faces and on the
  // edges
  bool alreadyProcessed[12];
  // Initialization of the boolean array of processed edges
  for (unsigned i = 0;i < 12;i++)
    alreadyProcessed[i] = false;
  
  std::vector<unsigned> edges;
  
  for (unsigned face = 1;face <= 6;face ++) {
    if (isAnInterestingFace(roots,face)) {

      // We insert the vertex on the face
      insertVertexInList(scn,onTheFaces[face - 1]);

      // And the vertices in the edges of that face which have not been
      // processed yet
      edgesOfAFace(face,edges);
      for (unsigned edge = 0;edge < 4;edge++) {
	if (!alreadyProcessed[edges[edge]]) {
	  for (unsigned i = 0;i < roots[edges[edge]].size();i++)
	    insertVertexInList(scn,roots[edges[edge]][i]);
	  alreadyProcessed[edges[edge]] = true;
	}
      }

      // We add all the corresponding triangles...
      for (unsigned edge = 0;edge < 4;edge++)
	for (unsigned i = 0;i < roots[edges[edge]].size();i++)
	  makeTriangle(scn,insideTheCube,onTheFaces[face - 1],
		       roots[edges[edge]][i]);
    }
  }
}

template <class MP,class REP3D>
bool processCell(scene<MP,REP3D> & scn,const cell3d<MP,REP3D> & cell)
{
  typedef typename MP::coeff_t C;
  bool weShouldMesh;

  Seq<Seq<topology::point<C> > > rootsOnTheEdges;
  Seq<topology::point<C> > onTheFaces;
  topology::point<C> insideTheCube(3);

  weShouldMesh = shouldWeMesh(cell,rootsOnTheEdges,onTheFaces,insideTheCube);

  if (weShouldMesh) {

    Seq<Seq<topology::vertex<topology::point<C> > * > > vRoots;
    Seq<topology::vertex<topology::point<C> > * > vOnTheFaces;
    topology::vertex<topology::point<C> > * vInsideTheCube;
    
    convertPointsToVertices(rootsOnTheEdges,onTheFaces,insideTheCube,
			    vRoots,vOnTheFaces,vInsideTheCube);
    
    buildMeshUsingHermiteData(scn,vRoots,vOnTheFaces,vInsideTheCube);
  }

//   if (weShouldMesh)
//     std::cout << "We are processing a singular cell now..." << std::endl;


  return weShouldMesh;
}
  
template <class C,class MP,class REP3D>
bool liesInsideTheCell(const cell3d<MP,REP3D> & cell,topology::point<C> p)
{
  return (p[0] >= cell.box.x0 && p[0] <= cell.box.x1) &&
    (p[1] >= cell.box.y0 && p[1] <= cell.box.y1) &&
    (p[2] >= cell.box.z0 && p[2] <= cell.box.z1);
}

template <class C,class MP,class REP3D>
void convert2DPointInto3DPoint(const cell3d<MP,REP3D> & cell,
			       const topology::point<C> & p2d,
			       topology::point<C> & p3d,
			       unsigned face)
{
  p3d.resize(3);
  switch (face) {
  case 1:
    p3d[0] = p2d[0];p3d[1] = p2d[1];p3d[2] = cell.box.z0;
    break;
  case 2:
    p3d[0] = p2d[0];p3d[1] = cell.box.y0;p3d[2] = p2d[1];
    break;
  case 3:
    p3d[0] = p2d[0];p3d[1] = p2d[1];p3d[2] = cell.box.z1;
    break;
  case 4:
    p3d[0] = cell.box.x1;p3d[1] = p2d[0];p3d[2] = p2d[1];
    break;
  case 5:
    p3d[0] = p2d[0];p3d[1] = cell.box.y1;p3d[2] = p2d[1];
    break;
  case 6:
    p3d[0] = cell.box.x0;p3d[1] = p2d[0];p3d[2] = p2d[1];
    break;
  }
}

template <class C,class MP,class REP3D>
void computeAndEvaluateDerivativesAtRoots(const cell3d<MP,REP3D> & cell,
					  const Seq<Seq<C> > & roots,
					  Seq<Seq<Seq<C> > > & evalDerivatives)
{
  linalg::rep3d<C> fx,fy,fz;
  
  fx.resize(cell.bezier.nbrow(),cell.bezier.nbcol(),cell.bezier.nbhgt());
  fy.resize(cell.bezier.nbrow(),cell.bezier.nbcol(),cell.bezier.nbhgt());
  fz.resize(cell.bezier.nbrow(),cell.bezier.nbcol(),cell.bezier.nbhgt());
  
  
  // We copy the tensor to calculate the derivatives
  for (unsigned i = 0;i < fx.nbrow();i++)
    for (unsigned j = 0;j < fx.nbcol();j++)
      for (unsigned k = 0;k < fx.nbhgt();k++) {
	fx(i,j,k) = cell.bezier(i,j,k);
	fy(i,j,k) = fx(i,j,k);
	fz(i,j,k) = fx(i,j,k);
      }
  
  BEZIER::diff(fx,0,fx.nbcol() - 1,fx.nbrow() - 1,fx.nbhgt() - 1);
  BEZIER::diff(fy,1,fy.nbcol() - 1,fy.nbrow() - 1,fy.nbhgt() - 1);
  BEZIER::diff(fz,2,fz.nbcol() - 1,fz.nbrow() - 1,fz.nbhgt() - 1);

  evalDerivatives.resize(3);
  evalDerivatives[0].resize(12);
  evalDerivatives[1].resize(12);
  evalDerivatives[2].resize(12);
  
  for (unsigned edge = 0;edge < 12;edge++) {
    Seq<C> sec(roots[edge].size());
    
    evalDerivatives[0][edge] = sec;
    evalDerivatives[1][edge] = sec;
    evalDerivatives[2][edge] = sec;

    
    for (unsigned i = 0;i < roots[edge].size();i++) {
      evalDerivatives[0][edge][i] = 
	evaluateTensorOnEdge(fx,edge + 1,roots[edge][i]);
      evalDerivatives[1][edge][i] = 
	evaluateTensorOnEdge(fy,edge + 1,roots[edge][i]);
      evalDerivatives[2][edge][i] =
	evaluateTensorOnEdge(fz,edge + 1,roots[edge][i]);
    }
  }  
}

template <class MP,class REP3D>
void convert2Dinto3DReversing(scene<MP,REP3D> & scn,
			      const 
			      topology::point_graph<typename MP::coeff_t> & pg,
			      const cell3d<MP,REP3D> & cell,
			      Seq<topology::vertex
			      <topology::point<typename MP::coeff_t> > * >
			      & vertices,
			      unsigned face)
{
//   std::cout << pg.Vertices << std::endl;
//   std::cout << cell.box.x0 << "," << cell.box.x1 
// 	    << "," << cell.box.y0 << "," << cell.box.y1 
// 	    << "," << cell.box.z0 << "," << cell.box.z1 << "," << std::endl;

  vertices.resize(pg.Vertices.size());
  for (int i = 0;i < vertices.size();i++) {
    switch(face) {
    case 1:
      vertices[i] = new topology::vertex<topology::point<typename MP::coeff_t> >
	(pg.Vertices[i][1],pg.Vertices[i][0],cell.box.z0);
      break;
    case 2:
      vertices[i] = new topology::vertex<topology::point<typename MP::coeff_t> >
	(pg.Vertices[i][1],cell.box.y0,pg.Vertices[i][0]);
      break;
    case 3:
      vertices[i] = new topology::vertex<topology::point<typename MP::coeff_t> >
	(pg.Vertices[i][1],pg.Vertices[i][0],cell.box.z1);
      break;
    case 4:
      vertices[i] = new topology::vertex<topology::point<typename MP::coeff_t> >
	(cell.box.x1,pg.Vertices[i][1],pg.Vertices[i][0]);
      break;
    case 5:
      vertices[i] = new topology::vertex<topology::point<typename MP::coeff_t> >
	(pg.Vertices[i][1],cell.box.y1,pg.Vertices[i][0]);
      break;
    case 6:
      vertices[i] = new topology::vertex<topology::point<typename MP::coeff_t> >
	(cell.box.x0,pg.Vertices[i][1],pg.Vertices[i][0]);
      break;
    }
    insertVertexInList(scn,vertices[i]);
  }
}

template <class MP,class REP3D>
void convert2Dinto3DReversing2(scene<MP,REP3D> & scn,
			      const 
			      topology::point_graph<typename MP::coeff_t> & pg,
			      const cell3d<MP,REP3D> & cell,
			      Seq<topology::vertex
			      <topology::point<typename MP::coeff_t> > * >
			      & vertices,
			      unsigned face)
{
//   std::cout << pg.Vertices << std::endl;
//   std::cout << cell.box.x0 << "," << cell.box.x1 
// 	    << "," << cell.box.y0 << "," << cell.box.y1 
// 	    << "," << cell.box.z0 << "," << cell.box.z1 << "," << std::endl;

  vertices.resize(pg.Vertices.size());
  for (int i = 0;i < vertices.size();i++) {
    switch(face) {
    case 1:
      vertices[i] = new topology::vertex<topology::point<typename MP::coeff_t> >
	(pg.Vertices[i][0],pg.Vertices[i][1],cell.box.z0);
      break;
    case 2:
      vertices[i] = new topology::vertex<topology::point<typename MP::coeff_t> >
	(pg.Vertices[i][0],cell.box.y0,pg.Vertices[i][1]);
      break;
    case 3:
      vertices[i] = new topology::vertex<topology::point<typename MP::coeff_t> >
	(pg.Vertices[i][0],pg.Vertices[i][1],cell.box.z1);
      break;
    case 4:
      vertices[i] = new topology::vertex<topology::point<typename MP::coeff_t> >
	(cell.box.x1,pg.Vertices[i][0],pg.Vertices[i][1]);
      break;
    case 5:
      vertices[i] = new topology::vertex<topology::point<typename MP::coeff_t> >
	(pg.Vertices[i][0],cell.box.y1,pg.Vertices[i][1]);
      break;
    case 6:
      vertices[i] = new topology::vertex<topology::point<typename MP::coeff_t> >
	(cell.box.x0,pg.Vertices[i][0],pg.Vertices[i][1]);
      break;
    }
    insertVertexInList(scn,vertices[i]);
  }
}

__END_NAMESPACE_SYNAPS

#endif // axel_mesh_hermite_H__
