/*********************************************************************
*      This file is part of the source code of AXEL library          *
*   Author(s): E. Tobis                                              *
*              B. Mourrain, GALAAD, INRIA                            *
**********************************************************************/

#ifndef axel_mesh_regularity_H__
#define axel_mesh_regularity_H__

#include <synaps/init.h>

#include <vector>
#include <synaps/base/Seq.h>
#include <synaps/topology/mesh/scene.h>
#include <synaps/topology/vertex.h>

__BEGIN_NAMESPACE_SYNAPS

template <class REP3D>
bool isARegularFaceOnThe1stDirection(const REP3D & cell,unsigned face)
{
  bool result;
  // We prepare the matrix that will hold the appropriate derivative
  REP3D derivative;
  derivative.resize(cell.nbrow(),cell.nbcol(),cell.nbhgt());
  for (int i = 0;i < cell.nbrow();i++)
    for (int j = 0;j < cell.nbcol();j++)
      for (int k = 0;k < cell.nbhgt();k++)
	derivative(i,j,k) = cell(i,j,k);
  
  switch (face) {
  case 1:
    // Face 1 means z=z0. The first variable is x
    diff(derivative,0,derivative.nbcol() - 1,
	 derivative.nbrow() - 1,derivative.nbhgt() - 1);
    
    result = varMaxByRowInFace_1(cell) <= 1 
      && varMaxByRowInFace_1(derivative) == 0
      && varMaxByColInFace_1(derivative) == 0;
    break;
  case 2:
    // Face 2 means y = y0. The first variable is x
    diff(derivative,0,derivative.nbcol() - 1,
	 derivative.nbrow() - 1,derivative.nbhgt() - 1);
    
    result = varMaxByHgtInFace_2(cell) <= 1
      && varMaxByHgtInFace_2(derivative) == 0
      && varMaxByColInFace_2(derivative) == 0;
    break;
  case 3:
    // Face 3 means z = z1. The first variable is x
    diff(derivative,0,derivative.nbcol() - 1,
	 derivative.nbrow() - 1,derivative.nbhgt() - 1);

    result = varMaxByRowInFace_3(cell) <= 1
      && varMaxByRowInFace_3(derivative) == 0
      && varMaxByColInFace_3(derivative) == 0;
    break;
  case 4:
    // Face 4 means x = x1. The first variable is y
    diff(derivative,1,derivative.nbcol() - 1,
	 derivative.nbrow() - 1,derivative.nbhgt() - 1);
    
    result = varMaxByHgtInFace_4(cell) <= 1
      && varMaxByHgtInFace_4(derivative) == 0
      && varMaxByRowInFace_4(derivative) == 0;
    break;
  case 5:
    // Face 5 means y = y1. The first variable is x
    diff(derivative,0,derivative.nbcol() - 1,
	 derivative.nbrow() - 1,derivative.nbhgt() - 1);

    result = varMaxByHgtInFace_5(cell) <= 1
      && varMaxByHgtInFace_5(derivative) == 0
      && varMaxByColInFace_5(derivative) == 0;
    break;
  case 6:
    // Face 6 means x = x0. The first variable is y
    // Face 4 means x = x1. The first variable is y
    diff(derivative,1,derivative.nbcol() - 1,
	 derivative.nbrow() - 1,derivative.nbhgt() - 1);

    result = varMaxByHgtInFace_6(cell) <= 1
      && varMaxByRowInFace_6(derivative) == 0
      && varMaxByHgtInFace_6(derivative) == 0;
    break;
  }
  return result;
}

template <class REP3D>
bool isARegularFaceOnThe2ndDirection(const REP3D & cell,unsigned face)
{
  bool result;
  // We prepare the matrix that will hold the appropriate derivative
  REP3D derivative;
  derivative.resize(cell.nbrow(),cell.nbcol(),cell.nbhgt());
  for (int i = 0;i < cell.nbrow();i++)
    for (int j = 0;j < cell.nbcol();j++)
      for (int k = 0;k < cell.nbhgt();k++)
	derivative(i,j,k) = cell(i,j,k);

  switch (face) {
  case 1:
    // Face 1 means z=z0. The second variable is y
    diff(derivative,1,derivative.nbcol() - 1,
	 derivative.nbrow() - 1,derivative.nbhgt() - 1);
    
    result = varMaxByColInFace_1(cell) <= 1 
      && varMaxByRowInFace_1(derivative) == 0
      && varMaxByColInFace_1(derivative) == 0;
    break;
  case 2:
    // Face 2 means y = y0. The second variable is z
    diff(derivative,2,derivative.nbcol() - 1,
	 derivative.nbrow() - 1,derivative.nbhgt() - 1);
    
    result = varMaxByColInFace_2(cell) <= 1
      && varMaxByHgtInFace_2(derivative) == 0
      && varMaxByColInFace_2(derivative) == 0;
    break;
  case 3:
    // Face 3 means z = z1. The second variable is y
    diff(derivative,1,derivative.nbcol() - 1,
	 derivative.nbrow() - 1,derivative.nbhgt() - 1);

    result = varMaxByColInFace_3(cell) <= 1
      && varMaxByRowInFace_3(derivative) == 0
      && varMaxByColInFace_3(derivative) == 0;
    break;
  case 4:
    // Face 4 means x = x1. The second variable is z
    diff(derivative,2,derivative.nbcol() - 1,
	 derivative.nbrow() - 1,derivative.nbhgt() - 1);
    
    result = varMaxByRowInFace_4(cell) <= 1
      && varMaxByHgtInFace_4(derivative) == 0
      && varMaxByRowInFace_4(derivative) == 0;
    break;
  case 5:
    // Face 5 means y = y1. The second variable is z
    diff(derivative,2,derivative.nbcol() - 1,
	 derivative.nbrow() - 1,derivative.nbhgt() - 1);

    result = varMaxByColInFace_5(cell) <= 1
      && varMaxByHgtInFace_5(derivative) == 0
      && varMaxByColInFace_5(derivative) == 0;
    break;
  case 6:
    // Face 6 means x = x0. The second variable is z
    diff(derivative,2,derivative.nbcol() - 1,
	 derivative.nbrow() - 1,derivative.nbhgt() - 1);

    result = varMaxByRowInFace_6(cell) <= 1
      && varMaxByRowInFace_6(derivative) == 0
      && varMaxByHgtInFace_6(derivative) == 0;
    break;
  }
  return result;
}

template <class MP,class REP3D,class C>
void meshRegularFace1stDirection(scene<MP,REP3D> & scn,
				 const Seq<Seq<topology::vertex<C> * > > & roots,
				 topology::vertex<C> * insideTheCube,
				 unsigned face)
{
  unsigned currentUp = 0;
  unsigned currentDown = 0;
  int currentSorted = 0;
  Seq<topology::vertex<C> *> sortedPoints;
  unsigned up,down,left,right,coord;

  switch (face) {
  case 1:
    up = 9;down = 8;left = 0;right = 2;coord = 1;
    break;
  case 2:
    up = 4;down = 6;left = 0;right = 1;coord = 2;
    break;
  case 3:
    up = 11;down = 10;left = 1;right = 3;coord = 1;
    break;
  case 4:
    up = 6;down = 7;left = 9;right = 11;coord = 2;
    break;
  case 5:
    up = 5;down = 7;left = 2;right = 3;coord = 2;
    break;
  case 6:
    up = 4;down = 5;left = 8;right = 10;coord = 2;
    break;
  }

  sortedPoints.resize(roots[up].size() + roots[down].size() +
		      roots[left].size() + roots[right].size());

  if (roots[left].size() != 0) {
    sortedPoints[0] = roots[left][0];
    currentSorted++;
  }

  while (currentUp != roots[up].size() && currentDown != roots[down].size()) {
    if (roots[up][currentUp]->point[coord] 
	< roots[down][currentDown]->point[coord]) {
      sortedPoints[currentSorted] = roots[up][currentUp];
      currentUp++;
      currentSorted++;
    } else {
      sortedPoints[currentSorted] = roots[down][currentDown];
      currentDown++;
      currentSorted++;
    }
  }
  
  while (currentUp != roots[up].size()) {
    sortedPoints[currentSorted] = roots[up][currentUp];
    currentSorted++;
    currentUp++;
  }      
  
  while (currentDown != roots[down].size()) {
    sortedPoints[currentSorted] = roots[down][currentDown];
    currentSorted++;
    currentDown++;
  }      
  
  if (roots[right].size() != 0)
    sortedPoints[currentSorted] = roots[right][0];
  
  currentSorted = 0;
  currentUp = 0;
  currentDown = 0;

  while (currentSorted < sortedPoints.size() - 1) {
    // We connect two consecutive vertices
    makeTriangle(scn,insideTheCube,sortedPoints[currentSorted],
		 sortedPoints[currentSorted + 1]);
    currentSorted += 2;
  }
}						 


template <class MP,class REP3D,class C>
void meshRegularFace2ndDirection(scene<MP,REP3D> & scn,
				 const Seq<Seq<topology::vertex<C> * > > & roots,
				 topology::vertex<C> * insideTheCube,
				 unsigned face)
{
  unsigned currentUp = 0;
  unsigned currentDown = 0;
  int currentSorted = 0;
  Seq<topology::vertex<C> *> sortedPoints;
  unsigned up,down,left,right,coord;

  switch (face) {
  case 1:
    up = 0;down = 2;left = 8;right = 9;coord = 0;
    break;
  case 2:
    up = 1;down = 0;left = 4;right = 6;coord = 0;
    break;
  case 3:
    up = 1;down = 3;left = 10;right = 11; coord = 0;
    break;
  case 4:
    up = 11;down = 10;left = 6;right = 7;coord = 1;
    break;
  case 5:
    up = 3;down = 3;left = 5;right = 7; coord = 0;
    break;
  case 6:
    up = 10;down = 8;left = 4;right = 5; coord = 0;
    break;
  }

  sortedPoints.resize(roots[up].size() + roots[down].size() +
		      roots[left].size() + roots[right].size());

  if (roots[left].size() != 0) {
    sortedPoints[0] = roots[left][0];
    currentSorted++;
  }

  while (currentUp != roots[up].size() && currentDown != roots[down].size()) {
    if (roots[up][currentUp]->point[coord] 
	< roots[down][currentDown]->point[coord]) {
      sortedPoints[currentSorted] = roots[up][currentUp];
      currentUp++;
      currentSorted++;
    } else {
      sortedPoints[currentSorted] = roots[down][currentDown];
      currentDown++;
      currentSorted++;
    }
  }
  
  while (currentUp != roots[up].size()) {
    sortedPoints[currentSorted] = roots[up][currentUp];
    currentSorted++;
    currentUp++;
  }      
  
  while (currentDown != roots[down].size()) {
    sortedPoints[currentSorted] = roots[down][currentDown];
    currentSorted++;
    currentDown++;
  }      
  
  if (roots[right].size() != 0)
    sortedPoints[currentSorted] = roots[right][0];
  
  currentSorted = 0;
  currentUp = 0;
  currentDown = 0;
  
  while (currentSorted < sortedPoints.size() - 1) {
    // We connect two consecutive vertices
    makeTriangle(scn,insideTheCube,sortedPoints[currentSorted],
		 sortedPoints[currentSorted + 1]);
    currentSorted += 2;
  }
}						 

__END_NAMESPACE_SYNAPS

#endif
