/*********************************************************************
*      This file is part of the source code of AXEL library          *
*   Author(s): L. Deschamps,                                         *
*              G. Gatellier                                          *
*              E. Tobis                                              *
*              B. Mourrain, GALAAD, INRIA                            *
**********************************************************************/
/** \file                                                           **/
#ifndef axel_mesh_subdivisor_H__
#define axel_mesh_subdivisor_H__
/********************************************************************/

#include <synaps/init.h>
#include <synaps/topology/mesh/bezier3d.h>
#include <synaps/topology/Octree.h>
#include <synaps/topology/mesh/scene.h>
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
// This parts were added by E. Tobis to incorporate his meshing of  //
// singular cells.
#include <synaps/topology/mesh/hermite.h>
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////



/********************************************************************/

__BEGIN_NAMESPACE_SYNAPS

template<class Cell>
struct subdivisor 
{

  typedef typename Cell::coeff_t  coeff_t;

  coeff_t epsMinRow, epsMinCol, epsMinHgt,
          epsMaxRow, epsMaxCol, epsMaxHgt;

  subdivisor(): 
    epsMinRow(pow(0.5, 2)),epsMinCol(pow(0.5, 2)), epsMinHgt(pow(0.5, 2)),
    epsMaxRow(pow(0.5, 7)),epsMaxCol(pow(0.5, 7)), epsMaxHgt(pow(0.5, 7)){}
  
  subdivisor(const coeff_t& min, const coeff_t& max): 
    epsMinRow(min),epsMinCol(min), epsMinHgt(min),
    epsMaxRow(max),epsMaxCol(max), epsMaxHgt(max){}
  
  bool firstCondition  (const Cell &bz3d ) const;
  bool secondCondition (const Cell &bz3d ) const;
  int  testForSubdivision(const Cell& cl )  const;
  void split (Cell &bz3d,
	      OCTREE::Node< Cell > * currentNode, 
	      std::list< OCTREE::Node<Cell >* > &listNodebezier3d ) const;

};

/** First condition tested in the adaptive subdivision.  
    Turn over true if the maximum threshold has been reached,
    forgery if not. **/
template< class Cell >
bool subdivisor<Cell>::firstCondition(const Cell &bz3d ) const 
{
  bool minRow, minCol, minHgt;
  
  // Verification sur la taille de la boite
  checkbox3d( bz3d.box, minRow, minCol, minHgt, epsMaxRow, 
	      epsMaxCol, epsMaxHgt);
  
  // Si c'est une courbe ou une droite ...
  if ( bz3d.bezier.nbrow() <= 1 ) minRow = true;
  if ( bz3d.bezier.nbcol() <= 1 ) minCol = true;
  if ( bz3d.bezier.nbhgt() <= 1 ) minHgt = true;
  
  // Si l'intervalle maximum a ete atteind
  if ( minRow && minCol && minHgt ) 
    return true;
  
  return false;
}

/** Second condition tested in the adaptive subdivision.
    Turn over true if the minimum threshold has been reached 
    and if one can guarantee topology, forgery if not. **/
template< class Cell >
bool subdivisor<Cell>::secondCondition(const Cell &bz3d ) const
{
  bool minRow, minCol, minHgt;
 
  // Verification sur la taille de la boite
  checkbox3d( bz3d.box, minRow, minCol, minHgt, epsMinRow, 
	      epsMinCol, epsMinHgt);
  
  // Si c'est une courbe ou une droite ...
  if ( bz3d.bezier.nbrow() <= 1 ) minRow = true;
  if ( bz3d.bezier.nbcol() <= 1 ) minCol = true;
  if ( bz3d.bezier.nbhgt() <= 1 ) minHgt = true;
  
  // Si l'intervalle minimum a ete atteind.
  if ( minRow && minCol && minHgt ) {
    // Si la topologie est garantie.
    if  ( testForSubdivision( bz3d )>0 )
      return true;
  }
  return false;
}

/** Function being useful has the general subdivision. 
    It includes different predicats and returns false if the 
    matrix of control must be to subdivide. **/
template < class Cell >
int subdivisor<Cell>::testForSubdivision(const Cell& cl )  const
{
    bool test = false;
    int totalVar=0;
    int var1  = varInEdge1 ( cl.bezier );
    int var2  = varInEdge2 ( cl.bezier );
    int var3  = varInEdge3 ( cl.bezier );
    int var4  = varInEdge4 ( cl.bezier );
    int var5  = varInEdge5 ( cl.bezier );
    int var6  = varInEdge6 ( cl.bezier );
    int var7  = varInEdge7 ( cl.bezier );
    int var8  = varInEdge8 ( cl.bezier );
    int var9  = varInEdge9 ( cl.bezier );
    int var10 = varInEdge10( cl.bezier );
    int var11 = varInEdge11( cl.bezier );
    int var12 = varInEdge12( cl.bezier );

    int reg = 0;

    if( lessThanTwoVarByRow( cl.bezier ) && 
        (( varMaxByHgtInFace_2( cl.bezier ) <= 1 ) ||
	 ( varMaxByColInFace_2( cl.bezier ) <= 1 ) ) &&
	(( varMaxByHgtInFace_5( cl.bezier ) <= 1 ) ||
	 ( varMaxByColInFace_5( cl.bezier ) <= 1 ) )
	) reg = 1;
    else if( lessThanTwoVarByCol( cl.bezier ) && 
        (( varMaxByHgtInFace_4( cl.bezier ) <= 1 ) ||
	 ( varMaxByRowInFace_4( cl.bezier ) <= 1 ) ) &&
	(( varMaxByHgtInFace_6( cl.bezier ) <= 1 ) ||
	 ( varMaxByRowInFace_6( cl.bezier ) <= 1 ) )
	) reg = 2;
    else if( lessThanTwoVarByHgt( cl.bezier ) && 
        (( varMaxByColInFace_1( cl.bezier ) <= 1 ) ||
	 ( varMaxByRowInFace_1( cl.bezier ) <= 1 ) ) &&
	(( varMaxByColInFace_3( cl.bezier ) <= 1 ) ||
	 ( varMaxByRowInFace_3( cl.bezier ) <= 1 ) )
	) reg = 3;
    else 
       return -4;

    int n =0;
    if( var1  > 1 ) n++;
    if( var2  > 1 ) n++;
    if( var3  > 1 ) n++;
    if( var4  > 1 ) n++;
    if( var5  > 1 ) n++;
    if( var6  > 1 ) n++;
    if( var7  > 1 ) n++;
    if( var8  > 1 ) n++;
    if( var9  > 1 ) n++;
    if( var10 > 1 ) n++;
    if( var11 > 1 ) n++;
    if( var12 > 1 ) n++;

    if(n>0) return -reg;


    //Une face doit avoir au maximum deux racines.
    if ( var1 + var3 + var9  + var10 > 2 )
          return -reg;
    if ( var1 + var2 + var5  + var6  > 2 )
          return -reg;
    if ( var2 + var4 + var11 + var12 > 2 )
          return -reg;
    if ( var6 + var8 + var10 + var12 > 2 )
          return -reg;
    if ( var3 + var4 + var7  + var8  > 2 )
          return -reg;
    if ( var5 + var7 + var9  + var11 > 2 )
          return -reg;
    
    return reg;
}

/** Subdivide the matrix of control of them eight and insere the new
    ones stamps in the octree. **/
template < class Cell >
void subdivisor<Cell>::split( Cell &bz3d,
			      OCTREE::Node< Cell > * currentNode, 
			      std::list< OCTREE::Node<Cell>* > &listNodebezier3d ) const
{

  bool minRow, minCol, minHgt;

  // Est-ce que l'intervalle minimum n'a pas ete atteind ?
  checkbox3d( bz3d.box, minRow, minCol, minHgt, epsMaxRow, epsMaxCol, epsMaxHgt);
  
  // Si c'est une courbe ou une droite ...
  if ( bz3d.bezier.nbrow() <= 1 ) minRow = true;
  if ( bz3d.bezier.nbcol() <= 1 ) minCol = true;
  if ( bz3d.bezier.nbhgt() <= 1 ) minHgt = true;

  // Exploitation de 2 listes pour permettre un partitionnement intelligent
  std::vector < Cell > listA, listB;
  std::vector < Cell > *mustBeDivided, *divided, *swap;
  Cell b0, b1;
  
  mustBeDivided = &listA;
  divided = &listB;
  
  if ( !minRow || !minCol || !minHgt ) 
    mustBeDivided->push_back( bz3d );
 

  // Division suivant les lignes
  if ( !minRow) {
   
    for ( int i = 0 ; i < mustBeDivided->size() ; ++i ) {
      b0 = (*mustBeDivided)[i];
      b1 = b0;

      bezier::deCasteljau( b0.bezier, b1.bezier, ByRow() );
      b0.box.x1 = ( b0.box.x1 + b0.box.x0 ) / 2;
      b1.box.x0 = ( b1.box.x1 + b1.box.x0 ) / 2;
      
      divided->push_back( b0 );
      divided->push_back( b1 );
    }

    mustBeDivided->clear();
    swap = mustBeDivided;
    mustBeDivided = divided;
    divided = swap;
    
    if ( !( divided->empty() ) )
      std::cerr << " Error : in split  \n";
  }

  
  // Division suivant les colonnes
  if ( !minCol ) {
    
    for ( int i = 0 ; i < mustBeDivided->size() ; ++i ) {
      b0 = (*mustBeDivided)[i];
      b1 = b0;

      bezier::deCasteljau( b0.bezier, b1.bezier, ByCol() );
      b0.box.y1 = ( b0.box.y1 + b0.box.y0 ) / 2;
      b1.box.y0 = ( b1.box.y1 + b1.box.y0 ) / 2;      

      divided->push_back( b0 );
      divided->push_back( b1 );
    }

    mustBeDivided->clear();
    swap = mustBeDivided;
    mustBeDivided = divided;
    divided = swap;
    
    if ( !( divided->empty() ) )
      std::cerr << " Error : in split  \n";
  }

  // Division suivant la profondeur
  if ( !minHgt ) {
    
    for ( int i = 0 ; i < mustBeDivided->size() ; ++i ) {
      b0 = (*mustBeDivided)[i];
      b1 = b0;
   
      bezier::deCasteljau( b0.bezier, b1.bezier, ByHgt() );
      b0.box.z1 = ( b0.box.z1 + b0.box.z0 ) / 2;
      b1.box.z0 = ( b1.box.z1 + b1.box.z0 ) / 2;

      divided->push_back( b0 );
      divided->push_back( b1 );
    }

    mustBeDivided->clear();
    swap = mustBeDivided;
    mustBeDivided = divided;
    divided = swap;
    
    if ( !( divided->empty() ) )
      std::cerr << " Error : in split  \n";
  }

  // Si ce n'est pas en 3D => probleme avec l'octree. 
  if ( mustBeDivided->size() != 8 ) {
    std::cerr << "Octree pas possible pour le moment :o(  \n";
    std::cout << " Erreur : Taille de la liste "  << mustBeDivided->size() << std::endl;
    exit(0);
  }

  // Construction de l'octree.
  if ( changeOfSign( (*mustBeDivided)[0].bezier ) ) {
    (*mustBeDivided)[0].setCubeCorners();
    OCTREE::Node< Cell >* node_1 = new  OCTREE::Node< Cell >(OCTREE::B_NW, currentNode, (*mustBeDivided)[0] );
    listNodebezier3d.push_back( node_1 );
  }

  if ( changeOfSign( (*mustBeDivided)[1].bezier ) ) {
    (*mustBeDivided)[1].setCubeCorners();
    OCTREE::Node< Cell >* node_2 = new  OCTREE::Node< Cell >(OCTREE::F_NW, currentNode, (*mustBeDivided)[1] );
    listNodebezier3d.push_back(node_2);
  }
  
  if ( changeOfSign( (*mustBeDivided)[2].bezier ) ) {
    (*mustBeDivided)[2].setCubeCorners();
    OCTREE::Node< Cell >* node_3 = new  OCTREE::Node< Cell >(OCTREE::B_SW, currentNode, (*mustBeDivided)[2] );
    listNodebezier3d.push_back(node_3);
  }
  
  if ( changeOfSign( (*mustBeDivided)[3].bezier ) ) {
    (*mustBeDivided)[3].setCubeCorners();
    OCTREE::Node< Cell >* node_4 = new  OCTREE::Node< Cell >(OCTREE::F_SW, currentNode, (*mustBeDivided)[3] );
    listNodebezier3d.push_back(node_4);
  }
  
  if ( changeOfSign( (*mustBeDivided)[4].bezier ) ) {
    (*mustBeDivided)[4].setCubeCorners();
    OCTREE::Node< Cell >* node_5 = new  OCTREE::Node< Cell >(OCTREE::B_NE, currentNode, (*mustBeDivided)[4] );
    listNodebezier3d.push_back(node_5);
  }
  
  if ( changeOfSign( (*mustBeDivided)[5].bezier ) ) {
   (*mustBeDivided)[5].setCubeCorners();
   OCTREE::Node< Cell >* node_6 = new  OCTREE::Node< Cell >(OCTREE::F_NE, currentNode, (*mustBeDivided)[5] );
    listNodebezier3d.push_back(node_6);
  }
  
  if ( changeOfSign( (*mustBeDivided)[6].bezier ) ) {
    (*mustBeDivided)[6].setCubeCorners();
    OCTREE::Node< Cell >* node_7 = new  OCTREE::Node< Cell >(OCTREE::B_SE, currentNode, (*mustBeDivided)[6] );
    listNodebezier3d.push_back(node_7);
  }
  
  if ( changeOfSign( (*mustBeDivided)[7].bezier ) ) {
    (*mustBeDivided)[7].setCubeCorners();
    OCTREE::Node< Cell >* node_8 = new  OCTREE::Node< Cell >(OCTREE::F_SE, currentNode, (*mustBeDivided)[7] );
    listNodebezier3d.push_back(node_8);
  }
}

//----------------------------------------------------------------------
/** Successive adaptive subdivision of the matrix 3d of control. **/
template <class Cell >
void adaptiveSplit(scene<typename Cell::polynomial_t, typename Cell::rep_t>& scn,
		   Cell &bz3d , const subdivisor<Cell> & mthd) 
{
  typedef typename Cell::polynomial_t polynomial_t;
  typedef typename Cell::coeff_t      coeff_t;
  typedef typename Cell::rep_t        rep_t;

  std::list< OCTREE::Node< Cell >* > listNodebezier3d;

  // Creation de la racine de l'octree.
  (scn.octree).root->SetCell( bz3d );

  // Insertion de la racine dans la liste des noeuds.
  listNodebezier3d.push_back( scn.octree.GetRoot() );
    
  // Tant qu'il reste des matrices a traiter.
  while ( !listNodebezier3d.empty() ) {
  
    OCTREE::Node< Cell > * currentNode;
    currentNode = listNodebezier3d.front();

    Cell* b3d;
    b3d = &currentNode->GetCell();

    topology::box3d< coeff_t >  *box;
    bezier::rep3d< polynomial_t, rep_t > *bezier;
  
    box = &( b3d->box );
    bezier = &( b3d->bezier );
  
    // Cas ou la surface est coplanaire a la cellule ou si
    // la cellule doit etre etudiee plus en detail.
    if  ( isCoplanar( *bezier ) || changeOfSign( *bezier ) ) {

	if ( mthd.firstCondition( *b3d )) {
	  insert(scn, currentNode );
	}
	else {
	  if ( mthd.secondCondition( *b3d ) ) {
	    insert(scn, currentNode);
	  }
	  else {

	      mthd.split( *b3d, currentNode, listNodebezier3d );
	  }
	}
    }
    
    // Destruction du noeud en cours dans la liste.
    // Ce noeud n'est plus util, il a ete soit subdiviser, soit inserer
    // dans la liste des noeuds.
    listNodebezier3d.pop_front();
  }
 
}
//----------------------------------------------------------------------
// Update the cube structure.
template<class Cell>
void process(topology::cube3d<typename Cell::polynomial_t,typename Cell::rep_t>* cube,
	     Cell* b3d,OCTREE::Node< Cell >* currentNode,
	     scene<typename Cell::polynomial_t, typename Cell::rep_t>& scn)
{
#define boulegue(X)\
  if  ( !cube->edge_##X##_isProcessed ) {			\
    if ( cube->rootOnEdge_##X != NULL ) std::cerr << "Erreur : buildMesh,  != NULL\n"; \
    if ( varInEdge##X( b3d->bezier ) == 1 ) {\
      b3d->computeRootOnEdge_##X(scn);\
      b3d->propagationOfEdge_##X( currentNode );\
    }\
    cube->edge_##X##_isProcessed = true;\
  }

  boulegue(1);
  boulegue(2);
  boulegue(3);
  boulegue(4);
  boulegue(5);
  boulegue(6);
  boulegue(7);
  boulegue(8);
  boulegue(9);
  boulegue(10);
  boulegue(11);
  boulegue(12);

}

//----------------------------------------------------------------------
/** Procedure which builds the mesh. **/
template < class Cell >
static void buildMesh( scene<typename Cell::polynomial_t, typename Cell::rep_t> & scn,
		       const Cell &b, const subdivisor<Cell> &mthd ) 
{

  typedef typename Cell::coeff_t      coeff_t;
  typedef typename Cell::polynomial_t polynomial_t;
  typedef typename Cell::rep_t        rep_t;

  // Pour chacune des matrices-3d solutions.
  for ( int i = scn.listOfNode.size()-1 ; i >= 0 ; i-- ) {
    
    OCTREE::Node< Cell > * currentNode;
    currentNode = scn.listOfNode[i];

    Cell *b3d;
    b3d = &currentNode->GetCell();

    // Pointeur sur le cube courant.
    topology::cube3d< polynomial_t, rep_t > *cube;
    cube = &(b3d->cube);

    // Cellule posant problemes...
    int reg=mthd.testForSubdivision( *b3d );
    //    cout <<reg<<endl;
    if  ( reg<=0 ) 
      switch( reg )
	{
	case -1:
	if   (! processRegularCell(scn,*b3d,0))  buildNotStudiedCells( *b3d, scn.XRegCells );
	break;
	case -2:
	  if (! processRegularCell(scn,*b3d,1))  buildNotStudiedCells( *b3d, scn.YRegCells );
	break;
	case -3:
	  if (! processRegularCell(scn,*b3d,2))  buildNotStudiedCells( *b3d, scn.ZRegCells );
	  break;
	default:
	  buildNotStudiedCells( *b3d, scn.SingCells );
	  //MATRIX3D::printsign(cout,b3d->bezier);
	}

    process(cube,b3d,currentNode,scn);
    if ( reg > 0 )   
      {
	buildNotStudiedCells(*b3d, scn.listRegCells);
	
	b3d->buildMeshInCube(scn, currentNode );
      }
  }
}

/** Construction of the cells whose study entirely 
    has not been possible. **/
template < class Cell >
void buildNotStudiedCells( Cell &bz3d, 
			   std::vector< topology::point< typename Cell::coeff_t > > &listCells) 
{

  typedef typename Cell::coeff_t coeff_t;
  typedef topology::vertex< topology::point<coeff_t > > vertex_t;
  std::vector<vertex_t*> roots;
  bz3d.cube.getAllRoots(roots);
  if(roots.size()==0)
    {
      // Coordonnees de la boite.
      coeff_t x0 =  (bz3d.box.x0+bz3d.box.x1)/2;
      coeff_t y0 =  (bz3d.box.y0+bz3d.box.y1)/2; 
      coeff_t z0 =  (bz3d.box.z0+bz3d.box.z1)/2; 

      // Les 2 points permettant de reconstituer le cube.
      listCells.push_back( topology::point< coeff_t >( x0, y0, z0 ) );
    }
  else
    {
      coeff_t x0, y0,z0;
      for(unsigned i=0;i<roots.size();i++)
	{
	  x0=roots[i]->point[0];
	  y0=roots[i]->point[1];
	  z0=roots[i]->point[2];
	  listCells.push_back( topology::point< coeff_t >( x0, y0, z0 ) );
	}
    }

}

/**********************************************************************/

__END_NAMESPACE_SYNAPS

#endif // axel_mesh_subdivisor_H__
