/*********************************************************************
*      This file is part of the source code of AXEL library          *
*   Author(s): M. Celikbas                                           *
*              G. Gatellier                                          *
*              B. Mourrain, GALAAD, INRIA                            *
**********************************************************************/
#ifndef synaps_upol_bspline_BSControl_2_H
#define synaps_upol_bspline_BSControl_2_H
/*********************************************************************/
#include <synaps/init.h>
#include <synaps/upol/bspline/BSPLINE.h>
/*********************************************************************/
__BEGIN_NAMESPACE_SYNAPS 
//--------------------------------------------------------------------
//! The BSpline surface container
/*!
  Here, we have implemented various functions we need for
  the spline surfaces.
 */
class BSControl_2
{
public:

  /*! The default constructor. */
  BSControl_2() {}

  /*! The default destructor.*/
  ~BSControl_2() {}

  /*!
    \param b the bspline surface.
    \param x the node we want to insert.

    see in the implementation module the function 
    \ref surface_insert_in_x "insert_in_x" for more details.
  */
  template<class C, class N, class Pt>
  void insert_in_x(bspline::surface<C,N,Pt>& b, const C& x)
    {
      using namespace BSPLINE;
      insert_in_x(b.points_,b.nodes1_,b.degree1_,x);
    }
  //--------------------------------------------------------------
   /*!
    \param b the bspline surface.
    \param y the node we want to insert.

    see in the implementation module the function 
    \ref surface_insert_in_y "insert_in_y" for more details.
  */
  template<class C, class N, class Pt>
  void insert_in_y(bspline::surface<C,N,Pt>& b, const C& y)
    {
      using namespace BSPLINE;
      insert_in__y(b.points_,b.nodes2_,b.degree2_,y);	
    }
  //--------------------------------------------------------------
  /*!
    \param b the bspline surface.
    
    see in the implementation module the function 
    \ref surface_derive_in_x "derive_in_x" for more details.
  */
  template<class C, class N, class Pt>
  void derive_in_x(bspline::surface<C,N,Pt>& b)
    {
      using namespace BSPLINE;
      derive_in_x<C,N,Pt>(b.points_,b.nodes1_,b.degree1_--);
    }
  //--------------------------------------------------------------
  /*!
    \param b the bspline surface.
    
    see in the implementation module the function 
    \ref surface_derive_in_y "derive_in_y" for more details.
  */
  template<class C, class N, class Pt>
  void derive_in_y(bspline::surface<C,N,Pt>& b)
    {
      using namespace BSPLINE;
      derive_in_y<C,N,Pt>(b.points_,b.nodes2_,b.degree2_--);
    }
  //--------------------------------------------------------------
  /*!
    \param result.
    \param b the bspline surface we want to evaluate at (x1,x2).
    \param x1 a node.
    \param x2 a node.

    see in the implementation module the function 
    \ref surface_value "value" for more details.
  */
  template<class C, class N, class Pt, class R, class V> 
  void value(R& result, const bspline::surface<C,N,Pt>& b,
	     const C& x1, const C&x2)
    {
      using namespace BSPLINE;
      value(result,b.points_,b.nodes1_,b.nodes2_,
	      b.degree1_,b.degree2_,x1,x2);
    }
  //--------------------------------------------------------------
  /*!
    \param f the file where we store the polyline of control.
    \param b the bspline surface.
    \param inf1.
    \param sup1.
    \param inf2.
    \param sup2.


    see in the implementation module the function 
    \ref surface_draw_polyline "draw_polyline" for more details.
  */
  template<class C, class N, class Pt>
  void draw_polyline(std::ostream& f, const bspline::surface<C,N,Pt>& b,
		     int inf1=-1, int sup1=-1, int inf2=-1, int sup2=-1)
    {
      using namespace BSPLINE;
      BSPLINE::draw_polyline(f, b.points_, inf1, sup1, inf2, sup2);
    }	
  //--------------------------------------------------------------
  /*!
    \param f the file where we store the bspline surface.
    \param b the bspline surface.
    \param inf1.
    \param sup1.
    \param inf2.
    \param sup2.
 

    see in the implementation module the function 
    \ref surface_draw_surface "draw_surface" for more details.
  */
  template<class C, class N, class Pt, class V>
  void draw_surface(std::ostream& f, const bspline::surface<C,N,Pt>& b,
		    int inf1=-1, int sup1=-1, int inf2=-1, int sup2=-1)
    {	
      using namespace BSPLINE;
      BSPLINE::draw_surface<C,N,Pt,V>(f,b.points_,b.nodes1_,b.nodes2_,
				      b.degree1_,b.degree2_,
				      inf1, sup1, inf2, sup2);
    }
  //--------------------------------------------------------------
  /*!
    \param bl it represents the left side of the bspline surface after we subdivide.
    \param br it represents the right side of the bspline surface after we subdivide.
    \param b the bspline surface.
    \param x the node where we subdivide the bspline surface.

    see in the implementation module the function 
    \ref surface_subdivide_in_x "subdivide_in_x" for more details.
  */
  template<class C, class N, class Pt, class V>
  void subdivide_in_x(bspline::surface<C,N,Pt>& bl, bspline::surface<C,N,Pt>& br,
		      const bspline::surface<C,N,Pt>& b, const C& x)
    {
      using namespace BSPLINE;
      subdivide_in_x<C,N,Pt,V>(bl.points_,bl.nodes1_,br.points_,br.nodes1_,
			       b.points_,b.nodes1_,b.degree1_,x);
      bl.degree1_=br.degree1_=b.degree1_;
      bl.degree2_=br.degree2_=b.degree2_;
      bl.nodes2_ =br.nodes2_ =b.nodes2_;
    }
  //--------------------------------------------------------------
  /*!
    \param bl it represents the left side of the bspline surface after we subdivide.
    \param br it represents the right side of the bspline surface after we subdivide.
    \param b the bspline surface.
    \param y  the node where we subdivide the bspline surface.

    see in the implementation module the function 
    \ref surface_subdivide_in_y "subdivide_in_y" for more details.
  */
  template<class C, class N, class Pt, class V>
  void subdivide_in_y(bspline::surface<C,N,Pt>& bl, bspline::surface<C,N,Pt>& br,
		      const bspline::surface<C,N,Pt>& b, const C& y)
    {
      using namespace BSPLINE;
      subdivide_in_y<C,N,Pt,V>(bl.points_,bl.nodes2_,br.points_,br.nodes2_,
			       b.points_,b.nodes2_,b.degree2_,y);
      bl.degree1_=br.degree1_=b.degree1_;
      bl.degree2_=br.degree2_=b.degree2_;
      bl.nodes1_ =br.nodes1_ =b.nodes1_;
    }
  //--------------------------------------------------------------
  /*!
    \param bb the bspline surface after incrementation of the degree.
    \param b the bspline surface.

    see in the implementation module the function 
    \ref surface_incr_degree_in_x "incr_degree_in_x" for more details.
  */
  template<class C, class N, class Pt, class V>
  void incr_degree_in_x(bspline::surface<C,N,Pt>& bb,
			const bspline::surface<C,N,Pt>& b)
    {
      using namespace BSPLINE;
      bb.degree1_=b.degree1_+1;
      bb.degree2_=b.degree2_;
      bb.nodes2_ =b.nodes2_;
      incr_degree_in_x<C,N,Pt>(bb.points_,bb.nodes1_,
			       b.points_,b.nodes1_,b.degree1_);
    }   
  //--------------------------------------------------------------
  /*!
    \param bb the bspline surface after incrementation of the degree.
    \param b the bspline surface.

    see in the implementation module the function 
    \ref surface_incr_degree_in_y "incr_degree_in_y" for more details.
  */
  template<class C, class N, class Pt, class V>
  void incr_degree_in_y(bspline::surface<C,N,Pt>& bb,
			const bspline::surface<C,N,Pt>& b)
    {
      using namespace BSPLINE;
      bb.degree1_=b.degree1_;
      bb.degree2_=b.degree2_+1;
      bb.nodes1_ =b.nodes1_;
      incr_degree_in_y<C,N,Pt>(bb.points_,bb.nodes2_,
			       b.points_,b.nodes2_,b.degree2_);
    }   
  
};
//--------------------------------------------------------------------
__END_NAMESPACE_SYNAPS 
/*********************************************************************/
#endif //
