/*********************************************************************
*      This file is part of the source code of AXEL library          *
*   Author(s): M. Celikbas                                           *
*              G. Gatellier                                          *
*              B. Mourrain, GALAAD, INRIA                            *
**********************************************************************/
#ifndef synaps_upol_bspline_BSPLINE_H
#define synaps_upol_bspline_BSPLINE_H
/*********************************************************************/
#include <iostream>
#include <vector>
#include <synaps/init.h>
#include <synaps/linalg/VECTOR.m>
/*********************************************************************/
__BEGIN_NAMESPACE_SYNAPS 
//--------------------------------------------------------------------
//! Module for the spline functions.
/*!
  Here, we have implemented various functions which can be useful for
  the spline curves and the spline surfaces.
*/
namespace BSPLINE
{
  
  /* functions usefull for spline */

  //--------------------------------------------------------------
  /*!
    \param v it represents the set of nodes.
    \param x it represents one node.
    
    \return The interval of \c x in the vector \c v.
  */
  template<class C, class N>
  int find_index(const N& v, const C&x)
  {
    int n=v.size();
    if (x<v[0] || x>v[n-1])
      {
	std::cerr<<"find_index : "<<x<<" is NOT in array"<<std::endl;
	exit(1);
      }
    
    int k;
    if (x==v[n-1])
      {
	for (k=n-2; k>=0 ; k--)
	  if(x>v[k] && x<=v[k+1]) break;
      }
    else
      {
	for (k=0; k<n-1; k++)
	  if (x>=v[k] && x<v[k+1]) break;
      }
    return k;  
  }

  //--------------------------------------------------------------
  /*!
    \param v it represents the set of nodes.
    \param k it represents an index in \c v.
    
    \return The multiplicity of the \c k-th element of \c v.
  */
  template<class N>
  int multiplicity(const N& v, int k)
  {
    int n=v.size(), m=0;
    if (k<0 || k>=n) return 0;
    for (int i=k; i<n && v[i]==v[k]; i++) m++;
    for (int i=k-1; i>=0 && v[i]==v[k]; i--) m++;
    return m;
  }
  
  //--------------------------------------------------------------
  /*!
    \param v it represents the set of nodes.
    \param k it represents an index in \c v.
    \param x it represents a node in \c v.
    
    \return The multiplicity in \c v of \c x which has the interval 
    <t>[v(k),v(k+1)]</t>.
  */
  template<class C, class N>
  int multiplicity(const N& v, int k, const C& x)
  {
    int n=v.size(), m=0;
    if (k<0 || k>=n) return 0;
    for (int i=k; i<n && v[i]==x; i++) m++;
    for (int i=k-1; i>=0 && v[i]==x; i--) m++;
    return m;
  }

  /* spline curves functions*/

  //--------------------------------------------------------------
  /*!
    \anchor curve_insert
    \param points the set of points of the spline.
    \param nodes the set of nodes of the spline.
    \param d the degree of the spline.
    \param x a node.
    
    Insert 1 node \c x, without changing the shape and the degree of the spline.
    You cannot insert more than <t>d+1</t> times a node. 
  */
  template<class C, class N, class Pt>
  void insert(Pt& points, N& nodes, int d, const C& x)
  {
    int k=find_index(nodes,x), n=points.size();
    if (k<d || k>n)
      {
	std::cout<<"insert : the given node must be between"
		 <<"nodes[d] = "<<nodes[d]<<" and nodes[n] = "
		 <<nodes[n]<<std::endl;
	return;
      }
    if (multiplicity(nodes,k,x)==d+1)
      {
	std::cout<<"insert : "<<x<<" has already the multiplicity "
		 <<d+1<<std::endl;
	return;
      }
    
    Pt new_pts(n+1);
    int a=k-d;
    
    // update points	
    for (int j=0; j<=a; j++)
      new_pts[j] = points[j];
    
    for (int j=a+1; j<=k; j++)
      {
	C t = (x-nodes[j])/(nodes[j+d]-nodes[j]);
	new_pts[j] = t*points[j] + (1.0-t)*points[j-1];
      }
    
    for (int j=k+1; j<n+1; j++)
      new_pts[j] = points[j-1];
    
    points = new_pts;
    
    // update nodes
    int m = nodes.size();
    N new_nodes(m+1);	
    for (int i=0; i<=k; i++) new_nodes[i] = nodes[i];
    new_nodes[k+1] = x;
    for (int i=k+2; i<m+1; i++) new_nodes[i] = nodes[i-1];
    nodes = new_nodes;	
  }
  
  //--------------------------------------------------------------
  /*!
    \anchor curve_derive
    \param points the set of points of the spline.
    \param nodes the set of nodes of the spline.
    \param d the degree of the spline.
    
    Derive the spline curve in place.
  */
  template<class C, class N, class Pt>
  void derive(Pt& points, N& nodes, int d)
  {
    using VECTOR::init;
    int dim=points[0].size(), n=points.size(), m=nodes.size(); 
    Pt new_pts(n-1);
    N new_nodes(m-2);
    
    for (int j=0; j<n-1; j++)
      {
	if (nodes[j+d+1]-nodes[j+1]>0)
	  {
	    new_pts[j] = points[j+1]-points[j];
	    new_pts[j]*= d/(nodes[j+d+1]-nodes[j+1]);
	  }
	else
	  {
	    Rep(new_pts[j]).resize(dim);
	    init(new_pts[j], 0);
	  }
      }
    
    for (int i=0; i<m-2; i++) // we take off the 1st and the last nodes
      new_nodes[i] = nodes[i+1];
    
    points = new_pts;
    nodes  = new_nodes;
  }        
  
  //--------------------------------------------------------------
  /*!
    \anchor curve_value
    \param result.
    \param points the set of points of the spline.
    \param nodes the set of nodes of the spline.
    \param d the degree of the spline.
    \param k .
    \param x a node.
    
    Evaluate the spline at \c x; the result is put in the vector result.
    The point \c x to be evaluated must be <t>nodes[k]<=x<nodes[k+1]</t>.
  */
  template<class C, class N, class Pt, class R>
  static void value(R& result,
		    const Pt& points, const N& nodes, int d,
		    int k, const C& x)
  {	
    using VECTOR::init;
    // version which manage the nodes vectors of type 1 and 2
    int dim=points[0].size(), n=points.size(), m=nodes.size();
    
    // addl = number of nodes[0] to add in order to make it of multiplicity d+1
    int addl = d+1-multiplicity(nodes,0);
    
    int kk = k + addl;
    
    Pt M(d+1);
    int a=kk-d;
    
    for (int j=a; j<=kk; j++)
      if (j<addl || j>=addl+n)
	{
	  M[j-a].rep().resize(dim);
	  init(M[j-a], 0);
	}
      else			
	M[j-a] = points[j-addl];
    
    for (int j=1; j<=d; j++)
      for (int i=kk ; i>=a+j; i--)
            {
	      C u1, u2, t;
	      int q=i+d+1-j;
	      
	      u1 = i<addl ? nodes[0] : i>=addl+m ? nodes[m-1] : nodes[i-addl];
	      u2 = q<addl ? nodes[0] : q>=addl+m ? nodes[m-1] : nodes[q-addl];
	      t  = (x-u1)/(u2-u1);
	      
	      M[i-a] = (1-t)*M[i-1-a]+t*M[i-a];
            }
    
    result = M[d];
  }    
  
  //--------------------------------------------------------------
  /*!
    \param result.
    \param points the set of points of the spline.
    \param nodes the set of nodes of the spline.
    \param d the degree of the spline.
    \param x a node.
    
    The other parameter is computed by \c find_index() in the function.
  */
  template<class C, class N, class Pt, class R>
  void value(R& result,
	     const Pt& points, const N& nodes, int d,
	     const C& x)
  {
    int k=find_index(nodes,x);
    value(result,points,nodes,d,k,x);
  }    

  //--------------------------------------------------------------
  /*!
    \anchor curve_draw_polyline
    \param f the file where we store the polyline of control
    \param points the set of points of the spline
    \param inf
    \param sup 
    

    The parameters \c inf et \s sup represent the size of the set of 
    points we want to represent(in most of cases, \c inf=0 and \c sup=n-1 
    if n is the size of the set).
    If you let the parameters with the value -1, the function recuperates
    these values.

    We store the polyline of control in the stream \c f.
  */
  template<class Pt>
  void draw_polyline(std::ostream& f, const Pt& points,
		     int inf=-1, int sup=-1)	      
  {
    int dim=points[0].size(), n = points.size();
    
    if (inf==-1) inf = 0; 
    if (sup==-1) sup = n-1; 
    
    for (int j=inf; j<=sup; j++)
      {
	for (int i=0; i<dim; i++)
	  f << points[j][i] << '\t';
	f<<std::endl;
      }
  }
  
  //--------------------------------------------------------------
  /*!
    \anchor curve_draw_curve
    \param f the file where we store the spline.
    \param points the set of points of the spline.
    \param nodes the set of nodes of the spline.
    \param d the degree of the spline.
    \param inf.
    \param sup.
    

    The parameters \c inf et \s sup represent the size of the set of 
    points we want to represent(in most of cases, \c inf=0 and \c sup=n-1 
    if n is the size of the set).
    If you let the parameters with the value -1, the function recuperates
    these values.\n
    Plot the spline curve on <t>nodes[inf]..nodes[sup]</t> in the stream \c f.
  */
  template<class C, class N, class Pt>
  void draw_curve(std::ostream& f,
		  const Pt& points, const N& nodes, int d,
		  int inf=-1, int sup=-1)	      
  {
    int dim=points[0].size(), n=points.size(), NbPt=100;
    
    // by default we plot on nodes[d]..nodes[n]
    if (inf==-1) inf = d;
    if (sup==-1) sup = n; 
    
    for (int k=inf; k<sup; k++)
      {
	if (nodes[k]<nodes[k+1])
	  {
	    C delta = (nodes[k+1]-nodes[k])/(C)(NbPt-1);
	    for (C u=nodes[k]; u<=nodes[k+1]; u+=delta)
	      {
		N v(dim);
		value(v, points, nodes, d, k, u);
		for (int i=0; i<dim; i++)
		  f << v[i] << '\t';
		f<<std::endl;
	      }
	  }
      }
  }
   
  //--------------------------------------------------------------
  /*!
    \anchor curve_subdivide
    \param points1 the set of points of the first spline.
    \param nodes1 the set of nodes of the first spline.
    \param points2 the set of points of the second spline.
    \param nodes2 the set of nodes of the second spline.
    \param points the set of points of the spline.
    \param nodes the set of nodes of the spline.
    \param d the degree of the spline.
    \param x a node.
    
    Subdivise a spline in 2 splines of same degree \c d, at \c x.
  */
  template<class C, class N, class Pt>
  void subdivide(Pt& points1, N& nodes1, Pt& points2, N& nodes2,
		 const Pt& points, const N& nodes, int d,
		 const C& x)
  {
    //  size(points1) == k+1, size(nodes1) == k+d+2
    //  size(points2) == n-k+d, size(nodes2) == n+2d-k+1
    
    int k=find_index(nodes,x), n=points.size();
    
    if (k<d || k>n)
      {
	std::cout<<"subdivide : you must give a node between"
		 <<" nodes[d] = "<<nodes[d]<<" and nodes[n] = "
		 <<nodes[n]<<std::endl;
	return;
      }
    
    int a=k-d;
    
    // update nodes1
    Rep(nodes1).resize(k+d+2);
    for (int i=0; i<=k; i++)
      nodes1[i] = nodes[i];
    
    for (int i=k+1; i<=k+d+1; i++)
      nodes1[i] = x;       
    
    // update nodes2
    Rep(nodes2).resize(n+2*d-k+1);
    for (int i=0; i<=d; i++)
      nodes2[i] = x;
    
    for (int i=d+1; i<=n-k+2*d; i++)
      nodes2[i] = nodes[i+a];
    
    // update points1[0]..points1[a-1]
    Rep(points1).resize(k+1);
    for (int j=0; j<a; j++)	    
      points1[j] = points[j];	
    
    // update points2[d+1]..points2[n-a-1]
    Rep(points2).resize(n-a);
    for (int j=d+1; j<n-a; j++)	    
      points2[j] = points[j+a];
    
    // update points1[a]..points1[k] et points2[0]..points2[d]
    // ie compute the extreme points of the algorithm of DeBoor
    Pt M(d+1);       
    
    points1[a] = points[a];
    points2[d] = points[k];
    
    for (int j=a; j<=k; j++)
      M[j-a] = points[j];
    
    for (int j=1; j<=d; j++)
      {	    
	for (int i=k ; i>=a+j; i--)
	  {
	    C t  = (x-nodes[i])/(nodes[i-j+d+1]-nodes[i]);	       
	    M[i-a] = (1-t)*M[i-1-a]+t*M[i-a];	
	  }
	points1[a+j] = M[j];
	points2[d-j] = M[d];
      }
  }
   
  //--------------------------------------------------------------
  /*!
    \anchor curve_incr_degree
    \param new_pts the new set of points of the spline.
    \param new_nodes the new set of nodes of the spline.
    \param points the set of points of the spline.
    \param nodes the set of nodes of the spline.
    \param d the degree of the spline.
    
    Elevate the degree of the spline without changing its shape.
    The new spline has <t>points.size()+k-1</t> points and <t>nodes.size()+k</t> 
    nodes where \c k is the number of different nodes in nodes.
  */    
  template<class C, class N, class Pt>
  void incr_degree(Pt& new_pts, N& new_nodes,
		   const Pt& points, const N& nodes, int d)
  {
    using VECTOR::init
    int dim=points[0].size(), n=points.size(), m=nodes.size();
    
    // k = number of elements different of nodes
    int k=1;
    for (int i=0; i<m-1; i++)
      if (nodes[i]!=nodes[i+1]) k++;
	
    // new_nodes = nodes where the multiplicity of each element has been increased
    int mm=m+k;
    Rep(new_nodes).resize(mm);
    int jj=0;
    for(int i=0; i<m-1; i++)
      {
	new_nodes[jj++] = nodes[i];
	if(nodes[i] != nodes[i+1]) new_nodes[jj++] = nodes[i];
      }
    new_nodes[jj++] = new_nodes[jj++] = nodes[m-1];
    
    // new_pts 
    int nn=n+k-1;
    Rep(new_pts).resize(nn);
    
    int dd=d+1;
    Pt cc(dd), cw(dd);	
    for (int i=0; i<dd; i++)
      Rep(cw[i]).resize(dim);
    
    for (int j=1; j<=nn; j++)
      {
	int r=find_index(nodes,new_nodes[j]) + 1;	   
	
	for (int i=0; i<dd; i++)
	  {		
	    init(cw[i],0);
	    cc[i] = cw[i];
	  }
	
	for (int i=std::max(1,dd+1-r); i<=std::min(dd,n+dd-r); i++)	    
	  cc[i-1] = cw[i-1] = points[i+r-dd-1];	    
	
	for (int k=dd-1; k>=1; k--)
	  {
	    C y1=new_nodes[j+k-1], y2=new_nodes[j+k];
	    for (int i=std::min(dd,n+2*dd-k-r); i>=std::max(dd-k+1,dd-r+1); i--)
	      {
		C yi=nodes[r+i-dd-1], yik=nodes[r+i+k-dd-1], den=1.0/(yik-yi); 
		cc[i-1] = (y2-yi)*den*cc[i-1] + (yik-y2)*den*cc[i-2];
		cw[i-1] = (y1-yi)*den*cw[i-1] + (yik-y1)*den*cw[i-2] + cc[i-1];
	      }
	  }
	new_pts[j-1] = cw[dd-1]/dd;	    
      }	
  }
   
  //--------------------------------------------------------------
  /*!
    \anchor curve_basis_functions
    \param res.
    \param nodes.
    \param degree.
    \param x.


    Compute the <t>N_{i}^{d}(x), i=0..n-1 (where n=nodes.size()-degree-1)</t>.\n
    The result is put in the vector \c res.\n
    These values could be usefull in interpolation and approximation problems.
  */
  template<class C, class N>
  void basis_functions(N& res, const N& nodes, int degree, C x)
  {
    using VECTOR::init;
    int m=nodes.size(), n=m-degree-1;
    
    Rep(res).resize(n);
    init(res,0);
    
    if (x<nodes[0] || x>nodes[m-1])  return;
    if (x==nodes[0])
      {
	if (multiplicity(nodes,0)==degree+1) res[0]=1.0;	   
	return;
      }	
    if (x==nodes[m-1])
      {
	if (multiplicity(nodes,m-1)==degree+1) res[n-1]=1.0; 
	return;
      }
    
    int k = find_index(nodes,x);
    
    res[k] = 1.0;
    for (int d=1; d<=degree; d++)
      {
	res[k-d] = (nodes[k+1]-x)/(nodes[k+1]-nodes[k-d+1]) * res[k-d+1];
	
	C temp = (x-nodes[k-d+1])/(nodes[k+1]-nodes[k-d+1]);
	for (int i=k-d+1; i<=k-1; i++)
	  {	
	    res[i] = temp*res[i];
	    temp =  (x-nodes[i+1])/(nodes[i+1+d]-nodes[i+1]);
	    res[i] += (1.0-temp)*res[i+1];	    
	  }
	
	res[k] = (x-nodes[k])/(nodes[k+d]-nodes[k]) * res[k];
      }
  }

  /* spline surfaces functions */

  //--------------------------------------------------------------
  /*!
    \anchor surface_insert_in_x
    \param points the set of points of the spline.
    \param nodes the set of nodes of the spline.
    \param d the degree of the spline.
    \param x a node.
    
    Insert 1 node \c x with respect to the 1st variable,
    without changing the shape of the surface and the degree in this
    direction.
  */
  template<class C, class N, class Pt>
  void insert_in_x(Pt& points, N& nodes, int d, const C& x)
  {
    int k=find_index(nodes,x), n1=points.nbrow(), n2=points.nbcol();
    if (k<d || k>n1)
      {
	std::cout<<"insert_in_x : the given node must be between"
		 <<" nodes[d] = "<<nodes[d]<<" and nodes[n] = "
		 <<nodes[n1]<<std::endl;
	return;
      }
    if (Multiplicity(nodes,k,x)==d+1)
      {
	std::cout<<"insert_in_x : "<<x<<" has already the multiplicity "
		 <<d+1<<std::endl;
	return;
      }
    
    Pt new_pts(n1+1,n2);
    int a=k-d;
    
    // update points
    for (int j=0; j<n2; j++)
      {
	for (int i=0; i<=a; i++)
	  new_pts(i,j) = points(i,j);
	
	for (int i=a+1; i<=k; i++)
	  {
	    C t = (x-nodes[i])/(nodes[i+d]-nodes[i]);
	    new_pts(i,j) = t*points(i,j) + (1.0-t)*points(i-1,j);
	  }
	
	for (int i=k+1; i<n1+1; i++)
	  new_pts(i,j) = points(i-1,j);
      }
    points = new_pts;
    
    // update nodes
    int m = nodes.size();
    N new_nodes(m+1);	
    for (int i=0; i<=k; i++) new_nodes[i] = nodes[i];
    new_nodes[k+1] = x;
    for (int i=k+2; i<m+1; i++) new_nodes[i] = nodes[i-1];
    nodes = new_nodes;	
  }

  //--------------------------------------------------------------
  /*!
    \anchor surface_insert_in_y
    \param points the set of points of the spline.
    \param nodes the set of nodes of the spline.
    \param d the degree of the spline.
    \param y a node.
  
    Insert 1 node \c y with respect to the 2nd variable,
    without changing the shape of the surface and the degree in this
    direction.
  */
  template<class C, class N, class Pt>
  void insert_in_y(Pt& points, N& nodes, int d, const C& y)
  {
    int k=find_index(nodes,y), n1=points.nbrow(), n2=points.nbcol();
    if (k<d || k>n2)
      {
	std::cout<<"insert_in_y : the given node must be between"
		 <<" nodes[d] = "<<nodes[d]<<" and nodes[n] = "
		 <<nodes[n2]<<std::endl;
	return;
      }	
    if (multiplicity(nodes,k,y)==d+1)
      {
	std::cout<<"insert_in_y : "<<y<<" has already the multiplicity "
		 <<d+1<<std::endl;
	return;
      }
    
    Pt new_pts(n1,n2+1);
    int a=k-d;
    
    // update points
    for (int i=0; i<n1; i++)
      {
	for (int j=0; j<=a; j++)
	  new_pts(i,j) = points(i,j);
	
	for (int j=a+1; j<=k; j++)
	  {
	    C t = (x-nodes[j])/(nodes[j+d]-nodes[j]);
	    new_pts(i,j) = t*points(i,j) + (1.0-t)*points(i,j-1);
	  }
	
	for (int j=k+1; j<n2+1; j++)
	  new_pts(i,j) = points(i,j-1);
      }
    points = new_pts;
	
    // update nodes
    int m = nodes.size();
    N new_nodes(m+1);	
    for (int i=0; i<=k; i++) new_nodes[i] = nodes[i];
    new_nodes[k+1] = y;
    for (int i=k+2; i<m+1; i++) new_nodes[i] = nodes[i-1];
    nodes = new_nodes;	
  }
  
  //--------------------------------------------------------------
  /*!
    \anchor surface_derive_in_x
    \param points the set of points of the spline.
    \param nodes the set of nodes of the spline.
    \param d the degree of the spline.
    
    Derive in place with respect to the 1st variable.
  */
  template<class C, class N, class Pt>
  void derive_in_x(Pt& points, N& nodes, int d)
  {
    int dim=points(0,0).size(),n1=points.nbrow(), n2=points.nbcol(),
      m=nodes.size(); 
    Pt new_pts(n1-1,n2);
    N new_nodes(m-2);
    
    for (int j=0; j<n2; j++)
      {
	for (int i=0; i<n1-1; i++)
	  {
	    if (nodes[i+d+1]-nodes[i+1]>0)
	      {
		new_pts(i,j) = points(i+1,j)-points(i,j);
		new_pts(i,j)*= d/(nodes[i+d+1]-nodes[i+1]);
	      }
	    else
	      {
		Rep(new_pts(i,j)).resize(dim);
		init(new_pts(i,j), 0);
	      }
	  }
      }
    
    for (int i=0; i<m-2; i++)
      new_nodes[i] = nodes[i+1];
    
    points = new_pts;
    nodes  = new_nodes;
  }
  
  //--------------------------------------------------------------
  /*!
    \anchor surface_derive_in_y
    \param points the set of points of the spline.
    \param nodes the set of nodes of the spline.
    \param d the degree of the spline.
     
    Derive in place with respect to the 2nd variable.
  */
  template<class C, class N, class Pt>
  void derive_in_y(Pt& points, N& nodes, int d)
  {
    int dim=points(0,0).size(),n1=points.nbrow(), n2=points.nbcol(),
      m=nodes.size(); 
    Pt new_pts(n1,n2-1);
    N new_nodes(m-2);
    
    for (int i=0; i<n1; i++)
      {
	for (int j=0; j<n2-1; j++)
	  {
	    if (nodes[j+d+1]-nodes[j+1]>0)
	      {
		new_pts(i,j) = points(i,j+1)-points(i,j);
		new_pts(i,j)*= d/(nodes[j+d+1]-nodes[j+1]);
	      }
	    else
	      {
		Rep(new_pts(i,j)).resize(dim);
		init(new_pts(i,j), 0);
	      }
	  }
      }
    
    for (int i=0; i<m-2; i++)
      new_nodes[i] = nodes[i+1];
    
    points = new_pts;
    nodes  = new_nodes;
  }
  
  //--------------------------------------------------------------
  /*!
    \anchor surface_value
    \param result.
    \param points the set of points of the spline.
    \param nodes1 the first set of nodes of the spline.
    \param nodes2 the second set of nodes of the spline.
    \param d1 the degree of the spline for the first set of nodes.
    \param d2 the degree of the spline for the second set of nodes.
    \param k1.
    \param k2.
    \param x a node.
    \param y a node.
    
    Evaluate the surface at the points \c (x,y) that verifies :\n
    <t>nodes1[k1]<=x<nodes1[k1+1] and nodes2[k2]<=y<nodes2[k2+1]</t>.\n
    The result is put in the vector result.
  */
  template<class C, class N, class Pt, class R, class V>
  void value(R& result,
	     const Pt& points, const N& nodes1, const N& nodes2,
	     int d1, int d2, int k1, int k2, const C& x, const C& y)
  {
    int n1 = points.nbrow(), n2 = points.nbcol(),
      m1 = nodes1.size(),  m2 = nodes2.size();
    
    int dim=points(0,0).size();
    
    int addl1=d1+1-multiplicity(nodes1,0), addl2=d2+1-multiplicity(nodes2,0);
    
    int kk1 = k1 + addl1, kk2 = k2 + addl2;
    int a1 = kk1 - d1, a2 = kk2 - d2;

    V M1(d1+1);

    for (int j1=a1; j1<=kk1; j1++)
      if (j1<addl1 || j1>=addl1+n1)
	{
	  M1[j1-a1].rep().resize(dim);	  
	  init(M1[j1-a1], 0);
	}
      else
	{
	  int w = j1-addl1;
	  V M2(d2+1);
  
	  for (int j2=a2; j2<=kk2; j2++)
	    if (j2<addl2 || j2>=addl2+n2)
	      {
		M2[j2-a2].rep().resize(dim);
		init(M2[j2-a2], 0);
	      }
	    else
	      M2[j2-a2] = points(w,j2-addl2);
	  
	  for (int j2=1; j2<=d2; j2++)
	    for (int i=kk2 ; i>=a2+j2; i--)
	      {
		C u1, u2, t, u0=nodes2[0], um=nodes2[m2-1];
		int q=i+d2+1-j2;			
		
		u1 = i<addl2 ? u0 : i>=addl2+m2 ? um : nodes2[i-addl2];
		u2 = q<addl2 ? u0 : q>=addl2+m2 ? um : nodes2[q-addl2];
		t  = (y-u1)/(u2-u1);
		
		M2[i-a2] = (1-t)*M2[i-1-a2]+t*M2[i-a2];
	      }
	  
	  M1[j1-a1] = M2[d2];
	}
    
    for (int j=1; j<=d1; j++)
      for (int i=kk1 ; i>=a1+j; i--)
	{
	  C u1, u2, t, u0=nodes1[0], um=nodes1[m1-1];
	  int q=i+d1+1-j;
	  
	  u1 = i<addl1 ? u0 : i>=addl1+m1 ? um : nodes1[i-addl1];
	  u2 = q<addl1 ? u0 : q>=addl1+m1 ? um : nodes1[q-addl1];
	  t  = (x-u1)/(u2-u1);
	  
	  M1[i-a1] = (1-t)*M1[i-1-a1]+t*M1[i-a1];
	}	
    
    result = M1[d1];
  }

  //--------------------------------------------------------------
  /*!
    \param result.
    \param points the set of points of the spline.
    \param nodes1 the first set of nodes of the spline.
    \param nodes2 the second set of nodes of the spline.
    \param d1 the degree of the spline for the first set of nodes.
    \param d2 the degree of the spline for the second set of nodes.
    \param x a node.
    \param y a node.
    
    Evaluate the surface at the points (x,y) that verifies :\n
    <t>nodes1[k1]<=x<nodes1[k1+1] and nodes2[k2]<=y<nodes2[k2+1]</t>.
    Here, the parameters \c k1 and \c k2 are computed in the function
    with the function \c find_index().
    The result is put in the vector \c result.
  */
  template<class C, class N, class Pt, class R, class V>
  void value(R& result,
	     const Pt& points, const N& nodes1, const N& nodes2,
	     int d1, int d2, const C& x, const C& y)
  {
    int k1=find_index(nodes1,x), k2=find_index(nodes2,y);
    value<C,N,Pt,R,V>(result,points,nodes1,nodes2,d1,d2,k1,k2,x,y);
  }
   
  //--------------------------------------------------------------
  /*!
    \anchor surface_draw_polyline
    \param f the file where we store the polyline of control.
    \param points the set of points of the spline.
    \param inf1.
    \param sup1.
    \param inf2.
    \param sup2.
    

    The optional inputs \c inf1, \c sup1, \c inf2 and \c sup2 have default values
    \c d1, \c d2, \c points.nbrow() and \c points.nbcol() wich corresponds 
    to the standard working intervals.

    We store the polyline of control in the stream \c f.
    The format is the "VECT" format of geomview.
  */
  template<class Pt>
  void draw_polyline(std::ostream& f, const Pt& points,
		     int inf1=-1, int sup1=-1, int inf2=-1, int sup2=-1)
  {
    int n1=points.nbrow(), n2=points.nbcol();
    int dim=points(0,0).size();
    
    if (inf1==-1) inf1 = 0; if (sup1==-1) sup1 = n1-1;
    if (inf2==-1) inf2 = 0; if (sup2==-1) sup2 = n2-1;
    
    n1 = sup1-inf1+1; n2 = sup2-inf2+1;
    
    f<<"VECT\n";
    f<<n1+n2<<"  "<<2*n1*n2<<"  "<<0<<'\n'; // nbpoly and nbvectice 
    for (int i=1; i<=n2; i++) f<<n1<<"  ";  // the columns
    for (int i=1; i<=n1; i++) f<<n2<<"  ";  // the lines
    f<<'\n';
    for (int i=1; i<=n1+n2; i++) f<<0<<"  "; // colors
    f<<'\n';
    // each column
    for (int j=inf2; j<=sup2; j++)
      for (int i=inf1; i<=sup1; i++)
	{		
	  for (int k=0; k<dim; k++)
	    f << points(i,j)[k] << "  ";
	  f<<'\n';
	}
    // each line
    for (int i=inf1; i<=sup1; i++)
      for (int j=inf2; j<=sup2; j++)
	{		
	  for (int k=0; k<dim; k++)
	    f << points(i,j)[k] << "  ";
	  f<<'\n';
	}
    f<<std::endl;
  }

  //--------------------------------------------------------------
  /*!
    \anchor surface_draw_surface
    \param f the file where we store the the spline.
    \param points the set of points.
    \param nodes1 the first set of nodes (in the x-coordinate).
    \param nodes2 the second set of nodes (in the y-coordinate).
    \param d1 the first degree (in the x-coordinate).
    \param d2 the second degree (in the y-coordinate).
    \param inf1.
    \param sup1.
    \param inf2.
    \param sup2.


    Plot the surface on 
    <t>(nodes1[inf1]..nodes1[sup1]) * (nodes2[inf2]..nodes2[sup2])</t>.\n
    The optional inputs \c inf1, \c sup1, \c inf2 and \c sup2 have default 
    values \c d1, \c d2, \c points.nbrow() and \c points.nbcol() wich corresponds 
    to the standard working intervals.\n
    The output format is the geomview "MESH" format.
  */
  template<class C, class N, class Pt, class V>
  void draw_surface(std::ostream& f,
		    const Pt& points, const N& nodes1, const N& nodes2,
		    int d1, int d2,
		    int inf1=-1, int sup1=-1, int inf2=-1, int sup2=-1) 
  {   
    int n1=points.nbrow(), n2=points.nbcol(), NbPt1=100, NbPt2=100;
    int dim=points(0,0).size();
    
    if (inf1==-1) inf1 = d1; if (sup1==-1) sup1 = n1;
    if (inf2==-1) inf2 = d2; if (sup2==-1) sup2 = n2;
    
    C delta1 = (nodes1[sup1]-nodes1[inf1])/(C)(NbPt1-1);
    C delta2 = (nodes2[sup2]-nodes2[inf2])/(C)(NbPt2-1);	
    f<<"MESH\n";
    f<<NbPt1<<"  "<<NbPt2<<'\n'; // at first u next v	
    for (int j=0; j<NbPt2; j++)
      {
	C y=nodes2[inf2] + delta2*j;
	for (int i=0; i<NbPt1; i++) // u
	  {
	    C x=nodes1[inf1] + delta1*i;
	    N v(dim);
	    value<C,N,Pt,N,V>(v,points,nodes1,nodes2,d1,d2,x,y);
	    for (int i=0; i<dim; i++)
	      f << v[i] << '\t';
	  }
	f<<'\n';
      }
    f<<endl;	
  }
  
  //--------------------------------------------------------------
  /*!
    \anchor surface_subdivide_in_x
    \param points1 the set of points of the first spline.
    \param nodes1 the set of nodes of the first spline.
    \param points2 the set of points of the second spline.
    \param nodes2 the set of nodes of the second spline.
    \param Pt points the set of points of the spline.
    \param N nodes the set of nodes of the spline.
    \param d the degree of the spline.
    \param x a node.

    Subdivise a spline in 2 splines of same degree d, at x (1st variable).
  */
  template<class C, class N, class Pt, class V>
  void subdivide_in_x(Pt& points1, N& nodes1, Pt& points2, N& nodes2,
		      const Pt& points, const N& nodes, int d,
		      const C& x)
  {
    int k=find_index(nodes,x), n1=points.nbrow(), n2=points.nbcol();
    if (k<d || k>n1)
      {
	std::cout<<"subdivide : you must give a node between"
		 <<" nodes[d] = "<<nodes[d]<<" and nodes[n] = "
		 <<nodes[n1]<<std::endl;
	return;
      }
    
    int a=k-d;
    
    // update nodes1
    Rep(nodes1).resize(k+d+2);
    for (int i=0; i<=k; i++)
      nodes1[i] = nodes[i];
    
    for (int i=k+1; i<=k+d+1; i++)
      nodes1[i] = x;       
    
    // update nodes2
    Rep(nodes2).resize(n1+2*d-k+1);
    for (int i=0; i<=d; i++)
      nodes2[i] = x;
    
    for (int i=d+1; i<=n1-k+2*d; i++)
      nodes2[i] = nodes[i+a];
    
    // update points1(0:a-1,0)..points1(0:a-1,n2-1)
    Rep(points1).resize(k+1,n2);
    for (int j=0; j<n2; j++)
      for (int i=0; i<a; i++)
	points1(i,j) = points(i,j);	
    
    // update points2(d+1:n1-a-1,0)..points2(d+1:n1-a-1,n2-1)
    Rep(points2).resize(n1-a,n2);
    for (int j=0; j<n2; j++)
      for (int i=d+1; i<n1-a; i++)	    
	points2(i,j) = points(i+a,j);
    
    // update points1(a:k,0)..points1(a:k,n2-1)
    // and points2(0:d,0)..points2(0:d,n2-1)
    // ie compute the extreme points of the algorithm of DeBoor
    V M(d+1);       
    
    for (int j=0; j<n2; j++)
      {
	points1(a,j) = points(a,j);
	points2(d,j) = points(k,j);
      }
    
    for (int j=0; j<n2; j++)
      {
	for (int i=a; i<=k; i++)
	  M[i-a] = points(i,j);	
	
	for (int jj=1; jj<=d; jj++)
	  {	    
	    for (int i=k ; i>=a+jj; i--)
	      {
		C t  = (x-nodes[i])/(nodes[i-jj+d+1]-nodes[i]);	       
		M[i-a] = (1-t)*M[i-1-a]+t*M[i-a];	
	      }
	    points1(a+jj,j) = M[jj];
	    points2(d-jj,j) = M[d];
	  } 
      }	
  }

  //--------------------------------------------------------------
  /*!
    \anchor surface_subdivide_in_y
    \param points1 the set of points of the first spline.
    \param nodes1 the set of nodes of the first spline.
    \param points2 the set of points of the second spline.
    \param nodes2 the set of nodes of the second spline.
    \param points the set of points of the spline.
    \param nodes the set of nodes of the spline.
    \param d the degree of the spline.
    \param y  a node.
 
    Subdivise a spline in 2 splines of same degree \c d, at \c y (2nd variable).
  */
  template<class C, class N, class Pt, class V>
  void subdivide_in_y(Pt& points1, N& nodes1, Pt& points2, N& nodes2,
		      const Pt& points, const N& nodes, int d,
		      const C& y)
  {
    int k=find_index(nodes,y), n1=points.nbrow(), n2=points.nbcol();	
    if (k<d || k>n2)
      {
	std::cout<<"subdivide : you must give a node between"
		 <<" nodes[d] = "<<nodes[d]<<" and nodes[n] = "
		 <<nodes[n2]<<std::endl;
	return;
      }
    
    int a=k-d;
    
    // update nodes1
    Rep(nodes1).resize(k+d+2);
    for (int i=0; i<=k; i++)
      nodes1[i] = nodes[i];
    
    for (int i=k+1; i<=k+d+1; i++)
      nodes1[i] = y;       
    
    // update nodes2
    Rep(nodes2).resize(n2+2*d-k+1);
    for (int i=0; i<=d; i++)
      nodes2[i] = y;
    
    for (int i=d+1; i<=n2-k+2*d; i++)
      nodes2[i] = nodes[i+a];
    
    // update points1(0,0:a-1)..points1(n1-1,0:a-1)
    Rep(points1).resize(n1,k+1);
    for (int i=0; i<n1; i++)
      for (int j=0; j<a; j++)
	points1(i,j) = points(i,j);	
    
    // update points2(0,d+1:n2-a-1)..points2(n1-1,d+1:n2-a-1)
    Rep(points2).resize(n1,n2-a);
    for (int i=0; i<n1; i++)
      for (int j=d+1; j<n2-a; j++)	    
	points2(i,j) = points(i,j+a);
    
    // update points1(0,a:k)..points1(n1-1,a:k)
    // and points2(0,0:d)..points2(n1-1,0:d)
    // ie compute the extreme points of the algorithm of DeBoor
    V M(d+1);       
    
    for (int i=0; i<n1; i++)
      {
	points1(i,a) = points(i,a);
	points2(i,d) = points(i,k);
      }
    
    for (int i=0; i<n1; i++)
      {
	for (int j=a; j<=k; j++)
	  M[j-a] = points(i,j);	
	
	for (int j=1; j<=d; j++)
	  {	    
	    for (int ii=k ; ii>=a+j; ii--)
	      {
		C t  = (y-nodes[ii])/(nodes[ii-j+d+1]-nodes[ii]);	       
		M[ii-a] = (1-t)*M[ii-1-a]+t*M[ii-a];	
	      }
	    points1(i,a+j) = M[j];
	    points2(i,d-j) = M[d];
	  } 
      }	
  }
   
  //--------------------------------------------------------------
  /*!
    \anchor surface_incr_degree_in_x
    \param new_pts the new set of points of the spline.
    \param new_nodes the new set of nodes of the spline.
    \param points the set of points of the spline.
    \param nodes the set of nodes of the spline.
    \param d the degree of the spline.

    Elevate the degree of the spline without changing its shape
    with respect to the 1st variable.
  */
  template<class C, class N, class Pt, class V>
  void incr_degree_in_x(Pt& new_pts, N& new_nodes,
			const Pt& points, const N& nodes, int d)
  {
    int n1=points.nbrow(), n2=points.nbcol(), m=nodes.size();
    int dim=points(0,0).size();
    
    // k = number of elements different of nodes
    int k=1;
    for (int i=0; i<m-1; i++)
      if (nodes[i]!=nodes[i+1]) k++;
    
    // new_nodes = nodes where the multiplicity of each element has been increased
    int mm=m+k;
    Rep(new_nodes).resize(mm);
    int jc=0;
    for(int i=0; i<m-1; i++)
      {
	new_nodes[jc++] = nodes[i];
	if(nodes[i] != nodes[i+1]) new_nodes[jc++] = nodes[i];
      }
    new_nodes[jc++] = new_nodes[jc++] = nodes[m-1];

    // new_pts 
    int nn=n1+k-1;
    Rep(new_pts).resize(nn,n2);
    
    int dd=d+1;
    V cc(dd), cw(dd);
    
    for (int i=0; i<dd; i++)
      Rep(cw[i]).resize(dim);
    
    for (int jj=0; jj<n2; jj++)
      {
	for (int j=1; j<=nn; j++)
	  {
	    int r=find_index(nodes,new_nodes[j]) + 1;	   
	    
	    for (int i=0; i<dd; i++)
	      {
		init(cw[i],0);
		cc[i] = cw[i];
	      }
	    
	    for (int i=std::max(1,dd+1-r); i<=std::min(dd,n1+dd-r); i++)	    
	      cc[i-1] = cw[i-1] = points(i+r-dd-1,jj);	    
	    
	    for (int k=dd-1; k>=1; k--)
	      {
		C y1=new_nodes[j+k-1], y2=new_nodes[j+k];
		for (int i=std::min(dd,n1+2*dd-k-r); i>=std::max(dd-k+1,dd-r+1); i--)
		  {
		    C yi=nodes[r+i-dd-1], yik=nodes[r+i+k-dd-1], 
		      den=1.0/(yik-yi); 
		    cc[i-1] = (y2-yi)*den*cc[i-1] + (yik-y2)*den*cc[i-2];
		    cw[i-1] = (y1-yi)*den*cw[i-1] + (yik-y1)*den*cw[i-2] + cc[i-1];
		  } 
	      }
	    new_pts(j-1,jj) = cw[dd-1]/dd;
	  }
      }
  }

  //--------------------------------------------------------------
  /*!
    \anchor surface_incr_degree_in_y
    \param new_pts the new set of points of the spline.
    \param new_nodes the new set of nodes of the spline.
    \param points the set of points of the spline.
    \param nodes the set of nodes of the spline.
    \param d the degree of the spline.
    
    Elevate the degree of the spline without changing its shape
    with respect to the 2nd variable.
  */
  template<class C, class N, class Pt, class V>
  void incr_degree_in_y(Pt& new_pts, N& new_nodes,
			const Pt& points, const N& nodes, int d)
  {
    using VECTOR::init;
    int n1=points.nbrow(), n2=points.nbcol(), m=nodes.size();
    int dim=points(0,0).size();
    
    // k = number of elements different of nodes
    int k=1;
    for (int i=0; i<m-1; i++)
      if (nodes[i]!=nodes[i+1]) k++;
    
    // new_nodes = nodes where the multiplicity of each element has been increased
    int mm=m+k;
    Rep(new_nodes).resize(mm);
    int jc=0;
    for(int i=0; i<m-1; i++)
      {
	new_nodes[jc++] = nodes[i];
	if(nodes[i] != nodes[i+1]) new_nodes[jc++] = nodes[i];
      }
    new_nodes[jc++] = new_nodes[jc++] = nodes[m-1];
    
    // new_pts 
    int nn=n2+k-1;
    Rep(new_pts).resize(n1,nn);
    
    int dd=d+1;
    V cc(dd), cw(dd);
    
    for (int i=0; i<dd; i++)
      Rep(cw[i]).resize(dim);
    
    for (int jj=0; jj<n1; jj++)
      {
	for (int j=1; j<=nn; j++)
	  {
	    int r=find_index(nodes,new_nodes[j]) + 1;	   
	    
	    for (int i=0; i<dd; i++)
	      {
		init(cw[i],0);
		cc[i] = cw[i];
	      }
	    
	    for (int i=std::max(1,dd+1-r); i<=std::min(dd,n2+dd-r); i++)	    
	      cc[i-1] = cw[i-1] = points(jj,i+r-dd-1);	    
	    
	    for (int k=dd-1; k>=1; k--)
	      {
		C y1=new_nodes[j+k-1], y2=new_nodes[j+k];
		for (int i=std::min(dd,n2+2*dd-k-r); i>=std::max(dd-k+1,dd-r+1); i--)
		  {
		    C yi=nodes[r+i-dd-1], yik=nodes[r+i+k-dd-1], den=1.0/(yik-yi); 
		    cc[i-1] = (y2-yi)*den*cc[i-1] + (yik-y2)*den*cc[i-2];
		    cw[i-1] = (y1-yi)*den*cw[i-1] + (yik-y1)*den*cw[i-2] + cc[i-1];
		  } 
	      }
	    new_pts(jj,j-1) = cw[dd-1]/dd;
	  }
      }
  }

}
//--------------------------------------------------------------------
__END_NAMESPACE_SYNAPS 
/*********************************************************************/
#endif // synaps_upol_bspline_BSCONTROL_H

