// Copyright (c) 2003  INRIA Sophia-Antipolis (France) and
//                     Department of Informatics and Telecommunications
//                     University of Athens (Greece).
// All rights reserved.
//
// Authors : Elias P. TSIGARIDAS <et@di.uoa.gr>
//           Athanasios Kakargias <grad0460@di.uoa.gr>

// Partially supported by INRIA's project <CALAMATA>, a
// bilateral collaboration with National Kapodistrian University of
// Athens.
// Partially supported by the IST Programme of the EU as a Shared-cost
// RTD (FET Open) Project under Contract No  IST-2000-26473
// (ECG - Effective Computational Geometry for Curves and Surfaces)
//

#ifndef synaps_arithm_Algebraic_h
#define synaps_arithm_Algebraic_h

#include <synaps/init.h>
#include <synaps/arithm/gmp.h>
#include <synaps/usolve/Exclusion.h>
#include <synaps/base/Seq.h>
#include <map>

#include <synaps/usolve/algebraic/Algebraic_config.h>
#include <synaps/usolve/algebraic/root_of.h>
#include <synaps/usolve/algebraic/Comparison.h>
#include <synaps/usolve/algebraic/comparison/Compare_operators.h>
#include <synaps/usolve/algebraic/comparison/Compare_RT_operators.h>
#include <synaps/usolve/algebraic/comparison/Compare_FT_operators.h>

#include <synaps/usolve/algebraic/sign_fct.h>
#include <synaps/usolve/algebraic/MPoly_2_2.h>
#include <synaps/usolve/algebraic/sign_fct_2.h>
#include <synaps/usolve/algebraic/Method_name.h>
#include <synaps/usolve/Sturm.h>
#include <synaps/msolve/BSolver.h>
#include <synaps/usolve/algebraic/Construct_rootof.h>


#include <synaps/arithm/Interval.h>
#include <synaps/arithm/traits.h>
#include <synaps/upol/UPolDse.h>
#include <synaps/usolve/sturm/solve_sturm.h>

__BEGIN_NAMESPACE_SYNAPS


template < class NT >
struct Algebraic : public Small_degree<NT>
{
    typedef Algebraic          self_t;
};

//----------------------------------------------------------------------
template < class NT > inline
Seq< typename Algebraic<NT>::RO_t > 
solve( const typename Algebraic<NT>::Poly& f,
       Algebraic<NT>)
{
    typedef Algebraic<NT>          K;
    typedef typename K::RO_t       RO_t;
    
    Seq<RO_t> sol;
   
    ALGEBRAIC::solve_sturm(f, std::back_inserter(sol.rep()), K());

    std::stable_sort( sol.begin(), sol.end(), ALGEBRAIC::Refine_compare());
    return sol;
}


template < class NT,
	   class OutputIterator> inline
OutputIterator
solve( const typename Algebraic<NT>::Poly& f,
       OutputIterator it,
       Algebraic<NT>)
{
    typedef Algebraic<NT>          K;

    return ALGEBRAIC::solve_sturm(f, it, K());
}


//----------------------------------------------------------------------
template < class NT >  inline
typename Algebraic<NT>::RO_t 
solve( const typename Algebraic<NT>::Poly& f,  Algebraic<NT>, unsigned i)
{
   typedef Algebraic<NT>          K;
   return solve( f, K())[i];
}

//----------------------------------------------------------------------
template < class NT >  inline
typename Sturm<NT>::RO_t 
RootOf( const typename Algebraic<NT>::Poly& f, unsigned i, Algebraic<NT>)
{
   typedef Algebraic<NT>          K;
   return solve( f, K())[i];
}
//----------------------------------------------------------------------

template < class NT > inline
int
count_real_roots( const typename Algebraic<NT>::Poly& f, Algebraic<NT>)
{
    typedef Algebraic<NT>          K;
    typedef typename K::RT            RT;
    typedef typename K::FT            FT;
    typedef typename K::FIT           FIT;

    SturmSeq<RT> sth( f, UPOLDAR::diff(f), SUBRES());
    FT b = bound<FT>( f, Cauchy());
    FT a = -b;

    return ( variation( sth, a) - variation( sth, b) );
}

__END_NAMESPACE_SYNAPS
#endif // _SYNAPS_SOLVE_STURM_H_

