// Copyright (c) 2003  INRIA Sophia-Antipolis (France) and
//                     Department of Informatics and Telecommunications
//                     University of Athens (Greece).
// All rights reserved.
//
// Authors : Elias P. TSIGARIDAS <et@di.uoa.gr>
//           Athanasios Kakargias <grad0460@di.uoa.gr>

// Partially supported by INRIA's project "CALAMATA", a
// bilateral collaboration with National Kapodistrian University of
// Athens.
// Partially supported by the IST Programme of the EU as a Shared-cost
// RTD (FET Open) Project under Contract No  IST-2000-26473
// (ECG - Effective Computational Geometry for Curves and Surfaces)
//


#ifndef SYNAPS_USOLVE_ALGEBRAIC_CONSTRUCT_ROOTOF_H
#define SYNAPS_USOLVE_ALGEBRAIC_CONSTRUCT_ROOTOF_H

#include <synaps/usolve/algebraic/root_of.h>
#include <synaps/usolve/algebraic/Method_name.h>
#include <synaps/usolve/Small_degree.h>
#include <synaps/msolve/method_name.h>

__ALGEBRAIC_BEGIN_NAMESPACE

using namespace SYNAPS;

#if 1


/** Functions for constructions of root_of classes.
 **/

/** Construction of an algebraic number from a number @a@.
 * The the number belongs to a fraction field.
 **/
template < typename NT >
typename Small_degree<NT>::RO_t
make_root_of_1(const typename Small_degree<NT>::FT& a, 
	       const Small_degree<NT>&)
{
  typedef typename Small_degree<NT>::RT     RT;
  typedef typename Small_degree<NT>::FT     FT;
  typedef typename Small_degree<NT>::Poly   Poly;
  typedef typename Small_degree<NT>::RO_t   RO_t;

  Poly f(2, AsSize());
  f[0] = -numerator(a);
  f[1] = denominator(a);
  return RO_t(f, 0, 1, a, a);
}

/** Construction of an algebraic number  from a number @a@.
 * The number belongs to a ring.
 **/
template < typename RT >
root_of<RT, typename Small_degree<RT>::Poly >
make_root_of_1(const RT& a, const Small_degree<RT>&)
{
  typedef NumberTraits<RT>              NTR;
  typedef typename NTR::FT               FT;
  typedef typename Small_degree<RT>::Poly   Poly;

  Poly f(2, AsSize());
  f[0] = -a;
  f[1] = 1;
  //  FT bound = Bound(f, Cauchy());
  return root_of<RT, Poly>(f, 0, 1, a, a);
}


/** Construction of an algebraic number from two numbers.
 * The numbers belong to a ring. The algebraic  number is actually
 * a number that belongs to the fraction field of the ring.
 * Optionally, we can set the multiplicity and the root position
 * (in the original equation). This function is useful for the @adjust_poly()@ functions.
 **/
template < typename RT >
root_of<RT, typename Small_degree<RT>::Poly >
make_root_of_1(const RT& c0, const RT& c1,
	       const Small_degree<RT>&)
{
  typedef NumberTraits<RT>     NTR;
  typedef typename NTR::FT      FT;
  typedef typename Small_degree<RT>::Poly   Poly;

  assert(c1 != 0);

  Poly f(2, AsSize());
  f[0] = c0;
  f[1] = c1;
  FT b = to_FT(-c0, c1);

  return root_of<RT, Poly>(f, 0, RT(1), b, b);
}

template < typename NT >
typename Small_degree<NT>::RO_t
make_root_of_1(const typename Small_degree<NT>::Poly& f, const Small_degree<NT>&)
{
  typedef typename Small_degree<NT>::RT     RT;
  typedef typename Small_degree<NT>::FT     FT;
  typedef typename Small_degree<NT>::Poly   Poly;
  typedef typename Small_degree<NT>::RO_t   RO_t;

  assert(UPOLDAR::degree(f) == 1);

  FT a = to_FT(-f[0], f[1]);
  return RO_t(f, 0, 1, a, a);
}



// Currently only for algebraic numbers of degree up to 2


template < typename NT_ >
typename Small_degree<NT_>::root_t
square( const typename Small_degree<NT_>::RO_t& r)
{
  typedef Small_degree<NT_>    MTH;
  typedef typename MTH::RT     RT;
  typedef typename MTH::FT     FT;

  typedef typename MTH::Poly   Poly;
  typedef typename MTH::RO_t   RO_t;

  assert (r.degree() <= 2);

  if (r.degree() == 1) {
    FT p = r.left() * r.left();
    
    Poly f(2, AsSize());
    f[0] = -denominator(p);
    f[1] = numerator(p);
    
    return RO_t(f, p, p);
  }
  
  if (r.degree() == 2) {
        
    Poly f(3, AsSize());
    f[0] = r[0] * r[0];
    f[1] = 2 * r[2] * r[0] - r[1] * r[1];
    f[2] = r[2] * r[2];
    //            (a.is_smaller() ^ (a[1]>0)));

    Seq<RO_t> sol = Solve(f, MTH());
    if (sol.size() == 1) return sol[0];
    
    return sol[ r.which() ^ (r[1] >0) ];
  }
  
  return RO_t();
}

#endif


/// MINE
#if 0


// /** Construction of a quadratic algebraic number.
//  * The coefficients of the corresponding polynomial belong to a ring.
//  **/
// template < typename RT,
// 	       typename P >
// root_of<RT, P>
// make_root_of_2(const RT&  c0, const RT&  c1, const RT&  c2,
// 			   const int& idx,
// 			   ALG<RT, P> mth)
// {
//   assert(c2 != RT(0));

//   P f(3, AsSize());
//   f[0] = c0;
//   f[1] = c1;
//   f[2] = c2;
//   adjust_sign_coeff(f);
//   return root_of<RT, P>(f, idx);
// }


// /** Construction of cubic algebraic number.
//  * The coefficients of the corresponding polynomial belong to a ring.
//  **/
// template < typename RT,
// 	       typename P >
// root_of<RT, P>
// make_root_of_3(const RT&  c0, const RT&  c1, const RT&  c2, const RT&  c3,
// 			   const int& idx,
// 			   ALG<RT, P> mth)
// {

//   assert(c3 != RT(0));

//   P f(4, AsSize());
//   f[0] = c0;
//   f[1] = c1;
//   f[2] = c2;
//   f[3] = c3;
//   adjust_sign_coeff(f);

//   return root_of<RT, P>(f, idx);
// }


// /** Construction of a quartic algebraic number.
//  * The coefficients of the corresponding polynomial belong to a ring.
//  **/
// template < typename RT,
// 	       typename P >
// root_of<RT, P>
// make_root_of_4( const RT&  c0, const RT&  c1, const RT&  c2, const RT&  c3, const RT&  c4,
// 				const int& idx,
// 			   ALG<RT, P> mth)
// {
//   assert(c4 != RT(0));

//   P f(5, AsSize());
//   f[0] = c0;
//   f[1] = c1;
//   f[2] = c2;
//   f[3] = c3;
//   f[4] = c4;
//   adjust_sign_coeff(f);

//   return root_of<RT, P>(f, idx);
// }

// #endif
// //MINE


// #if 0
// /** Generic constructor of an algebraic number as a root of a polynomial.
//  * The coefficients of the corresponding polynomial belong to a ring.
//  **/
// template< class UPoly > inline
// root_of< typename UPoly::value_type, UPoly >
// make_root_of_x(const UPoly& p,
// 	           const int&   idx = 0)
// {
//   typedef typename UPoly::value_type  RT;

//   assert(idx < UPOLDAR::degree(p));

//   UPoly f(p);
//   adjust_sign_coeff(f);
//   return root_of<RT, UPoly>(f, idx);
// }


// /** Construction of an algebraic number that is the square root of a number @a@
//  * that belongs to a ring.
//  * The algebraic number is of the form $\sqrt{a}$.
//  **/
// template< class RT > inline
// root_of<RT>
// sqrt_of(const RT& a)
// {
//   typedef typename root_of<RT>::UPoly UPoly;
//   typedef typename root_of<RT>::FT    FT;
//   typedef Rational_traits<FT>         Rat_traits;

//   assert(a >= RT(0));

//   if (a == RT(0)) {
// 	UPoly f(2, AsSize());
// 	f[0] = RT(0);
// 	f[1] = RT(1);
// 	FT left  = Rat_traits::make_rational(RT(-1), RT(1));
// 	FT right = Rat_traits::make_rational(RT( 1), RT(1));
// 	return root_of<RT>(f, 0, ROOT_ONE, 2, RT(1), left, right, true, POS_FIRST_LAST);
//   }

//   UPoly f(3, AsSize());
//   f[0] = -a;
//   f[1] = RT(0);
//   f[2] = RT(1);

//   FT left  = Rat_traits::make_rational(RT(0), RT(1));
//   FT right = Rat_traits::make_rational(a, RT(1));
//   return root_of<RT>(f, 1, ROOT_ONE_ONE, 1, MUL(RT(4), num), left, right, true, POS_LAST);
// }



// /** Construction of an algebraic number that is the square root of a number @a@
//  * that belongs to a fraction field.
//  * The algebraic number is of the form $\sqrt{a} = \sqrt{\frac{n}{d}}$.
//  **/
// template< class FT > inline
// root_of< typename Rational_traits<FT>::RT >
// rat_sqrt_of(const FT& a)
// {
//   typedef typename Rational_traits<FT>::RT  RT;
//   typedef typename root_of<RT>::UPoly       UPoly;
//   typedef Rational_traits<FT>               Rat_traits;

//   assert(num >= FT(0));

//   if (num == FT(0)) {
// 	UPoly f(2, AsSize());
// 	f[0] = RT(0);
// 	f[1] = RT(1);

// 	FT left  = Rat_traits::make_rational(RT(-1), RT(1));
// 	FT right = Rat_traits::make_rational(RT( 1), RT(1));
// 	return root_of<RT>(f, 0, ROOT_ONE, 2, RT(1), left, right, true, POS_FIRST_LAST);
//   }

//   UPoly f(3, AsSize());
//   f[0] = -Rat_traits::numerator(a);
//   f[1] = RT(0);
//   f[2] =  Rat_traits::denominator(a);
//   RT D = RT(2) * f[2] * (-f[0]);
//   FT left  = Rat_traits::make_rational(RT(0), RT(1));
//   FT right = Rat_traits::make_rational(-f[0], f[2]);
//   return root_of<RT>(f, 1, ROOT_ONE_ONE, 1, D, left, right, true, POS_LAST);
// }


// /** Construction of an algebraic number that is the  n-th root of a number @a@
//  * that belongs to a ring.
//  * The algebraic number is of the form $\sqrt[n]{a}$.
//  **/
// template< class RT > inline
// root_of<RT>
// sqrt_of_nth(const int& nth, const RT& a)
// {
//   typedef typename root_of<RT>::UPoly UPoly;
//   typedef typename root_of<RT>::FT    FT;
//   typedef Rational_traits<FT>         Rat_traits;

//   assert(nth >= 0);

//   if (nth == 0) return make_root_of_1(RT(1));
//   if (nth == 1) return make_root_of_1(a);
//   if (nth == 2) return sqrt(a);

//   if (a == RT(0)) {
// 	UPoly f(2, AsSize());
// 	f[0] = RT(0);
// 	f[1] = RT(1);
// 	FT left  = Rat_traits::make_rational(RT(-1), RT(1));
// 	FT right = Rat_traits::make_rational(RT( 1), RT(1));
// 	return root_of<RT>(f, 0, ROOT_ONE, nth, RT(1), left, right, true, POS_FIRST_LAST);
//   }

//   RT D = pow(MUL(a * (-nth)), nth-1) * (-nth);
//   if ( nth % 2 == 0 ){
// 	assert(a > 0);

// 	UPoly f(nth + 1, AsSize());
// 	f[0] = -a;
// 	f[nth] = RT(1);

// 	FT left  = Rat_traits::make_rational(RT(0), RT(1));
// 	FT right = Rat_traits::make_rational( a,  RT(1));
// 	return root_of<RT>(f, 1, ROOT_UNKNOWN, nth, D, RT(0), a, true, POS_LAST);
//   }

//   UPoly f(nth + 1, AsSize());
//   f[0] = -a;
//   f[nth] = RT(1);

//   if (a > RT(0)) {
// 	FT left  = Rat_traits::make_rational(RT(0), RT(1));
// 	FT right = Rat_traits::make_rational( a,  RT(1));
// 	return root_of<RT>(f, 0, ROOT_UNKNOWN, 1, D, left, right, true, POS_FIRST_LAST);
//   }

//   FT left  = Rat_traits::make_rational(-a, RT(1));
//   FT right = Rat_traits::make_rational(RT(0),  RT(1));
//   return root_of<RT>(f, 0, ROOT_UNKNOWN, 1, D, left, right, true, POS_FIRST_LAST);
// }


// //TODO
// /* Construction of an algebraic number that is the  n-th root of a number
//  * that belongs to a fraction field.
//  * The algebraic number is of the form $\sqrt[n]{a} = \sqrt[n]{\frac{n}{d}}$.
//  */


#endif

__ALGEBRAIC_END_NAMESPACE

#endif // SYNAPS_USOLVE_ALGEBRAIC_ROOTOF_H


