/*******************************************************************
 *   This file is part of the source code of the SYNAPS kernel.
 *   Author(s): B. Mourrain, GALAAD, INRIA
 ********************************************************************/
#ifndef _synaps_usolve_bezier_IslBzBdg_h_
#define _synaps_usolve_bezier_IslBzBdg_h_
//===================================================================
#include <synaps/usolve/bezier/bounding.h>
//===================================================================
__BEGIN_NAMESPACE_SYNAPS
//===================================================================
template<class O>
struct IslBzBdg : public bezier::Isole<double>
{
  double eps;
  bool isole;
  unsigned iter ;
  IslBzBdg():isole(true), eps(1e-6), iter(0){}
  IslBzBdg(double e): isole(true), eps(e), iter(0) {}

  template<class VECT,class POL>
  void init_pol(VECT& ubp, VECT& dbp, const POL& r,  unsigned sz,
		bool neg = false); 

  template<class VECT,class POL,class C>
  void init_pol(VECT& ubp, VECT& dbp, const POL& r,  unsigned sz,
		const C& u, const C&v);

  template < typename POL >
  void solve(Seq<Interval<O> >& sol, const POL& r, bool neg = false);

  template < typename POL ,typename C>
  void solve(Seq<Interval<O> >& sol, const POL& r, const C& u,const C& v);
    
};

//--------------------------------------------------------------------
template<class O>
template<class VECT,class POL>
void IslBzBdg<O>::init_pol(VECT& ubp, VECT& dbp, const POL& p,unsigned sz, bool neg)
{
  typedef typename POL::value_type    RT;
  typedef typename arithm::kernelof<RT>::T kernel_t;
  typedef typename kernel_t::integer  integer_t;
  typedef typename kernel_t::rational rational_t;
  using let::assign;

 rational_t mn,mx,cf;

 assign(mx,VECTOR::max_abs(p));
 if( mx == 0 ) 
   return;
 
 std::vector<rational_t> bp( sz );
 for ( unsigned i = 0; i < sz; i ++ )
   {
     bp[i]=p[i];
     bp[i]/= binomial<integer_t>(sz-1,i);
     bp[i]/=mx;
   }
 
 if(neg)
   for(unsigned i=1;i<sz;i+=2)
     bp[i]=-bp[i];
 {
   numerics::rounding<double> rnd( numerics::rnd_up() );
   for ( unsigned i = 0; i < sz; i ++ )
     {
       assign(ubp[i],bp[i]);
     }
 };
	
    {
      numerics::rounding<double> rnd( numerics::rnd_dw() );
      for ( unsigned i = 0; i < sz; i ++ )
	{
	  assign(dbp[i],bp[i]);
	}
    };

}

//--------------------------------------------------------------------
template<class O>
template<class VECT,class POL,class C>
void IslBzBdg<O>::init_pol(VECT& ubp, VECT& dbp, const POL& r,unsigned sz,const C& u, const C&v)
{
  typedef typename POL::value_type    RT;
  typedef typename arithm::kernelof<RT>::T kernel_t;
  typedef typename kernel_t::integer  integer_t;
  typedef typename kernel_t::rational rational_t;
  using let::assign;
  
#if 0
    std::vector<rational_t> bp( sz );
    BEZIER::monomials2bezier(bp,r,sz,u,v);
    rational_t mx=VECTOR::max_abs(bp);
    VECTOR::div(bp,mx);
    {
      numerics::rounding<double> rnd( numerics::rnd_up() );
      for ( unsigned i = 0; i < sz; i ++ )
	assign(ubp[i],bp[i]);
    }
    {
      numerics::rounding<double> rnd( numerics::rnd_up() );
      for ( unsigned i = 0; i < sz; i ++ )
	assign(dbp[i],bp[i]);
    };
    
#else
    O ud, vd;
    {
      numerics::rounding<double> rnd( numerics::rnd_dw() );
      assign(ud,u);
    };
    {
      numerics::rounding<double> rnd( numerics::rnd_up() );
      assign(vd,v);
      BEZIER::monomials2bezier(ubp,r,r.size(),ud,vd);
    };
    {
      numerics::rounding<double> rnd( numerics::rnd_dw() );
      BEZIER::monomials2bezier(dbp,r,r.size(),ud,vd);
    };
#endif
    for ( unsigned i= 0; i < ubp.size(); i ++ )
      if ( dbp[i] > ubp[i] ) { 
	std::swap(dbp[i], ubp[i]);
      };
}
//--------------------------------------------------------------------
template<class O>
template < typename Poly >
void IslBzBdg<O>::solve(Seq<Interval<O> >& sol, const Poly& r, bool neg)
{
    typedef typename Poly::value_type    RT;
    typedef typename arithm::kernelof<RT>::T kernel_t;
    typedef typename kernel_t::integer  integer_t;
    typedef typename kernel_t::rational rational_t;
    typedef Interval<O> sol_t;
    using let::assign;

    unsigned sz=r.size();
    std::vector<double> ubp( sz );
    std::vector<double> dbp( sz );


    this->init_pol(ubp,dbp,r,sz,neg);

    for ( unsigned i= 0; i < sz; i ++ )
      {
	if ( dbp[i] == ubp[i] && dbp[i] !=0 )  
	  {
	    //	    std::cout<<"Pb"<<std::endl;
	    if(dbp[i]>0) 
	      dbp[i]*=(1-1e-15);
	    else
	      dbp[i]*=(1+1e-15);

	    if(ubp[i]>0) 
	      ubp[i]*=(1+1e-15);
	    else
	      ubp[i]*=(1-1e-15);
	  }
	
      }
    
    this->Run(ubp,dbp,eps/16,true);

    Seq<O> sl;
    this->GetIntervals(sl);

    O mx;
    assign(mx, SYNAPS::bound<rational_t>(r,Cauchy()));
    if(neg)
      RConvert(sl,sl.size(),mx);
    else
      RConvert(sl,0,mx);
    
    for(unsigned i = 0; i < sl.size(); i += 2 )
      {
	if (this->m_res[i/2].t || (sl[i+1]-sl[i])< this->eps )
	  sol.push_back( sol_t(sl[i],sl[i+1]) );
	else 
	  {
	    //std::cout <<"\t"<<sl[i]<<" "<<sl[i+1]<<std::endl;
	    this->solve(sol,r,sl[i],sl[i+1]);
	  }
      }
}

template <class O>
template <typename POL ,typename C>
void IslBzBdg<O>::solve(Seq<Interval<O> >& sol, const POL& r, const C& u,const C& v)
{
    typedef typename POL::value_type    RT;
    typedef typename arithm::kernelof<RT>::T kernel_t;
    typedef typename kernel_t::integer  integer_t;
    typedef typename kernel_t::rational rational_t;    
    typedef Interval<O> sol_t;
    
    using let::assign;

    unsigned sz= r.size();
    double ud,vd,ad,bd, eps;
    std::vector<double> ubp( sz );
    std::vector<double> dbp( sz );

    this->init_pol(ubp,dbp,r,sz,u,v);
    this->Run(ubp,dbp,this->eps,true);

    Seq<O> sl;
    this->GetIntervals(sl);
    RConvert(sl,u,v);
    for(unsigned i = 0; i < sl.size(); i += 2 )
      {
	if (this->m_res[i/2].t )
	  sol.push_back( sol_t(sl[i],sl[i+1]) );
	else 
	  {
	    this->solve(sol,r,sl[i],sl[i+1]);
	  }
      }
}
//--------------------------------------------------------------------

template<class O, class POL> 
Seq<Interval<O> > solve(const POL& p, IslBzBdg<O> mth)
{
  Seq<Interval<O> > sol;
  mth.solve(sol,p,true);
  mth.solve(sol,p);
  return sol;
}

template<class O, class POL,class C> 
Seq<Interval<O> > solve(const POL& p, IslBzBdg<O> mth,const C& a, const C& b)
{
  Seq<Interval<O> > sol;
  mth.solve(sol,p,a,b);
  return sol;
}

//===================================================================
__END_NAMESPACE_SYNAPS
#endif //_synaps_usolve_bezier_IslBzBdg_h_
