
#ifndef __SYNAPS_SOLVE_STURM_SOLVE_D3_H__
#define __SYNAPS_SOLVE_STURM_SOLVE_D3_H__


#include <synaps/init.h>
#include <synaps/usolve/algebraic/Algebraic_config.h>
#include <synaps/usolve/Util.h>
#include <synaps/upol/bound.h>
#include <synaps/usolve/sturm/construct_root_of.h>


__ALGEBRAIC_BEGIN_NAMESPACE


//! Computes the roots of a polynomial of degree 3
template < typename K_,
	   typename OutputIterator >
void
solve_adjust_poly_3(const typename K_::Poly& f,
		    OutputIterator  sol,
		    unsigned mult,
		    K_)
{
    typedef K_                 K;
    typedef typename K::RT     RT;
    typedef typename K::FT     FT;
    typedef typename K::Poly   Poly;
    typedef typename K::RO_t   RO_t;
    
    //DEBUG::print(__FUNCTION__);
    RT a = RT( 3) * f[3];
    RT b = RT(-1) * f[2];
    RT c = RT( 1) * f[1];
    RT d = RT(-3) * f[0];
    RT D2x2;
    RT D2 = calc_D2(a, b, c);
    RT D3 = calc_D2(b, c, d);
    RT W = calc_brackets(b, c, a, d);
    RT D2x4 = MUL( RT(4), D2 );
    RT D = calc_D2(D2x4, W, D3);

    D2x2 = MUL( RT(2), D2 );
    if (D > RT(0)) {
	//	r.mu_ = ROOT_ONE_COMPLEX_COMPLEX;
	if (f[0] > RT(0)) { //the real root is negative
	    FT lb = -bound<FT>(f, Cauchy());
	    FT ub = to_FT(RT(0));
	    *sol++ = RO_t(f, 0, D, lb, ub, 1*mult);
	    return;
	}
	if (f[0] < RT(0)) { // the real root is positive
	    FT lb = to_FT(RT(0));
	    FT ub = bound<FT>(f, Cauchy());
	    *sol++ = RO_t(f, 0, D, lb, ub, 1*mult);
	    return;
	}
	//else if (r[0] == RT(0))  // the real root is zero
	RO_t r;
	r.set_multiplicity( mult);
	*sol++ = r;
	return;
    }
    if (D < RT(0)) {
	//	r.mu_ = ROOT_ONE_ONE_ONE;
	D2x2 = MUL( RT(2), D2 );
	RT f_ba = calc_F3_BA(a, b, W, D2x2);
	if (f_ba > RT(0)) {
	    FT is3 = bound<FT>(f, Cauchy());
	    FT is2 = to_FT(W, D2x2);
	    FT is1 = to_FT(b, a);
	    FT is0 = -is3;
	    *sol++ = RO_t( f, 0, D, is0, is1, 1*mult);
	    *sol++ = RO_t( f, 1, D, is1, is2, 1*mult);
	    *sol++ = RO_t( f, 2, D, is2, is3, 1*mult);
	    return;
	}
	if (f_ba < RT(0)) {
	    FT is3 = bound<FT>(f, Cauchy());
	    FT is2 = to_FT(b, a);
	    FT is1 = to_FT(W, D2x2);
	    FT is0 = -is3;
	    *sol++ = RO_t( f, 0, D, is0, is1, 1*mult);
	    *sol++ = RO_t( f, 1, D, is1, is2, 1*mult);
	    *sol++ = RO_t( f, 2, D, is2, is3, 1*mult);
	    return;
	}
	// if b/a is a root
	FT rb = bound<FT>(f, Cauchy());
	FT lb = -rb;
	Poly p(3, AsSize());
	p[2] = SQR(RT(3 * f[3]));
	p[1] = 6 * f[3] * f[2];
	p[0] = 9 * f[3] * f[1] - 2 * SQR(f[2]);
	RT DD = SQR(p[1]) - 4 * p[2] * p[0];

	FT mid = to_FT(b, a);
	Poly h(2, AsSize());
	h[0] = -b;
	h[1] = a;

	*sol++ = RO_t(p, 0, DD, lb, mid, 1*mult);
	*sol++ = RO_t(h, 0, RT(1), mid, mid, 1*mult);
	*sol++ = RO_t(p, 1, DD, mid, rb, 1*mult);
	return;
    }
    // if (D1 == RT(0))
    // If the cubic has multiple roots then compute them as rationals
    if (D2 == RT(0)) {
	// The cubic has one triple root
	Poly p(2, AsSize());
	p[0] = -b;
	p[1] = a;
	FT isol = to_FT(b, a);
	*sol++ = RO_t(p, 0, RT(1), isol, isol, 3*mult);
	return;
    }
    //	if (D2 != 0)
    RT P = calc_P(a, W, b, D2);
    RT Dr = calc_brackets(RT(2), P, a, W);
    if (Dr > RT(0)) {
	//  r.mu_ = ROOT_DOUBLE_ONE;
	Poly p(2, AsSize());
	p[0] = -W;
	p[1] = D2x2;
	FT isol = to_FT(-p[0], p[1]);
	*sol++ = RO_t(p, 0, RT(1), isol, isol, 2*mult);

	p[0] = -P;
	p[1] = a * D2;
	isol = to_FT(-p[0], p[1]);
	*sol++ = RO_t(p, 0, RT(1), isol, isol, 1*mult);
	return;
    }
    //	 else
    //  r.mu_ = ROOT_ONE_DOUBLE;
    Poly p(2, AsSize());
    p[0] = -P;
    p[1] = a * D2;
    FT isol = to_FT(-p[0], p[1]);
    *sol++ = RO_t(p, 0, RT(1), isol, isol, 1*mult);

    p[0] = -W;
    p[1] = D2x2;
    isol = to_FT(-p[0], p[1]);
    *sol++ = RO_t(p, 0, RT(1), isol, isol, 2*mult);

    return;
}




__ALGEBRAIC_END_NAMESPACE


#endif // __SYNAPS_SOLVE_STURM_SOLVE_D3_H__
