
#ifndef __SYNAPS_SOLVE_STURM_SOLVE_STURM_H__
#define __SYNAPS_SOLVE_STURM_SOLVE_STURM_H__

#include <synaps/init.h>
#include <synaps/upol/SturmSeq.h>
#include <synaps/upol/square_free.h>
#include <synaps/upol/bound.h>
#include <synaps/usolve/sturm/common.h>
#include <synaps/usolve/sturm/solve_small.h>
#include <synaps/usolve/Sturm.h>
#include <synaps/usolve/sturm/construct_root_of.h>

#include <stack>

__ALGEBRAIC_BEGIN_NAMESPACE



template < typename FT >
struct Quadtuple {
    FT   a_;
    FT   b_;
    int  Va_;
    int  Vb_;

    Quadtuple(const FT& a, const FT& b, int Va, int Vb)
	: a_(a), b_(b), Va_(Va), Vb_(Vb)
	{ }

    const FT& a() const { return a_; }
    const FT& b() const { return b_; }
    int Va() const { return Va_; }
    int Vb() const { return Vb_; }
};



template < typename RT > inline
UPolDse<RT>
translate( const UPolDse<RT>& f, const typename rationalof<RT>::T& q)
{
    typedef UPolDse<RT>          Poly;
    //    typedef NumberTraits<RT>    NTR;
    typedef typename rationalof<RT>::T  FT;

    FT b = numerator(q);
    FT c = denominator(q);

    Poly h(2, AsSize());
    h[1] = c;
    h[0] = b;

    int p = UPOLDAR::degree(f);
    RT d(1);
    Poly T0(1, AsSize());
    T0[0] = f[p];
    Poly(T1);

    for (int i = 1; i <= p; ++i) {
	d *= c;
	T1 = h * T0 + d * f[p - i];
	std::swap(T0, T1);
    }
    return T0;
}




template < typename K,
	   typename OutputIterator > inline
OutputIterator
sturm_isolation( const SturmSeq< typename K::RT >& sth,
		 const typename K::FIT& I, 
		 OutputIterator isol,
		 K)
{
    typedef typename K::RT    RT;
    typedef typename K::FT    FT;
    typedef typename K::FIT   FIT;
    typedef typename K::Poly  Poly;
    
    typedef Quadtuple<FT>   Quad;

    std::stack<Quad>   Work;
    std::vector<FIT>   result;

    FT a = lower(I);
    FT b = upper(I);

    Poly h = sth[0];

    Poly x(2, AsSize()); x[1] = RT(1);

    int sgn_a = ::SYNAPS::sign( homogeneous_eval_horner<K>( h, a));
    
    if ( singleton(I) ) {
	if ( sgn_a  == ZERO) {
	    *isol++ = I;
	}
	return isol;
    }
    
    if ( sgn_a == ZERO ) {
	*isol++ = FIT( a, a);
	
	Poly f = translate(sth[0], a)/x;
	FT M = min_bound<FT>(f, Cauchy());
	a = a + M;
    }
    

    int Va = VAR<K>(sth, a);
    int Vb = VAR<K>(sth, b);
    int V = Va - Vb;

    if (V == 0) return isol;
    if (V == 1) {
	*isol++ = FIT(a, b);
	return isol;
    }
    Work.push( Quad(a, b, Va, Vb) );


    while ( !Work.empty() ) {
	Quad q = Work.top();
	Work.pop();
    
	FT c = (q.a() + q.b())/2;

//	std::cout << "Search: " << a << ", " << c << ", " << b << std::endl; 
	
	if ( ::SYNAPS::sign( homogeneous_eval_horner<K>( sth[0], c)) == ZERO ) {
	    *isol++ = FIT( c, c);
	    
	    //    result.push_back( FIT(c, c) );

	    Poly t(2, AsSize());
	    t[1] = denominator(c);
	    t[0] = -numerator(c);
	    h = EUCLIDIAN::pquo(h, t);

	    Poly f = translate(sth[0], c)/x;
	    FT M = min_bound<FT>(f, Cauchy());
	    //int Vc_plus = Count(sth, FT(c + M));
	    int Vc_plus = VAR<K>(sth, FT(c + M));
	    //int Vc_minus = Var(sth, FT(c - M));
	    int Vc_minus = Vc_plus + 1;

	    if (q.Va() == Vc_minus + 1) {
		result.push_back( FIT(q.a(), FT(c - M)) );
	    }
	    if (q.Va() > Vc_minus + 1) {
		Work.push( Quad(q.a(), FT(c - M), q.Va(), Vc_minus) );
	    }
	    if (q.Vb() == Vc_plus - 1) {
		result.push_back( FIT(FT(c + M), q.b()) );
	    }
	    if (q.Vb() < Vc_plus - 1) {
		Work.push( Quad(c + M, q.b(), Vc_plus, q.Vb()) );
	    }
	} else {
	    int Vc = VAR<K>(sth, c);
	    if (q.Va() == Vc + 1) {
		result.push_back( FIT(q.a(), c) );
	    }
	    if (q.Vb() == Vc - 1) {
		result.push_back( FIT(c, q.b()) );
	    }

	    if (q.Vb() < Vc - 1) {
		Work.push( Quad(c, q.b(), Vc, q.Vb()) );
	    }
	    if (q.Va() > Vc + 1) {
		Work.push( Quad(q.a(), c, q.Va(), Vc) );
	    }
	}

    }
    
    typename std::vector<FIT>::const_iterator it;
    for (it = result.begin(); it != result.end(); ++it) {
	*isol++ = *it;
    }
    
    return isol;
}



template < typename K,
	   typename OutputIterator > inline
OutputIterator
sturm_rootof( const SturmSeq< typename K::RT>& sth,
	      const typename K::FIT& I, 
	      unsigned mult, OutputIterator sol, K)
{
    typedef typename K::RT    RT;
    typedef typename K::FT    FT;
    typedef typename K::FIT   FIT;
    typedef typename K::RO_t  RO_t;
    typedef typename K::Poly  Poly;
    
    std::vector<FIT> isol;
    
    sturm_isolation( sth, I, std::back_inserter( isol), K());
    sort( isol.begin(), isol.end());
    
    
    typename std::vector<FIT>::const_iterator it;
    for (it = isol.begin(); it != isol.end(); ++it) {
	if (singleton(*it) ) {
	    RO_t r = construct_root_of( RT(-numerator(it->lower())), denominator(it->lower()), K());
	    r.set_multiplicity(mult); 
	    *sol++ = r;
	} else {
	    *sol++ = RO_t(sth[0], it->lower(), it->upper(), mult);
	}
    }
    return sol;
}



template < typename K,
	   typename OutputIterator > inline
OutputIterator
sturm_rootof( const typename K::Poly& F, unsigned mult, OutputIterator sol, K)
{
    typedef typename K::RT    RT;
    typedef typename K::FT    FT;
    typedef typename K::FIT   FIT;
    typedef typename K::RO_t  RO_t;
    typedef typename K::Poly  Poly;

    SturmSeq<RT> sth(F, UPOLDAR::diff(F), SUBRES());
    
    FT b = max_bound<FT>(sth[0], Cauchy());
    FT a = -b;
    FIT I( a, b);
    
    return sturm_rootof( sth, I, mult, sol, K());
}





template < typename K,
           typename OutputIterator > inline
OutputIterator
solve_sturm( const typename K::Poly& h, OutputIterator sol, K)
       
{
    typedef typename K::RT    RT;
    typedef typename K::FT    FT;
    typedef typename K::FIT   FIT;
    typedef typename K::RO_t  RO_t;
    typedef typename K::Poly  Poly;
    if (UPOLDAR::degree(h) < 5) {
	return solve_small_degree(h, 1, sol, K());
    }
    // Check if x = 0, is a root
    int mult = 0;    
    if ( h[0] == RT(0)) {
	mult = 1;
	RO_t r = construct_root_of( RT(0), RT(1), K());
	
	while (h[mult] == RT(0)) {
	    r.set_multiplicity( ++mult); 
	}
	*sol++ = r;
    }
    int dg = UPOLDAR::degree( h) - mult;
    
    Poly f( dg+1, AsSize());
    for (int i = 0; i <= dg; ++i) { f[i]  = h[i+mult]; }

    SturmSeq<RT>   sth;
    Seq<Poly> PL = square_free_factorization( f, sth);
    if (PL.size() == 1) {
	if (UPOLDAR::degree(PL[0]) < 5) {
	    solve_small_degree( PL[0], 1, sol, K());
	} else {
	  FT b = max_bound<FT>( sth[0], Cauchy());
	  FT a = -b;
	  FIT I( a, b);
	  sturm_rootof( sth, I, 1, sol, K());
	}
	return sol;
    } else {
	for (int i = 0; i < PL.size(); ++i) {
	    if (UPOLDAR::degree( PL[i]) < 5) {
		solve_small_degree( PL[i], i+1, sol, K());
	    } else {
		sturm_rootof( PL[i], i+1, sol, K());
	    }
	}	
    }
    
    return sol; 
}


__ALGEBRAIC_END_NAMESPACE

#endif //__SYNAPS_SOLVE_STURM_SOLVE_STURM_H__
