
#ifndef _SYNAPS_SOLVE_STURM_STURM_BEZIER_SOLVER_H_
#define _SYNAPS_SOLVE_STURM_STURM_BEZIER_SOLVER_H_


#include <synaps/init.h>
#include <synaps/usolve.h>
#include <synaps/arithm/Interval.h>
#include <synaps/usolve/Aberth.h>
#include <synaps/usolve/sturm/common.h>
#include <synaps/usolve/sturm/solve_sturm.h>
#include <synaps/usolve/sturm/construct_root_of.h>
#include <synaps/usolve/bezier/IslBzBdg.h>

__ALGEBRAIC_BEGIN_NAMESPACE


template < typename K,
	   typename OutputIterator > inline
OutputIterator
solve_sturm_bezier( const typename K::Poly& h, OutputIterator sol, K)
{
    typedef typename K::RT           RT;
    typedef typename K::FT           FT;
    typedef typename K::XT           XT;
    typedef typename K::RL           RL;

    typedef typename K::FIT          FIT;
    typedef typename K::XIT          XIT;
    typedef typename K::RIT          RIT;
  
    typedef typename K::Poly         Poly;
    typedef typename K::RO_t         RO_t;
    typedef typename K::solution_type         solution_type;
  
  
    if (UPOLDAR::degree(h) < 5) {
	solve_small_degree(h, 1, sol, K());
	return sol;
    }

    // Check if x = 0, is a root
    int mult = 0;    
    if ( h[0] == RT(0)) {
	mult = 1;
	RO_t r = construct_root_of( RT(0), RT(1), K());
	
	while (h[mult] == RT(0)) {
	    r.set_multiplicity( ++mult); 
	}
	*sol++ = r;
    }
    int dg = UPOLDAR::degree( h) - mult;
    
    Poly f( dg+1, AsSize());
    for (int i = 0; i <= dg; ++i) { f[i]  = h[i+mult]; }

  
    
    IslBzInteger<XT> Bzsolver;
    Seq<XIT> apxI = solve( f, Bzsolver);
    Seq<FT> exIsol;

    bool sth_computed = false;
    
    // Compute "simple" rational isolating points
    if ( !(apxI.size() == 0) ) {
      exIsol.resize( apxI.size() + 1);
      
      // the first point
      exIsol[0] =  
	simplest_rational_in_interval<K>( lower( apxI[0]) - 1, lower( apxI[0]));
      
      // the last point
      exIsol[exIsol.size()-1]
	= simplest_rational_in_interval<K>( upper( apxI[apxI.size()-1]), 
					    upper( apxI[apxI.size()-1]) + 1);
      
      // the middle points
      for (int j = 0; j < apxI.size()-1; ++j) {
	exIsol[j+1] = simplest_rational_in_interval<K>( upper( apxI[j]), lower( apxI[j+1]));
      }
      
    }
    
    //construct the isolating interval
    Seq<FIT> SI ( apxI.size() );
    for (int j = 0; j < apxI.size(); ++j) {
      if ( (width( apxI[j]) > Bzsolver.eps) || singleton( apxI[j]) ) {
	SI[j] = FIT( lower( apxI[j]), upper( apxI[j]));
	continue;
      }
      
      SI[j] = FIT( exIsol[j], exIsol[j+1]);
    }
    

    // Isolate the real roots, using Sturm sequeneces
    for (int j = 0; j < SI.size(); ++j) {
      if ( singleton( SI[j]) ) {
	RO_t r = construct_root_of( lower( SI[j]) , K());
	*sol++ = r;
	continue;
      }
      if ( width( apxI[j]) > Bzsolver.eps ) {
	RO_t r = construct_root_of( f, SI[j] , K());
	*sol++ = r;
	continue;	    
      }
      SturmSeq<RT> sth;
      if (sth_computed == false) {
	Seq<Poly> PL = SquareFreeFactorization( f, sth);
	sth_computed = true;
      }
      sturm_rootof( sth, SI[j],  1, sol, K());
    }
    
    return sol;
}



#if 0
template < typename K,
	   typename OutputIterator > inline
OutputIterator
solve_sturm_bezier( const typename K::Poly& h, OutputIterator sol, K)
{
    typedef typename K::RT           RT;
    typedef typename K::FT           FT;
    typedef typename K::XT           XT;
    typedef typename K::RL           RL;

    typedef typename K::FIT          FIT;
    typedef typename K::XIT          XIT;
    typedef typename K::RIT          RIT;
  
    typedef typename K::Poly         Poly;
    typedef typename K::RO_t         RO_t;
    typedef typename K::solution_type         solution_type;
  
  
    if (UPOLDAR::degree(h) < 5) {
	solve_small_degree(h, 1, sol, K());
	return sol;
    }

    // Check if x = 0, is a root
    int mult = 0;    
    if ( h[0] == RT(0)) {
	mult = 1;
	RO_t r = construct_root_of( RT(0), RT(1), K());
	
	while (h[mult] == RT(0)) {
	    r.set_multiplicity( ++mult); 
	}
	*sol++ = r;
    }
    int dg = UPOLDAR::degree( h) - mult;
    
    Poly f( dg+1, AsSize());
    for (int i = 0; i <= dg; ++i) { f[i]  = h[i+mult]; }

    SturmSeq<RT> sth;
    Seq<Poly> PL = SquareFreeFactorization( f, sth);

    for (int i = 0; i < PL.size(); ++i) {
      if (UPOLDAR::degree( PL[i]) < 1) continue;
      
      if (UPOLDAR::degree( PL[i]) < 5) {
	solve_small_degree( PL[i], i+1, sol, K());
	continue;
      }

      IslBzInteger<XT> Bzsolver;
      Seq<XIT> apxI = solve( PL[i], Bzsolver);
      Seq<FT> exIsol;

      // Compute "simple" rational isolating points
      if ( !(apxI.size() == 0) ) {
	exIsol.resize( apxI.size() + 1);
      
	// the first point
	exIsol[0] =  
	  simplest_rational_in_interval<K>( lower( apxI[0]) - 1, lower( apxI[0]));

	// the last point
	exIsol[exIsol.size()-1]
	  = simplest_rational_in_interval<K>( upper( apxI[apxI.size()-1]), 
					      upper( apxI[apxI.size()-1]) + 1);
      
	// the middle points
	for (int j = 0; j < apxI.size()-1; ++j) {
	  exIsol[j+1] = simplest_rational_in_interval<K>( upper( apxI[j]), lower( apxI[j+1]));
	}
      
      }
    
      //construct the isolating interval
      Seq<FIT> SI ( apxI.size() );
      for (int j = 0; j < apxI.size(); ++j) {
	if ( (width( apxI[j]) > Bzsolver.eps) || singleton( apxI[j]) ) {
	  SI[j] = FIT( lower( apxI[j]), upper( apxI[j]));
	  continue;
	}
      
	SI[j] = FIT( exIsol[j], exIsol[j+1]);
      }

      // Do we need to recompute the Sturm sequence ?
      if (PL.size() == 1) {
	;
      } else {
	sth = SturmSeq<RT>( PL[i], UPOLDAR::diff( PL[i]), SUBRES());
      }
    
      // Isolate the real roots, using Sturm sequeneces
      for (int j = 0; j < SI.size(); ++j) {
	if ( singleton( SI[j]) ) {
	  RO_t r = construct_root_of( lower( SI[j]) , K());
	  r.set_multiplicity( i);
	  *sol++ = r;
	  continue;
	}
	if ( width( apxI[j]) > Bzsolver.eps ) {
	  RO_t r = construct_root_of( PL[i], SI[j] , K());
	  r.set_multiplicity( i+1);
	  *sol++ = r;
	  continue;	    
	}
	sturm_rootof( sth, SI[j],  i+1, sol, K());
      }
    }
    
    return sol;
}

#endif
__ALGEBRAIC_END_NAMESPACE

#endif // _SYNAPS_SOLVE_STURM_STURM_BEZIER_SOLVER_H_

